/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */

#include "generaldef.h"


#include <string.h>
#include <string>
#include <fstream>
#include <strstream>
#include <time.h>
#include "mzwrap.h"
#include "pipeline3d.h"

#include "hpc.h"
#include "xcglue.h"

static Scheme_Object *polComplex;


//predeclarations of all functions
static int FFI_BuildPolComplexCube,FFI_MkPolComplex,FFI_BuildPolComplexSimplex,FFI_StructPolComplex;
static int FFI_GetSpatialDimension,FFI_GetPointDimension;
static int FFI_ScalePolComplex,FFI_TranslatePolComplex,FFI_RotatePolComplex,FFI_ShearPolComplex,FFI_PolComplexMat;
static int FFI_SetNamePolComplex,FFI_AddPropertyPolComplex,FFI_GetPropertyPolComplex;
static int FFI_EmbedPolComplex;
static int FFI_DumpPolComplex;
static int FFI_GetNum,FFI_GetFirst,FFI_GetSucc;
static int FFI_Integrate,FFI_GetLimits;
static int FFI_Join,FFI_Power,FFI_Skeleton,FFI_MapPolComplex,FFI_MapPolComplex2;
static int FFI_CopyPolComplex;
static int FFI_Or,FFI_And,FFI_Xor,FFI_Diff,FFI_PDiff;
static int FFI_ExportVRML,FFI_ExportSVG,FFI_ExportSVG3D,FFI_Pipeline3D,FFI_ExportXML,FFI_XMLImport,FFI_HpcConfig;
static int FFI_UkPolComplex,FFI_UkPolComplexf;
static int FFI_ViewOpenGL;
static int FFI_PovRay;
static int FFI_BenchmarkCount,FFI_BenchmarkReset,FFI_BenchmarkLog,FFI_BenchmarkTime;
static int FFI_PolComplexFinilizer;


typedef struct 
{
	int   code;
	char* class_name;
	char* ffi_name;
	Scheme_Object* (*fn)(int argc, Scheme_Object **argv);
	int   minarg,maxarg;
}
mzwrap_fn_t;

static int				number_of_registered_hpc=0;
static int				table_dim =0;
static mzwrap_fn_t	table_vec [1024];
static int           table_msec[1024],table_global_msec[1024];
static int           table_gross_msec[1024],table_gross_global_msec[1024];
static int           table_call[1024],table_global_call[1024];
static int           total_call=0    ,total_global_call=0;
static int           total_msec=0    ,total_global_msec=0;
static int           total_gross_msec=0    ,total_gross_global_msec=0;


#define 	ffi_type_setup() 



void PolComplexFinilizer(void *p, void *data) ;

///////////////////////////////////////////////////////////////////////////////////////////

class ffi_type
{
	int code;
	int narg;
	int argc;
	Scheme_Object** argv;
	clock_t t1;

	inline const char* fName()
		{return code>=0?::table_vec[code].ffi_name:"<noname>";}

public:

	int outside_time;

	//from a command arg
	inline ffi_type(int _code,int _argc,Scheme_Object** _argv)
		:code(_code),t1(0),outside_time(0),narg(0),argc(_argc),argv(0)
	{
		if (argc<0)
			scheme_signal_error("empty list in %s function",fName());

		if (code>=0) t1=clock();

		if (argc)
		{
			argv=new Scheme_Object*[argc];
			for (int i=0;i<argc;i++) argv[i]=_argv[i];
		}
	}

	//copy constructor
	inline ffi_type(const ffi_type& src)
		:code(src.code),narg(0),argc(src.argc),argv(0),t1(0)
	{
		if (argc)
		{
			this->argv =new Scheme_Object*[argc];
			for (int i=0;i<argc;i++) this->argv[i]=src.argv[i];
		}
	}

	//from a scheme_object-list
	inline ffi_type(int code,Scheme_Object* list)
		:code(code),narg(0),argv(0),t1(0)
	{
		if (!SCHEME_LISTP(list))
			scheme_signal_error("invalid list %V in %s function",list,fName());

		this->argc=scheme_proper_list_length(list);

		if (this->argc<0)
			scheme_signal_error("empty list in %s function",fName());

		if (argc)
		{
			this->argv=new Scheme_Object*[this->argc];

			for (int i=0;i<this->argc;i++,list = scheme_cdr(list))
				this->argv[i]=scheme_car(list);
		}
	}

	//destructor
	inline ~ffi_type()
	{
		if (this->argv)
			delete [] this->argv;

		if (t1 && code>=0)
		{
			int msec=((int)(((clock()-t1)/(double)CLOCKS_PER_SEC)*1000))-outside_time;

			table_msec[code]+=msec;
			table_global_msec[code]+=msec;
			
						table_gross_msec[code]+=msec+outside_time;
			table_gross_global_msec[code]+=msec+outside_time;
			
			table_call[code]++;
			table_global_call[code]++;

			total_msec+=msec;
			total_global_msec+=msec;
			
			total_gross_msec+=msec+outside_time;
			total_gross_global_msec+=msec+outside_time;

			total_call++;
			total_global_call++;
		}
	}

	//check number of arguments
	inline void checkNumArgs(int minc,int maxc)
	{
		if (minc!=-1 && !(argc>=minc))
			scheme_wrong_count(fName(),minc,maxc,argc,argv);

		if (maxc!=-1 && !(argc<=maxc))
			scheme_wrong_count(fName(),minc,maxc,argc,argv);
	}

	//remaining arg
	inline int remaining()
		{return argc-narg;}

	//access the real value
	static inline real getReal(Scheme_Object* obj)
	{
		if (SCHEME_INTP(obj))		return (real)SCHEME_INT_VAL(obj);
		if (SCHEME_DBLP(obj))		return (real)SCHEME_DBL_VAL(obj);
		if (SCHEME_RATIONALP(obj)) return (real)scheme_rational_to_double(obj);
		return 0;
	}

	inline int getInt()
	{
		if (narg>=argc)
			scheme_signal_error("too few paramenters for %s function",fName());

		if (!SCHEME_INTP(argv[narg]))
			scheme_wrong_type(fName(),"integer",narg,argc,argv);

		return SCHEME_INT_VAL(argv[narg++]);
	}

	inline char* getString()
	{
		if (narg>=argc)
			scheme_signal_error("too few paramenters for %s function",fName());

		if (SCHEME_BYTE_STRINGP(argv[narg]))
		{
			return SCHEME_BYTE_STR_VAL(argv[narg++]);
		}
		else if (SCHEME_CHAR_STRINGP(argv[narg]))
		{
			return SCHEME_BYTE_STR_VAL(scheme_char_string_to_byte_string(argv[narg++]));
		}
		else if (SCHEME_SYMBOLP(argv[narg]))
		{
			return SCHEME_SYM_VAL(argv[narg++]);
		}

		scheme_wrong_type(fName(),"string",narg,argc,argv);
		return 0;
	}


	inline real getReal()
	{
		if (narg>=argc)
			scheme_signal_error("too few paramenters for %s function",fName());

		Scheme_Object* obj=argv[narg];

		if (!SCHEME_INTP(obj) && !SCHEME_DBLP(obj) && !SCHEME_RATIONALP(obj))
			scheme_wrong_type(fName(),"real",narg,argc,argv);

		narg++;

		if (SCHEME_INTP(obj))
			return (real)SCHEME_INT_VAL(obj);

		if (SCHEME_DBLP(obj))
			return (real)SCHEME_DBL_VAL(obj);
		
		return (real)scheme_rational_to_double(obj);
	}

	inline Hpc* getHpc()
	{
		if (narg>=argc)
			scheme_signal_error("too few paramenters for %s function",fName());

		if (!objscheme_is_a(argv[narg],polComplex))
			scheme_wrong_type(fName(),"pol-complex%",narg,argc,argv);

		Hpc* ret=(Hpc*)OBJSCHEME_GET_CPP_OBJ(argv[narg++]);
		return ret;
	}

	inline Scheme_Object* getObj(bool bAdvance=true)
	{
		if (narg>=argc)
			scheme_signal_error("too few paramenters for %s function",fName());

		Scheme_Object* ret=argv[narg];
		if (bAdvance)
			narg++;

		return ret;
	}

	inline Scheme_Object* getFunction()
	{
		 return getObj();
	}


	ffi_type getList()
	{
		if (narg>=argc)
			scheme_signal_error("too few paramenters for %s function",fName());
			
		Scheme_Object* list=argv[narg++];

		if (!SCHEME_LISTP(list))
			scheme_signal_error("invalid list %V in %s function",list,fName());

		return ffi_type(code,list);
	}


	static inline Scheme_Object* wrap(int value)
		{return scheme_make_integer_value(value);}

	static inline Scheme_Object* wrap(real value)
	{
	#if 1
		value = floor( 1e6l * value + 0.5l ) / 1e6l;
	#endif
		return scheme_make_double(value);
	}

	static inline Scheme_Object* wrap(const char* value)
	{
		return scheme_byte_string_to_char_string(scheme_make_byte_string(value));
	}

	static inline Scheme_Object* wrap(Hpc* what,Scheme_Object* obj=0)
	{
		if (!obj)
		{
			obj= objscheme_make_uninited_object(polComplex);
			++number_of_registered_hpc;
		}
		else
		{
			scheme_remove_all_finalization(obj);
		}

		OBJSCHEME_SET_CPP_OBJ(obj,(Scheme_Object *)what);
		scheme_add_finalizer(obj, PolComplexFinilizer, 0);//for deallocation
		return obj;
	}
};

///////////////////////////////////////////////////////////////////////////////////////

Scheme_Object *BenchmarkCount (int argc, Scheme_Object **argv)
{
	return ffi_type::wrap(number_of_registered_hpc);
}

Scheme_Object *BenchmarkReset (int argc, Scheme_Object **argv)
{
	total_msec=0;
	total_gross_msec=0;
	total_call=0;
	memset(table_msec,0,sizeof(table_msec));
	memset(table_gross_msec,0,sizeof(table_msec));
	memset(table_call,0,sizeof(table_call));
	return scheme_void;
}

Scheme_Object *BenchmarkLog(int argc, Scheme_Object **argv)
{
	if (argc!=1)
		scheme_wrong_count("hpc-benchmark-log",1,1,argc,argv);

	if (!SCHEME_BOOLP(argv[0]))
		scheme_wrong_type("hpc-benchmark-log","boolean",0,argc,argv);

	bool bLocal=SCHEME_FALSEP(argv[0]);

	FILE* file=stdout;
	for (int i=0;i<table_dim;i++)
	{
		int msec  =bLocal?table_call[i]:table_global_call[i];
		int ncalls=bLocal?table_msec[i]:table_global_msec[i];

		if (msec || ncalls)
			fprintf(file,"[%32s] %6d calls %6d msec\n",table_vec[i].ffi_name,msec,ncalls);
	}

	int msec  =bLocal?(int)total_call:(int)total_global_call;
	int ncalls=bLocal?(int)total_msec:(int)total_global_msec;

	fprintf(file,"[%32s] \n","---------------------");
	fprintf(file,"[%32s] %6d calls %6d msec\n\n","Total",msec,ncalls);
	fflush(file);	
	return scheme_void;
}

Scheme_Object *BenchmarkTime (int argc, Scheme_Object **argv)
{
	return ffi_type::wrap(total_msec);
}

Scheme_Object *BenchmarkGrossTime (int argc, Scheme_Object **argv)
{
	return ffi_type::wrap(total_gross_msec);
}

///////////////////////////////////////////////////////////////////////////////////////////

void PolComplexFinilizer(void *p, void *data) 
{
	clock_t t1=clock();

	Hpc* pc=(Hpc*)OBJSCHEME_GET_CPP_OBJ((Scheme_Object*)p);
	--number_of_registered_hpc;

	if (pc) 
		pc->release();

	int msec=((int)(((clock()-t1)/(double)CLOCKS_PER_SEC)*1000));
	table_msec[FFI_PolComplexFinilizer]+=msec;
	table_global_msec[FFI_PolComplexFinilizer]+=msec;

	table_call[FFI_PolComplexFinilizer]++;
	table_global_call[FFI_PolComplexFinilizer]++;

	total_msec+=msec;
	total_global_msec+=msec;

	total_call++;
	total_global_call++;
}




/////////////////////////////////////////////////////////////////////////

/*
(send (build-pol-complex-cube 2) viewopengl)
*/
Scheme_Object *BuildPolComplexCube (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_BuildPolComplexCube,argc,argv);
	p.checkNumArgs(1,1);

	int dim =p.getInt();
	Hpc *ret = Hpc::cube(dim);

	return ffi_type::wrap(ret);
}


/*
(send (build-pol-complex-simplex 3) viewopengl)
*/

Scheme_Object *BuildPolComplexSimplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_BuildPolComplexSimplex,argc,argv);
	p.checkNumArgs(1,1);

	int dim =p.getInt();
	Hpc *ret = Hpc::simplex(dim);

	return ffi_type::wrap(ret);
}


/////////////////////////////////////////////////////////////////////////

/*
(send (make-object pol-complex% 2) get-npol-dim)
*/
Scheme_Object *GetSpatialDimension (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_GetSpatialDimension,argc,argv);
	p.checkNumArgs(1,1);

	const Hpc* src=p.getHpc();
	int spacedim=src->spacedim;

	return ffi_type::wrap(spacedim);	
}


/*
(send (build-pol-complex-cube 2) get-dpol-dim)
*/
Scheme_Object *GetPointDimension (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_GetPointDimension,argc,argv);
	p.checkNumArgs(1,1);

	const Hpc* src=p.getHpc();
	int pointdim=src->pointdim;

	return ffi_type::wrap(pointdim);	
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
(send (send (build-pol-complex-cube 3) scale-pol-complex 0 0.5) viewopengl)
*/
Scheme_Object *ScalePolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_ScalePolComplex,argc,argv);
	p.checkNumArgs(3,3);

	Hpc*         src   =p.getHpc();
	int          ref   =p.getInt()+1;
	real         factor=p.getReal();

	Hpc* ret=Hpc::scale(max2(ref,src->spacedim),ref,factor);
	ret=ret->add(src);

	return ffi_type::wrap(ret);
}

/*
(send (send (build-pol-complex-cube 3) translate-pol-complex 0 0.5) viewopengl)
*/
Scheme_Object *TranslatePolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_TranslatePolComplex,argc,argv);
	p.checkNumArgs(3,3);

	Hpc*         src   =p.getHpc   ();
	int          ref   =p.getInt   ()+1;
	real         factor=p.getReal  ();

	Hpc* ret=Hpc::translate(max2(ref,src->spacedim),ref,factor);
	ret=ret->add(src);

	return ffi_type::wrap(ret);
}

/*
(send (send (build-pol-complex-cube 3) rotate-pol-complex 0 1 (/ 3.14 2)) viewopengl)
*/
Scheme_Object *RotatePolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_RotatePolComplex,argc,argv);
	p.checkNumArgs(4,4);

	Hpc*         src   =p.getHpc   ();
	int          ref1  =p.getInt   ()+1;
	int          ref2  =p.getInt   ()+1;
	real         factor=-1*p.getReal();

	Hpc* ret=Hpc::rotate(max3(ref1,ref2,src->spacedim),ref1,ref2,factor);
	ret=ret->add(src);

	return ffi_type::wrap(ret);

}

/*
(send (send (build-pol-complex-cube 3) shear-pol-complex 0 2.5) viewopengl)
*/

Scheme_Object *ShearPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_ShearPolComplex,argc,argv);
	Hpc*          src=p.getHpc();
	int           ref=p.getInt()+1;
	int n=0;
	real factor[MAX_GEOMETRIC_DIMENSION+1];
	memset(factor,0,sizeof(factor));
	while (p.remaining()) factor[n++]=p.getReal();

	Hpc* ret=Hpc::shear(max2(ref,src->spacedim),ref,factor);
	ret=ret->add(src);

	return ffi_type::wrap(ret);
}


/*
(send (send (build-pol-complex-cube 2) pol-complex-mat (list (list 1 0 0) (list 0.5 2 10) (list 1.5 10 3))) viewopengl)
*/

Scheme_Object *PolComplexMat (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_PolComplexMat,argc,argv);

	Hpc* src=p.getHpc();
	
	ffi_type prows=p.getList();

	real mat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
	int n=0;
	int nrows=prows.remaining();
	int ncols=-1;

	while (prows.remaining())
	{
		ffi_type pcols=prows.getList();
		if (ncols==-1) ncols=pcols.remaining();

		if (ncols!=nrows || pcols.remaining()!=ncols)
			scheme_signal_error("not square matrix in function PolComplexMat");

		while (pcols.remaining()) mat[n++]=(real)pcols.getReal();
	}

	Hpc* ret=Hpc::transform(nrows-1,mat);
	ret=ret->add(src);

	return ffi_type::wrap(ret);
}


/*
(send (send (build-pol-complex-cube 2) embed-pol-complex 1) dump)
*/

Scheme_Object *EmbedPolComplex(int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_EmbedPolComplex,argc,argv);
	p.checkNumArgs(2,2);

	const Hpc* src=p.getHpc();
	int        dim=p.getInt();

	Hpc* ret=src->clone(); //create a new object which is the embed of source
	ret=ret->embed(src->spacedim+dim);

	return ffi_type::wrap(ret);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
(send (send (build-pol-complex-cube 2) set-name-pol-complex "ciao") dump-pol-complex)
*/

Scheme_Object *SetNamePolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_SetNamePolComplex,argc,argv);
	p.checkNumArgs(2,2);

	Hpc* src=p.getHpc();

#if 0
	char* name=p.getString();
#else
	//TODO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	//memory deallocation!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	char* _name=p.getString();
	char* name=new char[strlen(_name)+1];
	strcpy(name,_name);
#endif

	src=src->addProperty("name",name);
	return ffi_type::wrap(src,argv[0]);
}


/*
(send (send (build-pol-complex-cube 2) add-prop-pol-complex '("chiave" "valore")) dump-pol-complex)
*/

Scheme_Object *AddPropertyPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_AddPropertyPolComplex,argc,argv);
	p.checkNumArgs(2,2);

	Hpc *src = p.getHpc();
	ffi_type subp=p.getList();
	subp.checkNumArgs(2,2);

	char* pname  = subp.getString();
	void* pvalue = subp.getObj();
	scheme_dont_gc_ptr(pvalue);

	src=src->addProperty(pname,pvalue);
	return ffi_type::wrap(src,argv[0]);
}

/*
(send (send (build-pol-complex-cube 2) add-prop-pol-complex '("chiave" "valore")) get-prop-pol-complex "chiave")
*/

Scheme_Object *GetPropertyPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_GetPropertyPolComplex,argc,argv);
	p.checkNumArgs(2,2);

	Hpc   *src  = p.getHpc();
	char  *pname=p.getString();

	void* pvalue=src->getProperty(pname);
	return pvalue?(Scheme_Object*)pvalue:(Scheme_Object*)scheme_void;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/* 
(send (build-pol-complex-cube 2) dump-pol-complex)
*/
Scheme_Object *DumpPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_DumpPolComplex,argc,argv);
	p.checkNumArgs(1,1);

	Hpc* src=p.getHpc();
	src->print();

#if 1
	{
		FILE* file=fopen("dump.tmp","wt"); //for debugging!
		src->print(0,file);
		fclose(file);
	}
#endif

	return scheme_void;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (struct-pol-complex (list (build-pol-complex-cube 2) (build-pol-complex-cube 1))) dump-pol-complex)
*/
Scheme_Object *StructPolComplex(int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_StructPolComplex,argc,argv);
	p.checkNumArgs(1,1);

	ffi_type pstruct=p.getList();

	Hpc* ret=new Hpc;
	while (pstruct.remaining())
	{
		Hpc* src=pstruct.getHpc();
		ret=ret->add(src);
	}
	
	return ffi_type::wrap(ret);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (build-pol-complex-cube 2) get-limits '(1 2) 'med)
(send (build-pol-complex-cube 2) get-limits '(1 2) 'min-max)
*/

Scheme_Object *GetLimits (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_GetLimits,argc,argv);




	//optimization
	Hpc* src=p.getHpc();

	ffi_type psub=p.getList();
	std::vector<int> selection;
	while (psub.remaining()) selection.push_back(psub.getInt());
	char* symbol=p.getString();
	real m[MAX_GEOMETRIC_DIMENSION+1],M[MAX_GEOMETRIC_DIMENSION+1];
	src->limits(m,M);

	if (!strcmp(symbol,"max"))	
	{
		Scheme_Object* vmax[MAX_GEOMETRIC_DIMENSION+1];
		for (int i=0;i<selection.size();i++) vmax[i]=ffi_type::wrap(M[selection[i]]);
		Scheme_Object* aux=scheme_build_list(selection.size(),vmax);
		return scheme_build_list(1,&aux);	
	}
	
	else if (!strcmp(symbol,"min"))
	{
		Scheme_Object* vmin[MAX_GEOMETRIC_DIMENSION+1];
		for (int i=0;i<selection.size();i++) vmin[i]=ffi_type::wrap(m[selection[i]]);
		Scheme_Object* aux=scheme_build_list(selection.size(),vmin);
		return scheme_build_list(1,&aux);	
	}
	else if (!strcmp(symbol,"med"))
	{
		Scheme_Object* vmed[MAX_GEOMETRIC_DIMENSION+1];
		for (int i=0;i<selection.size();i++) vmed[i]=ffi_type::wrap(0.5f*m[selection[i]]+0.5f*M[selection[i]]);
		Scheme_Object* aux=scheme_build_list(selection.size(),vmed);
		return scheme_build_list(1,&aux);	
	}
	else if (!strcmp(symbol,"size"))	
	{
		Scheme_Object* vsiz[MAX_GEOMETRIC_DIMENSION+1];
		for (int i=0;i<selection.size();i++) vsiz[i]=ffi_type::wrap(M[selection[i]]-m[selection[i]]);
		Scheme_Object* aux=scheme_build_list(selection.size(),vsiz);
		return scheme_build_list(1,&aux);	
	}

	else if (!strcmp(symbol,"min-max"))	
	{
		Scheme_Object* vmax[MAX_GEOMETRIC_DIMENSION+1];
		Scheme_Object* vmin[MAX_GEOMETRIC_DIMENSION+1];
		for (int i=0;i<selection.size();i++) {vmin[i]=ffi_type::wrap(m[selection[i]]);vmax[i]=ffi_type::wrap(M[selection[i]]);}
		Scheme_Object* aux[2]={scheme_build_list(selection.size(),vmin),scheme_build_list(selection.size(),vmax)};
		return scheme_build_list(2,aux);
	}
	
	scheme_wrong_type("GetLimits","wrong-symbol",2,argc,argv);
	return scheme_void;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (struct-pol-complex (list (build-pol-complex-cube 2) (build-pol-complex-cube 1))) get-num)
*/
Scheme_Object *GetNum ( int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_GetNum,argc,argv);
	p.checkNumArgs(1,1);

	const Hpc* src=p.getHpc();
	int num=src->num();

	return ffi_type::wrap(num);
}

/*
(send (struct-pol-complex (list (build-pol-complex-cube 2) (build-pol-complex-cube 1))) get-first)
*/

Scheme_Object *GetFirst (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_GetFirst,argc,argv);
	p.checkNumArgs(1,1);

	const Hpc* src=p.getHpc();
	int  first=0;

	return ffi_type::wrap(first);
}

/*
(send (struct-pol-complex (list (build-pol-complex-cube 2) (build-pol-complex-cube 1))) get-succ 0)
*/

Scheme_Object *GetSucc (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_GetSucc,argc,argv);
	p.checkNumArgs(2,2);

	const Hpc* src=p.getHpc();
	int n=p.getInt();
	int succ=n+1;

	return ffi_type::wrap(succ);
}

/*
(send (build-pol-complex-cube 2) Integrate (list 0 1))
*/

Scheme_Object *Integrate (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_Integrate,argc,argv);
	return ffi_type::wrap(0.0f);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(let ((pc1 (build-pol-complex-cube 2)) (pc2 (send (send (build-pol-complex-cube 2) translate-pol-complex 0 2) translate-pol-complex 2 2))) ((begin (send (send pc1 join (list pc2)) viewopengl))))
*/

Scheme_Object *Join(int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_Join,argc,argv);
	Hpc* src=p.getHpc();
	int n=0;
	ffi_type psub=p.getList();
	Hpc** pols=new Hpc*[psub.remaining()+1];
	pols[n++]=src;
	while (psub.remaining()) pols[n++]=psub.getHpc();
	Hpc* ret=Hpc::join(n,pols);
	delete [] pols;
	return ffi_type::wrap(ret);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (build-pol-complex-cube 2) viewopengl)
*/

Scheme_Object *ViewOpenGL(int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(-1,argc,argv);
	p.checkNumArgs(1,1);

	const Hpc* src=p.getHpc();
	//too much transformation can bug OpenGL
	Hpc* rf=src->reducedForm(false,true); //!bStrict,bShareGeometry
	rf->view();
	rf->dropSharedGeometry();
	rf->release(); 

	return scheme_void;
}



/*
(send (build-pol-complex-cube 2) viewopengl)
*/

Scheme_Object *PovRay(int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(-1,argc,argv);
	p.checkNumArgs(1,1);

	const Hpc* src=p.getHpc();
	printf("Exporting to povray....\n");fflush(stdout);
	src->povray(); 
	printf("...done\n");fflush(stdout);

	return scheme_void;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (build-pol-complex-cube 2) copy-pol-complex)
*/

Scheme_Object *CopyPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_CopyPolComplex,argc,argv);
	p.checkNumArgs(1,1);

	const Hpc *src = p.getHpc();
	Hpc *ret = src->clone();

	return ffi_type::wrap(ret);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (send (build-pol-complex-cube 3) skeleton 1) viewopengl)
probabilmente faccio casini con le dimensioni perche' le cell di <g> non sono piu' piene
*/


Scheme_Object *Skeleton (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_Skeleton,argc,argv);
	p.checkNumArgs(2,2);

	Hpc *src=p.getHpc();
	int dim  =p.getInt();

	Hpc* ret=src->skeleton(dim);
	return ffi_type::wrap(ret);
}



Scheme_Object *ExportSVG3D (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_ExportSVG3D,argc,argv);
	p.checkNumArgs(4,4);

	Hpc *src=p.getHpc();

	char* outfilename=p.getString();
	bool  back_to_front=!strcmp(p.getString(),"back-to-front");

	ffi_type psub=p.getList();

	char* type       =psub.getString();

	double vrp[3];vrp[0]=psub.getReal();vrp[1]=psub.getReal();vrp[2]=psub.getReal();
	double vuv[3];vuv[0]=psub.getReal();vuv[1]=psub.getReal();vuv[2]=psub.getReal();
	double vpn[3];vpn[0]=psub.getReal();vpn[1]=psub.getReal();vpn[2]=psub.getReal();
	double cop[3];cop[0]=psub.getReal();cop[1]=psub.getReal();cop[2]=psub.getReal();

	double front=psub.getReal();
	double back =psub.getReal();

	double range_u[2]={psub.getReal(),psub.getReal()};
	double range_v[2]={psub.getReal(),psub.getReal()};
	double range_x[2]={psub.getReal(),psub.getReal()};
	double range_y[2]={psub.getReal(),psub.getReal()};

	int   width =(int)(range_x[1]-range_x[0]);if (width<0 ) width*=-1;
	int   height=(int)(range_y[1]-range_y[0]);if (height<0) height*=-1;

	src->exportSvg(outfilename,type,width,height,back_to_front,vrp,vuv,vpn,cop,front,back,range_u,range_v,range_x,range_y);

	return scheme_void;
}



Scheme_Object *Pipeline3D (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_Pipeline3D,argc,argv);
	p.checkNumArgs(2,2);

	Hpc *src=p.getHpc();

	ffi_type psub=p.getList();

	char* type       =psub.getString();

	double vrp[3];vrp[0]=psub.getReal();vrp[1]=psub.getReal();vrp[2]=psub.getReal();
	double vuv[3];vuv[0]=psub.getReal();vuv[1]=psub.getReal();vuv[2]=psub.getReal();
	double vpn[3];vpn[0]=psub.getReal();vpn[1]=psub.getReal();vpn[2]=psub.getReal();
	double cop[3];cop[0]=psub.getReal();cop[1]=psub.getReal();cop[2]=psub.getReal();

	double front=psub.getReal();
	double back =psub.getReal();

	double range_u[2]={psub.getReal(),psub.getReal()};
	double range_v[2]={psub.getReal(),psub.getReal()};
	double range_x[2]={psub.getReal(),psub.getReal()};
	double range_y[2]={psub.getReal(),psub.getReal()};

	int   width =range_x[1]-range_x[0];if (width<0 ) width*=-1;
	int   height=range_y[1]-range_y[0];if (height<0) height*=-1;

	int spacedim=src->spacedim;

	//valid only in 3 dimension
	if (spacedim!=3)
		return scheme_void;

	real Tv[16]={1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,0,1};
	real Th[16]={1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,0,1};
	real observer[4]={1,0,0,0};
	makePipeline3d(Tv,&observer[1],type,vrp,vuv,vpn,cop,front,back,range_u,range_v,range_x,range_y);

	int permutation[MAX_GEOMETRIC_DIMENSION+1];
	int matdim=spacedim+1;
	permutation[0]=matdim-1;
	for (int I=1;I<matdim;I++) permutation[I]=I-1;
	tmat<real>::swap_rows(matdim,Tv,permutation);
	tmat<real>::swap_cols(matdim,Tv,permutation);
	tmat<real>::invert(spacedim+1,Th,Tv); 

	Hpc* ret=Hpc::transform(spacedim,Tv);
	ret=ret->add(src);

	return ffi_type::wrap(ret);
}




////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(let ((pc1 (build-pol-complex-cube 2))  (pc2 (send (send (build-pol-complex-cube 2) translate-pol-complex 0 0.5) translate-pol-complex 1 0.5)))  ((send (send pc1 or (list pc2)) viewopengl)))
(let ((pc1 (build-pol-complex-cube 2))  (pc2 (send (send (build-pol-complex-cube 2) translate-pol-complex 0 0.5) translate-pol-complex 1 0.5)))  ((send (send pc1 and (list pc2)) viewopengl)))
(let ((pc1 (build-pol-complex-cube 2))  (pc2 (send (send (build-pol-complex-cube 2) translate-pol-complex 0 0.5) translate-pol-complex 1 0.5)))  ((send (send pc1 xor (list pc2)) viewopengl)))
(let ((pc1 (build-pol-complex-cube 2))  (pc2 (send (send (build-pol-complex-cube 2) translate-pol-complex 0 0.5) translate-pol-complex 1 0.5)))  ((send (send pc1 diff (list pc2)) viewopengl)))
*/

static Scheme_Object *DoBoolOp (int argc, Scheme_Object **argv,int operation)
{
	ffi_type_setup();
	ffi_type p(-1,argc,argv);
	p.checkNumArgs(2,2);

	Hpc* src=p.getHpc();
	ffi_type psub=p.getList();
	int n=0;

	Hpc** pols=new Hpc*[psub.remaining()+1];
	pols[n++]=src;
	while (psub.remaining()) pols[n++]=psub.getHpc();

	Hpc* ret=Hpc::booleanOp(operation,n,pols);
	
	delete [] pols;
	return ffi_type::wrap(ret);
}

Scheme_Object *Or(int argc, Scheme_Object **argv)		
{
	ffi_type p(FFI_Or,argc,argv);
	return DoBoolOp(argc,argv,bsp_bool::BOOL_OR);
}
Scheme_Object *And(int argc, Scheme_Object **argv)
{
	ffi_type p(FFI_And,argc,argv);
	return DoBoolOp(argc,argv,bsp_bool::BOOL_AND);
}
Scheme_Object *Diff(int argc, Scheme_Object **argv)		
{
	ffi_type p(FFI_Diff,argc,argv);
	return DoBoolOp(argc,argv,bsp_bool::BOOL_DIFF);
}

Scheme_Object *Xor(int argc, Scheme_Object **argv)		
{
	ffi_type p(FFI_Xor,argc,argv);
	return DoBoolOp(argc,argv,bsp_bool::BOOL_XOR);
}





Scheme_Object *PDiff	(int argc, Scheme_Object **argv)	
{
	ffi_type_setup();
	ffi_type p(FFI_PDiff,argc,argv);
	p.checkNumArgs(2,2);

	Hpc* src=p.getHpc();
	ffi_type psub=p.getList();
	int n=0;
	Hpc** pols=new Hpc*[psub.remaining()+1];
	pols[n++]=src;
	while (psub.remaining()) pols[n++]=psub.getHpc();

	Scheme_Object** retlist=new Scheme_Object*[n];
	Hpc** args=new Hpc*[n];

	int i,j;
	for (i=0;i<n;i++)
	{
		int narg=0;
		args[narg++]=pols[i];
		for (j=0;j<i;j++) args[narg++]=pols[j];
		retlist[i]=ffi_type::wrap(Hpc::booleanOp(bsp_bool::BOOL_DIFF,narg,args));
	}

	Scheme_Object* ret=scheme_build_list(n,retlist);

	delete [] args;
	delete [] pols;
	delete [] retlist;
	return ret;

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
(send (send (build-pol-complex-cube 2) power (struct-pol-complex (list (build-pol-complex-cube 1) (send (build-pol-complex-cube 1) translate-pol-complex 0 2)))) viewopengl)
*/


Scheme_Object *Power(int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_Power,argc,argv);
	p.checkNumArgs(2,2);

	Hpc* arg1=p.getHpc();
	Hpc* arg2=p.getHpc();

	//arg1->print();
	//arg2->print();

	Hpc* ret=Hpc::power(arg1,arg2);
	return ffi_type::wrap(ret);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (build-pol-complex-cube 2) export-vrml-pol-complex "untitled.wrl" 'stream)
*/

Scheme_Object *ExportVRML (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_ExportVRML,argc,argv);
	p.checkNumArgs(3,3);

	const Hpc*  pc=p.getHpc();
	const char* filename=p.getString();
	const char* other=p.getString();

	if (!pc->vrml(filename)) scheme_signal_error("error exporting VRML file");
	return scheme_void;
}


/*
(send (build-pol-complex-cube 2) export-svg-pol-complex 640 "out.svg")
*/

Scheme_Object *ExportSVG (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_ExportSVG,argc,argv);
	p.checkNumArgs(3,3);

	const Hpc* pc=p.getHpc();
	real width=p.getReal();
	const char* filename=p.getString();

	if (!pc->svg(width,filename)) scheme_signal_error("error exporting SVG file");
	return scheme_void;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (make-pol (list (list 0 0) (list 1 0) (list 1 1) (list 0 1)) (list (list 1 2 3 4))  (list (list 1))) viewopengl)
*/

Scheme_Object *MkPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_MkPolComplex,argc,argv);
	p.checkNumArgs(3,3);

	int u;
	ffi_type pverts =p.getList();
	ffi_type pcells =p.getList();
	ffi_type ppols  =p.getList();

	//parse points
	int npoints=pverts.remaining();
	real* Vpoints=0,*Vp=0;
	int pointdim=-1;
	
	for (u=0;u<npoints;u++)
	{
		ffi_type ppoint=pverts.getList();

		if (pointdim==-1) 
		{
			pointdim=ppoint.remaining();
			Vpoints=new real[pointdim*npoints];
			Vp=Vpoints;
		}
		else
		{
			if (pointdim!=ppoint.remaining())
				scheme_signal_error("error in MkPolComplex. vertices don't have the same dimensions %V",argv[0]);
		}

		for (int i=0;i<pointdim;i++) *Vp++=ppoint.getReal();	
	}
	checkpoint((Vp-Vpoints)==(npoints*pointdim));

	//parse hulls
	int nhulls=pcells.remaining();
	int** Vhulls=new int*[nhulls];
	int** Vh=Vhulls;

	int midx=-1,Midx=-1;
	while (pcells.remaining())
	{
		ffi_type phull=pcells.getList();
		*Vh=new int[phull.remaining()+1];
		int* h=*Vh++;
		*h++=phull.remaining();

		while (phull.remaining()) 
		{
			int idx=phull.getInt();
			midx=(midx==-1)?(idx):(min2(midx,idx));
			Midx=(Midx==-1)?(idx):(max2(Midx,idx));
			*h++=idx;
		}
	}
	checkpoint((Vh-Vhulls)==nhulls);

	if (midx==1) //bug fix
	{
		for (u=0;u<nhulls;u++) 
		for (int j=1;j<=Vhulls[u][0];j++) --Vhulls[u][j];
		midx--;Midx--;
	}

	//safety check
	if (midx<0 || midx>=npoints || Midx<0 || Midx>=npoints)
		scheme_signal_error("error in MkPolComplex. hulls reference a point outside the valid range %V",argv[1]);

	//parse pols
	int npols=ppols.remaining();
	int** Vpols=new int*[npols];
	int** Vo=Vpols;
	midx=-1,Midx=-1;
	while (ppols.remaining())
	{
		ffi_type ppol=ppols.getList();
		*Vo=new int[ppol.remaining()+1];
		int* p=*Vo++;
		*p++=ppol.remaining();

		while (ppol.remaining()) 
		{
			int idx=ppol.getInt();
			midx=(midx==-1)?(idx):(min2(midx,idx));
			Midx=(Midx==-1)?(idx):(max2(Midx,idx));
			*p++=idx;
		}
	}
	checkpoint((Vo-Vpols)==npols);

	if (midx==1) //bug fix
	{
		for (u=0;u<npols;u++) 
		for (int j=1;j<=Vpols[u][0];j++) --Vpols[u][j];
		--midx;
		--Midx;
	}

	if (midx<0 || midx>=nhulls || Midx<0 || Midx>=nhulls)
		scheme_signal_error("error in MkPolComplex. Pols reference an hull outside the valid range %V",argv[2]);


	//simplify hulls not referenced
	bool* bHull=new bool[nhulls];
	memset(bHull,0,sizeof(bool)*nhulls);
	for (u=0;u<npols;u++) 
	{for (int j=1;j<=Vpols[u][0];j++) bHull[Vpols[u][j]]=true;}
	for (u=0;u<nhulls;u++) if (!bHull[u]) Vhulls[u][0]=0;

	Hpc *pc = Hpc::mkpol(pointdim,npoints,(const real*)Vpoints,nhulls,(const int**)Vhulls);
	
	delete [] bHull;
	delete [] Vpoints;
	for (u=0;u<nhulls;u++) delete [] Vhulls[u];delete [] Vhulls;
	for (u=0;u<npols ;u++) delete [] Vpols [u];delete [] Vpols;

	return ffi_type::wrap(pc);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (build-pol-complex-cube 2) unmake-pol)
(send (build-pol-complex-cube 2) unmake-pol-f)
*/

static Scheme_Object *UkPolComplex (int argc, Scheme_Object **argv,bool bCovector)
{
	int i;
	ffi_type p(-1,argc,argv);
	p.checkNumArgs(1,1);

	Hpc* src=p.getHpc();

	real *nodes;
	int dim,nnodes,nhulls,**hulls;
	src->ukpol(dim,nnodes,nodes,nhulls,hulls,bCovector);

	Scheme_Object **In[3],*Out[3];
	In[0]=new Scheme_Object*[nnodes];
	In[1]=new Scheme_Object*[nhulls];
	In[2]=new Scheme_Object*[nhulls];

	//vertices
	Scheme_Object* v[MAX_GEOMETRIC_DIMENSION+1];
	real* ptr=nodes;
	
	for (i=0;i<nnodes;i++)
	{
		Scheme_Object** pv=v;
		if (bCovector) *pv++=ffi_type::wrap(-1*(*ptr++)); 
		for (int u=0;u<dim;u++) *pv++=ffi_type::wrap((bCovector?-1:1)*(*ptr++));
		if (bCovector) Swap(Scheme_Object*,v[0],v[dim]); //expend homog coordinate as last
		In[0][i]=scheme_build_list(dim+(bCovector?1:0),v);
	}
	checkpoint((ptr-nodes)==nnodes*(dim+(bCovector?1:0)));
	
	//hulls/pols
	for (i=0;i<nhulls;i++)
	{
		int* h=hulls[i];
		int num=*h++;
		Scheme_Object** aux=new Scheme_Object*[num];
		for (int j=1;j<=num;j++,h++) aux[j-1]=ffi_type::wrap(*h+1);
		In[1][i]=scheme_build_list(num,aux);
		Scheme_Object* tmp=ffi_type::wrap(i+1);In[2][i]=scheme_build_list(1,&tmp);
		delete [] aux;
	}
		
	for (int u=0;u<nhulls;u++) delete [] hulls[u];
	delete [] hulls;
	delete [] nodes;

	Out[0] = scheme_build_list(nnodes,In[0]);delete [] In[0] ;
	Out[1] = scheme_build_list(nhulls,In[1]);delete [] In[1];
	Out[2] = scheme_build_list(nhulls,In[2]);delete [] In[2];

	Scheme_Object *ret=scheme_build_list(3,Out);
	
	return ret;
}


Scheme_Object *UkPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_UkPolComplex,argc,argv);
	return UkPolComplex(argc,argv,false);
}

Scheme_Object *UkPolComplexFaces (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_UkPolComplexf,argc,argv);
	return UkPolComplex(argc,argv,true);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/*
(send (send (build-pol-complex-cube 2) map-pol-complex (lambda(x) (list (car x) (+ 1 (cadr x)) ))) viewopengl)
*/

Scheme_Object *MapPolComplex (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_MapPolComplex,argc,argv);
	p.checkNumArgs(2,2);

	int U;
	Hpc* src=p.getHpc()->reducedForm(false,false);//!bStrict,!bShareGeometry

	Scheme_Object* scheme_coords[MAX_GEOMETRIC_DIMENSION+1];
	Scheme_Object* schemefun=p.getFunction();

	Hpc *ret = new Hpc(src->spacedim);

	int time_in_scheme=0;

	int npoints=0;
	for (int i=0;i<src->num();i++)
	{
		Hpc* sub=src->get(i)->clone();
		ggraph* g=sub->g;

		g->triangulate(); //triangulate in its full dimension
		g->embed(src->spacedim); //go to spacedim of the complex

		if (sub->vmat) 
		{
			g->transform(sub->vmat,sub->hmat); 
			delete [] sub->vmat;sub->vmat=0;
			delete [] sub->hmat;sub->hmat=0;
		}

		//map points
		for (ggraph::list_iterator  it=g->each(0);!it.end();it++,++npoints)
		{
			real* vg=g->getGeom(*it);
			for (U=0;U<ret->spacedim;U++) scheme_coords[U]=ffi_type::wrap(vg[U+1]);
			Scheme_Object* scheme_point=scheme_build_list(ret->spacedim,scheme_coords);
			clock_t t1=clock();
			Scheme_Object* tmp=_scheme_apply(schemefun,1,&scheme_point);
			time_in_scheme+=(int)(((clock()-t1)/(double)CLOCKS_PER_SEC)*1000);
			int newdim=scheme_proper_list_length(tmp);
			if (newdim > g->pointdim) {g->embed(newdim);vg=g->getGeom(*it);sub->pointdim=sub->spacedim=newdim;}
			for (U=0;U<newdim;U++,tmp=scheme_cdr(tmp)) vg[U+1]=ffi_type::getReal(scheme_car(tmp));
		}

		//if the output g is not in full dimension the points could have been relocated everywhere. I need
		//a new embedding in the space

		if (g->each(g->pointdim).end())
		{
			int pointdim,npoints,nhulls,**hulls;real* points;
			sub->ukpol(pointdim,npoints,points,nhulls,hulls,false);
			Hpc* aux=sub->mkpol(pointdim,npoints,points,nhulls,(const int**)hulls);
			ret=ret->add(aux);
			aux->release();
			sub->release();
			delete [] points;
			for (int u=0;u<nhulls;u++) delete [] hulls[u];delete [] hulls;
		}
		else
		{
			//reset plane information, they aren't valid anymore
			for (ggraph::list_iterator  it=g->each(g->pointdim-1);!it.end();it++)
				g->releaseGeom(*it);
			ret=ret->add(sub);
			sub->release();
		}

	}

	src->release();

	p.outside_time=time_in_scheme;
	return ffi_type::wrap(ret);
}


Scheme_Object *MapPolComplex2 (int argc, Scheme_Object **argv)
{
	int i,j;
	ffi_type_setup();
	ffi_type p(FFI_MapPolComplex,argc,argv);
	p.checkNumArgs(2,2);

	/* Get Arguments */
	Hpc* src=p.getHpc()->reducedForm(false,false);//!bStrict,!bShareGeometry
	Scheme_Object* schemefun=p.getFunction();
	
	/* Ukpol */
	int pointdim,npoints,nhulls,**hulls;
	real* src_points; 
	src->ukpol(pointdim,npoints,src_points,nhulls,hulls,false);
	
	/* Make initial c-array of scheme vectors of coordinates*/
	Scheme_Object* src_scheme_coords[MAX_GEOMETRIC_DIMENSION+1];
	for( j=0;j<pointdim;j++) {
		src_scheme_coords[j]=scheme_make_vector(npoints, scheme_make_double(0.0));
	}
	
	
	//	for (int u=0;u<dim;u++) *pv++=ffi_type::wrap((bCovector?-1:1)*(*ptr++));
	/* fill the  c-array of scheme vectors of coordinates*/
	for( i=0;i<npoints;i++) {
		for( j=0;j<pointdim;j++) {
			//SCHEME_VEC_ELS(src_scheme_coords[j])[i]=ffi_type::wrap(*(src_points+(i*(pointdim+1))+j+1)); 
			SCHEME_VEC_ELS(src_scheme_coords[j])[i]=ffi_type::wrap(*(src_points++)); 
		}
	}
	
	/* Make list out of c-array */
	Scheme_Object* scheme_src=scheme_build_list(src->spacedim,src_scheme_coords);
	
	/* Call back scheme */
	
	clock_t t1=clock();
	Scheme_Object* scheme_ret=_scheme_apply(schemefun,1,&scheme_src);
	int time_in_scheme=(int)(((clock()-t1)/(double)CLOCKS_PER_SEC)*1000);
	
	/* Rebuild c-array of scheme vectors of coordinates */
	int newdim=scheme_proper_list_length(scheme_ret);
	Scheme_Object* tmp=scheme_ret;
	Scheme_Object* ret_scheme_coords[MAX_GEOMETRIC_DIMENSION+1];
	for ( j=0;j<newdim;j++,tmp=scheme_cdr(tmp)) ret_scheme_coords[j]=scheme_car(tmp);
	
	/* Rebuild points */
	real *ret_points=0;
	ret_points=new real[npoints*newdim];
	for( i=0;i<npoints;i++) {
		for( j=0;j<newdim;j++) {
			tmp=ret_scheme_coords[j];
			if (SCHEME_VECTORP(tmp)) {
				*(ret_points+(i*newdim)+j)=ffi_type::getReal(SCHEME_VEC_ELS(tmp)[i]); 
			} else {
				*(ret_points+(i*newdim)+j)=ffi_type::getReal(tmp); 
			}
				
			//*(ret_points++)=ffi_type::getReal(SCHEME_VEC_ELS(ret_scheme_coords[j])[i]);
		}
	}

	
	/* Make pol complex again */
	Hpc* ret=ret->mkpol(newdim,npoints,ret_points,nhulls,(const int**)hulls);
	delete [] ret_points;
	
	p.outside_time=time_in_scheme;
	return ffi_type::wrap(ret);
}

///////////////////////////////////////////////////////////////////////////////////////
// TODO
///////////////////////////////////////////////////////////////////////////////////////


Scheme_Object *ExportXML (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_ExportXML,argc,argv);
	scheme_signal_error("XML export not yet implemented!");
	return scheme_void;
}

Scheme_Object *XMLImport (int argc, Scheme_Object **argv)
{
	ffi_type_setup();
	ffi_type p(FFI_XMLImport,argc,argv);
	scheme_signal_error("XML import not yet implemented!");
	return scheme_void;
}


Scheme_Object* HpcConfig(int argc,Scheme_Object** argv)
{
	ffi_type_setup();
	ffi_type p(FFI_HpcConfig,argc,argv);
	p.checkNumArgs(1,1);

	ffi_type subp=p.getList();

	while (subp.remaining())
	{
		char* name =subp.getString();
		real  value=subp.getReal();

		if (!strcmp(name,"HPC_USE_SYMBOLIC_MKPOL"))
		{
			HPC_USE_SYMBOLIC_MKPOL=value>0;
		}
		else if (!strcmp(name,"HPC_TOLERANCE"))
		{
			HPC_TOLERANCE=value;
		}
		else if (!strcmp(name,"HPC_TOLERANCE_BOOL"))
		{
			HPC_TOLERANCE_BOOL=value;
		}
		else if (!strcmp(name,"HPC_USE_INTERNAL_FACES"))
		{
			HPC_USE_INTERNAL_FACES=value>0;
		}
		else if (!strcmp(name,"HPC_USE_PARTITION_PLANES"))
		{
			HPC_USE_PARTITION_PLANES=value>0;
		}
	}

	//print actual configuration
	printf("bool HPC_USE_SYMBOLIC_MKPOL  =%s\n", HPC_USE_SYMBOLIC_MKPOL?"true":"false");
	printf("real HPC_TOLERANCE           =%e\n", HPC_TOLERANCE);
	printf("real HPC_TOLERANCE_BOOL      =%e\n", HPC_TOLERANCE_BOOL);
	printf("bool HPC_USE_INTERNAL_FACES  =%s\n", HPC_USE_INTERNAL_FACES?"true":"false");
	printf("bool HPC_USE_PARTITION_PLANES=%s\n", HPC_USE_PARTITION_PLANES?"true":"false");

	return scheme_void;
}





///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////

Scheme_Object *scheme_SimpleXn_initialize (Scheme_Env *env)
{
	table_dim=0;

	#define make_ffi_fn(_code,_class_name,_ffi_name,_fn,_minarg,_maxarg) \
	{\
		_code=table_dim++;\
		mzwrap_fn_t _tmp_={_code,_class_name,_ffi_name,_fn,_minarg,_maxarg};\
		table_vec[_code]=_tmp_;\
	}\

	make_ffi_fn(FFI_XMLImport							,0					,"xml-pol-import"						,XMLImport,1,1);
	make_ffi_fn(FFI_HpcConfig							,0					,"hpc-config"						   ,HpcConfig,1,1);
	make_ffi_fn(FFI_BuildPolComplexCube				,0					,"build-pol-complex-cube"			,BuildPolComplexCube,1,1);
	make_ffi_fn(FFI_BuildPolComplexSimplex			,0					,"build-pol-complex-simplex"		,BuildPolComplexSimplex,1,1);
	make_ffi_fn(FFI_StructPolComplex					,0					,"struct-pol-complex"				,StructPolComplex,1,1);
	make_ffi_fn(FFI_MkPolComplex						,0					,"make-pol"								,MkPolComplex,3,3);
	make_ffi_fn(FFI_BenchmarkCount					,0					,"hpc-count"							,BenchmarkCount,0,0);
	make_ffi_fn(FFI_BenchmarkReset					,0					,"hpc-benchmark-reset"				,BenchmarkReset,0,0);
	make_ffi_fn(FFI_BenchmarkLog						,0					,"hpc-benchmark-log"					,BenchmarkLog,1,1);
	make_ffi_fn(FFI_BenchmarkTime						,0					,"hpc-benchmark-time"				,BenchmarkTime,0,0);
	make_ffi_fn(FFI_BenchmarkTime						,0					,"hpc-benchmark-gross-time"		,BenchmarkGrossTime,0,0);
	make_ffi_fn(FFI_GetSpatialDimension				,"pol-complex%","get-npol-dim"						,GetSpatialDimension,0,0);
	make_ffi_fn(FFI_GetPointDimension				,"pol-complex%","get-dpol-dim"						,GetPointDimension,0,0);
	make_ffi_fn(FFI_Power								,"pol-complex%","power"									,Power,1,1);
	make_ffi_fn(FFI_GetFirst							,"pol-complex%","get-first"							,GetFirst,0,0);
	make_ffi_fn(FFI_GetSucc								,"pol-complex%","get-succ"								,GetSucc,1,1);
	make_ffi_fn(FFI_GetNum								,"pol-complex%","get-num"								,GetNum,0,0);
	make_ffi_fn(FFI_ScalePolComplex					,"pol-complex%","scale-pol-complex"					,ScalePolComplex,2,2);
	make_ffi_fn(FFI_TranslatePolComplex				,"pol-complex%","translate-pol-complex"			,TranslatePolComplex,2,2);
	make_ffi_fn(FFI_RotatePolComplex					,"pol-complex%","rotate-pol-complex"				,RotatePolComplex,3,3);
	make_ffi_fn(FFI_ShearPolComplex					,"pol-complex%","shear-pol-complex"					,ShearPolComplex,2,2);
	make_ffi_fn(FFI_EmbedPolComplex					,"pol-complex%","embed-pol-complex"					,EmbedPolComplex,1,1);
	make_ffi_fn(FFI_Skeleton							,"pol-complex%","skeleton"								,Skeleton,1,1);
	make_ffi_fn(FFI_Or									,"pol-complex%","or"										,Or,1,1);
	make_ffi_fn(FFI_And									,"pol-complex%","and"									,And,1,1);
	make_ffi_fn(FFI_Diff									,"pol-complex%","diff"									,Diff,1,1);
	make_ffi_fn(FFI_Xor									,"pol-complex%","xor"									,Xor,1,1);
	make_ffi_fn(FFI_PDiff								,"pol-complex%","pdiff"									,PDiff,1,1);
	make_ffi_fn(FFI_Integrate							,"pol-complex%","integrate"							,Integrate,1,1);
	make_ffi_fn(FFI_Join									,"pol-complex%","join"									,Join,1,1);
	make_ffi_fn(FFI_GetLimits							,"pol-complex%","get-limits"							,GetLimits,2,2);
	make_ffi_fn(FFI_UkPolComplex						,"pol-complex%","unmake-pol"							,UkPolComplex,0,0);
	make_ffi_fn(FFI_UkPolComplexf						,"pol-complex%","unmake-pol-f"						,UkPolComplexFaces,0,0);
	make_ffi_fn(FFI_MapPolComplex						,"pol-complex%","map-pol-complex"					,MapPolComplex,1,1);
	make_ffi_fn(FFI_MapPolComplex2					,"pol-complex%","map-pol-complex2"					,MapPolComplex2,1,1);
	make_ffi_fn(FFI_ExportVRML							,"pol-complex%","export-vrml-pol-complex"			,ExportVRML,2,2);
	make_ffi_fn(FFI_ExportSVG							,"pol-complex%","export-svg-pol-complex"			,ExportSVG,2,2);
	make_ffi_fn(FFI_ExportSVG3D                  ,"pol-complex%","export-svg3d-pol-complex"      ,ExportSVG3D,3,3);
	make_ffi_fn(FFI_Pipeline3D                   ,"pol-complex%","pipeline3d-pol-complex"        ,Pipeline3D,1,1);
	make_ffi_fn(FFI_DumpPolComplex					,"pol-complex%","dump-pol-complex"					,DumpPolComplex,0,0);
	make_ffi_fn(FFI_PolComplexMat						,"pol-complex%","pol-complex-mat"					,PolComplexMat,1,1) ;
	make_ffi_fn(FFI_SetNamePolComplex				,"pol-complex%","set-name-pol-complex"				,SetNamePolComplex,1,1) ; 
	make_ffi_fn(FFI_CopyPolComplex					,"pol-complex%","copy-pol-complex"					,CopyPolComplex,0,0);
	make_ffi_fn(FFI_AddPropertyPolComplex			,"pol-complex%","add-prop-pol-complex"				,AddPropertyPolComplex,1,1) ; 
	make_ffi_fn(FFI_GetPropertyPolComplex			,"pol-complex%","get-prop-pol-complex"				,GetPropertyPolComplex,1,1);
	make_ffi_fn(FFI_ExportXML							,"pol-complex%","save-xml"								,ExportXML,1,1);
	make_ffi_fn(FFI_ViewOpenGL							,"pol-complex%","viewopengl"							,ViewOpenGL,0,0);
	make_ffi_fn(FFI_PovRay                       ,"pol-complex%","povray"                        ,PovRay,0,0);

	scheme_register_extension_global(&polComplex,sizeof(polComplex));

	int num_methods_pol_nodes=0,num_methos_pol_complex=0;

	int i;
	for (i=0;i<table_dim;i++)
	{
		char* class_name=table_vec[i].class_name;

		if (!class_name) continue;
		if (!strcmp(class_name,"pol-complex%")) ++num_methos_pol_complex;
	}

	polComplex = objscheme_make_class("pol-complex%" ,NULL,NULL,num_methos_pol_complex);
	
	for (i=0;i<table_dim;i++)
	{
		char* class_name=table_vec[i].class_name;
		char* ffi_name=table_vec[i].ffi_name;
		int   minarg=table_vec[i].minarg;
		int   maxarg=table_vec[i].maxarg;
		Scheme_Object* (*fn)(int argc, Scheme_Object **argv)=table_vec[i].fn;

		if (!class_name)
			scheme_add_global(ffi_name,scheme_make_prim_w_arity(fn,ffi_name,minarg,maxarg),env);
		else if (!strcmp(class_name,"pol-complex%")) 
			objscheme_add_method_w_arity(polComplex,ffi_name,fn,minarg,maxarg);
	}

	scheme_add_global("pol-complex-primitive-class",polComplex,env);

	//finilization
	{
		FFI_PolComplexFinilizer=table_dim++;
		mzwrap_fn_t _tmp_={FFI_PolComplexFinilizer,0,"finilize-pol-complex",0,1,1};
		table_vec[FFI_PolComplexFinilizer]=_tmp_;
	}



	return scheme_void;
}

