/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */

#include "generaldef.h"


#include <locale.h>
#include <scheme.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "hpc.h"
#include "gbool.h"

extern "C" {Scheme_Object *scheme_Plasm_initialize(Scheme_Env *env);}
Scheme_Object *scheme_SimpleXn_initialize(Scheme_Env *env);


///////////////////////////////////////////////////////////////////////////////

void test_hpc(int cont=1,bool bView=true)
{

REDO:

	{
		//simplex
		printf("testing simplex...\n");
		Hpc* obj;
		obj=Hpc::simplex(0);if (bView) obj->view();obj->release();
		obj=Hpc::simplex(1);if (bView) obj->view();obj->release();
		obj=Hpc::simplex(2);if (bView) obj->view();obj->release();
		obj=Hpc::simplex(3);if (bView) obj->view();obj->release();

	}
	
	{
		//cube
		printf("testing cube...\n");
		Hpc* obj;
		obj=Hpc::cube(0);if (bView) obj->view();obj->release();
		obj=Hpc::cube(1);if (bView) obj->view();obj->release();
		obj=Hpc::cube(2);if (bView) obj->view();obj->release();
		obj=Hpc::cube(3);if (bView) obj->view();obj->release();
	}

	{
		//mkpol
		printf("testing mkpol...\n");
		const int nhulls=1;
		real points[]={0,0, 1,0, 1,1, 0,1,};
		const int  hulls0[]={4,  0,1,2,3};
		const int* hulls []={hulls0};
		Hpc* obj=Hpc::mkpol(2,4,points,nhulls,hulls);
		if (bView) obj->view();obj->release();
		
	}

	{
		//mkpol
		const int nhulls=26;
		const int npoints=8;
		const int pointdim=3;
		real points[npoints*pointdim]={0,0,0, 1,0,0, 1,1,0, 0,1,0,0,0,1, 1,0,1, 1,1,1, 0,1,1};
		const int hullsf0[]={4, 0,1,2,3};
		const int hullsf1[]={4, 0,1,5,4};
		const int hullsf2[]={4, 1,2,6,5};
		const int hullsf3[]={4, 2,3,7,6};
		const int hullsf4[]={4, 0,3,7,4};
		const int hullsf5[]={4, 4,5,6,7};
		const int hullse0[]={2, 4,5};int hullse1[]={2, 5,6};int hullse2 []={2, 6,7};int hullse3 []={2, 7,4};
		const int hullse4[]={2, 0,1};int hullse5[]={2, 1,2};int hullse6 []={2, 2,3};int hullse7 []={2, 3,0};
		const int hullse8[]={2, 0,4};int hullse9[]={2, 1,5};int hullse10[]={2, 2,6};int hullse11[]={2, 3,7};
		const int hullsp0[]={1,0};int hullsp1[]={1,1};int hullsp2[]={1,2};int hullsp3[]={1,3};
		const int hullsp4[]={1,4};int hullsp5[]={1,5};int hullsp6[]={1,6};int hullsp7[]={1,7};

		const int* hulls[26]=
		{
			hullsf0,hullsf1,hullsf2,hullsf3,hullsf4,hullsf5,
			hullse0,hullse1,hullse2,hullse3,hullse4,hullse5,hullse6,hullse7,hullse8,hullse9,hullse10,hullse11,
			hullsp0,hullsp1,hullsp2,hullsp3,hullsp4,hullsp5,hullsp6,hullsp7
		};

		Hpc* obj=Hpc::mkpol(pointdim,npoints,points,nhulls,hulls);
		if (bView) obj->view();
		obj->release();
	}

	{
		//mkpol
		const int pointdim=3;
		const int npoints=8;
		const int nhulls=1;
		const real points[]={0,0,0, 1,0,0, 1,1,0, 0,1,0, 0,0,1, 1,0,1, 1,1,1, 0,1,1};
		const int hulls0[]={8, 0,1,2,3,4,5,6,7};
		const int* hulls[nhulls]={hulls0};
		Hpc* obj=Hpc::mkpol(pointdim,npoints,points,nhulls,hulls);
		if (bView) obj->view();obj->release();
	}

	{
		//embed
		printf("testing embed...\n");
		Hpc* cube2=Hpc::cube(2);
		cube2=cube2->embed(3);

		Hpc* obj=Hpc::translate(3,3,0.5)->add(cube2);
		if (bView) obj->view();
		
		cube2->release();
		obj->release();
	}


	{
		//transformations
		printf("testing transformation...\n");
		Hpc* obj=Hpc::cube(3);

		real vt3[16],ht3[16],vtt[16],htt[16];
		tmat<real>::identity(4,vtt);
		tmat<real>::identity(4,htt);
		real vtrans3[]={0.1f,0.2f,0.3f};
		real vscale3[]={1,2,3};

		tmat<real>::scale    (4,1,vscale3[0],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 
		tmat<real>::scale    (4,2,vscale3[1],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 
		tmat<real>::scale    (4,3,vscale3[2],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 
		tmat<real>::translate(4,1,vtrans3[0],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3);
		tmat<real>::translate(4,2,vtrans3[1],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3);   
		tmat<real>::translate(4,3,vtrans3[2],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 

		Hpc* tmp[1024];int ntmp=0;

		tmp[0]=Hpc::transform(3,htt)->add(
				tmp[1]=Hpc::translate(3,3,vtrans3[2])->add(
					tmp[2]=Hpc::translate(3,2,vtrans3[1])->add(
						tmp[3]=Hpc::translate(3,1,vtrans3[0])->add(
							tmp[4]=Hpc::scale(3,3,vscale3[2])->add(
								tmp[5]=Hpc::scale(3,2,vscale3[1])->add(
									tmp[6]=Hpc::scale(3,1,vscale3[0])->add(
										obj )))))));
		ntmp=7;
		if (bView) tmp[ntmp-1]->view();
		obj->release();


		
		for (int i=0;i<ntmp;i++)
			tmp[i]->release();
	}

	{
		//limits
		Hpc* tmp[1024];int ntmp=0;
		printf("testing limits...\n");

		Hpc* cube0=
			tmp[0]=Hpc::rotate(3,1,2,3.14f/6)->add(
				tmp[1]=Hpc::cube(3));

		real m[MAX_GEOMETRIC_DIMENSION+1],M[MAX_GEOMETRIC_DIMENSION+1];
		cube0->limits(m,M);

		Hpc* cube1=
			tmp[2]=Hpc::translate(3,3,M[3])->add(
				tmp[3]=Hpc::translate(3,2,M[2])->add(
					tmp[4]=Hpc::translate(3,1,M[1])->add(
						tmp[5]=Hpc::cube(3))));

		Hpc* cube2=
			tmp[6]=Hpc::translate(3,3,+m[3])->add(
				tmp[7]=Hpc::translate(3,2,+m[2])->add(
					tmp[8]=Hpc::translate(3,1,+m[1])->add(
						tmp[9]=Hpc::scale(3,3,-1)->add(
							tmp[10]=Hpc::scale(3,1,-1)->add(
								tmp[11]=Hpc::scale(3,2,-1)->add(
									tmp[12]=Hpc::cube(3)))))));

		Hpc* obj=tmp[13]=new Hpc;
		obj->add(cube0);
		obj->add(cube1);
		obj->add(cube2);

		if (bView) obj->view();

		ntmp=14;
		for (int i=0;i<ntmp;i++)
			tmp[i]->release();
	}

	{
		//join

		Hpc* tmp[1024];int ntmp=0;

		printf("testing join...\n");
		Hpc* cube0=
			tmp[0]=Hpc::translate(3,3,1)->add(
				tmp[1]=Hpc::cube(0));

		Hpc* cube1=
			tmp[2]=Hpc::translate(3,3,1.7f)->add(
				tmp[3]=Hpc::translate(3,1,0)->add(
					tmp[4]=Hpc::translate(3,2,0.5f)->add(
						tmp[5]=Hpc::cube(1))));

		Hpc* cube2=
				tmp[6]=Hpc::cube(2);

		Hpc* args[]={cube0,cube1,cube2};
		Hpc* obj=
			tmp[7]=Hpc::join(3,args);

		if (bView) obj->view();
		ntmp=8;
		for (int i=0;i<ntmp;i++)
			tmp[i]->release();
	}

	{	
		Hpc* tmp[1024];int ntmp=0;

		//boolean operation
		printf("testing boolean ops...\n");
		Hpc* cube0=
			tmp[0]=Hpc::cube(3);

		Hpc* cube1=
			tmp[1]=Hpc::translate(3,1,0.5)->add(
				tmp[2]=Hpc::translate(3,2,0.5)->add(
					tmp[3]=Hpc::translate(3,3,0.5)->add(
						tmp[4]=Hpc::cube(3))));

		int ops[]={bsp_bool::BOOL_OR,bsp_bool::BOOL_AND,bsp_bool::BOOL_DIFF,bsp_bool::BOOL_XOR};

		  int i;
		for (i=0;i<4;i++)
		{
			Hpc* args[2];
			args[0]=cube0->clone();
			args[1]=cube1->clone();
			Hpc* obj=Hpc::booleanOp(ops[i],2,args);
			if (bView) obj->view();
			obj->release();
			args[0]->release();
			args[1]->release();
		}
		
		ntmp=5;
		for (i=0;i<ntmp;i++)
			tmp[i]->release();
	}

	{
		//ukpol
		Hpc* tmp[1024];int ntmp=0;

		printf("testing ukpol...\n");

		Hpc* obj1=
			(tmp[0]=(new Hpc()))->add(
				(tmp[1]=Hpc::cube(0))->add(
				(tmp[2]=Hpc::translate(3,2,1))->add(tmp[3]=Hpc::cube(1))->add(
				(tmp[4]=Hpc::translate(3,2,2))->add(tmp[5]=Hpc::cube(2))->add(
				(tmp[6]=Hpc::translate(3,2,3))->add(tmp[7]=Hpc::cube(3))))));

		real* points;
		int pointdim,npoints,**hulls,nhulls;
		Hpc* clone=tmp[8]=obj1->clone();
	
		clone->ukpol(pointdim,npoints,points,nhulls,hulls,false);

		Hpc* obj2=
			(tmp[9]=Hpc::translate(3,1,1.5))->add(
				tmp[10]=Hpc::mkpol(pointdim,npoints,points,nhulls,(const int**)hulls));
		  
int i;
		for ( i=0;i<nhulls;i++) delete [] hulls[i];
		delete [] hulls;
		delete [] points;

		Hpc* obj=(tmp[11]=(new Hpc))->add(obj1)->add(obj2);

		if (bView) obj->view();
	
		ntmp=12;
		for ( i=0;i<ntmp;i++)
			tmp[i]->release();
	}

	{
		//ukpolf
		Hpc* tmp[1024];int ntmp=0;

		printf("testing ukpolf...\n");
		Hpc* obj1=tmp[0]=Hpc::cube(3);

		real* faces;
		int pointdim,nfaces,**hulls,nhulls;
		Hpc* clone=tmp[1]=obj1->clone();
		clone->ukpol(pointdim,nfaces,faces,nhulls,hulls,true);

		int i,j,dst;
		real* src=faces;
		for (i=0,dst=0;i<nfaces   ;i++)
		for (j=0      ;j<=pointdim;j++,src++)
			if (j) faces[dst++]=*src;

		Hpc* obj2=tmp[2]=Hpc::mkpol(pointdim,nfaces,faces,nhulls,(const int**)hulls);

		for (i=0;i<nhulls;i++) delete [] hulls[i];
		delete [] hulls;
		delete [] faces;

		Hpc* obj=(tmp[3]=(new Hpc))->add(
			(tmp[4]=Hpc::translate(3,1,2))->add(obj1))->add(obj2);
		if (bView) obj->view();

		ntmp=5;
		for (i=0;i<ntmp;i++)
			tmp[i]->release();

	}

	{
		//power
		Hpc* tmp[1024];int ntmp=0;

		printf("testing power...\n");
		Hpc* cube2  =tmp[0]=Hpc::cube(2);
		Hpc* cube2s1=tmp[1]=cube2->skeleton(1);
		Hpc* cube2s0=tmp[2]=cube2->skeleton(0);
		
		Hpc* cube1=tmp[3]=Hpc::cube(1);

		Hpc* obj=
			(tmp[4]=(new Hpc))->add(
				(tmp[5]=Hpc::power(cube2s0,cube1)))->add(
				(tmp[6]=Hpc::power(cube2s1,cube1)))->add(
				(tmp[7]=Hpc::translate(3,1,2))->add(
					(tmp[8]=(new Hpc))->add(
						(tmp[9]=Hpc::power(cube1,cube2s0)))->add(
						(tmp[10]=Hpc::power(cube1,cube2s1)))))
			;

		if (bView) obj->view();

		ntmp=11;
		for (int i=0;i<ntmp;i++)
			tmp[i]->release();
	}

	{

		Hpc* tmp[1024];int ntmp=0;

		//skeleton
		printf("testing skeleton...\n");
		Hpc* cube0=
			tmp[0]=Hpc::cube(3);

		Hpc* obj=
			(tmp[1]=(new Hpc))->add(
				(tmp[2]=cube0->skeleton(0)))->add(
					(tmp[3]=cube0->skeleton(1)))->add(
						(tmp[4]=cube0->skeleton(2)));

		if (bView) obj->view();

		ntmp=5;
		for (int i=0;i<ntmp;i++)
			tmp[i]->release();
	}

	cont--;
	if (cont) goto REDO;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


Scheme_Object *scheme_Tree_initialize(Scheme_Env *env);
void objscheme_init();
void objscheme_add_procedures(Scheme_Env *env);


int main(int argc, char *argv[])
{
#if 0
	test_hpc(1,true);
#else

	char* mzschemepath=getenv("MZSCHEME");

	if (!mzschemepath || !strlen(mzschemepath))
	{
		printf(
			"can't find mzscheme directory. PLaSM need to access files in mzscheme/collects.\n"
			"Please install mzscheme and execute in your shell the following commands:\n"
			"\n"
			"   export MZSCHEME=/lib/plt             (bash)\n"
			"   set MZSCHEME /lib/plt                (tclsh)\n"
			"   SET MZSCHEME=\"c:\\Programs\\plt\"   (Win32\n"
			"\n\n"
		);

		return -1;
	}

#ifdef WIN32
	for (int u=0;u<strlen(mzschemepath);u++) 
		if (mzschemepath[u]=='\\') mzschemepath[u]='/';
#endif

	printf("Using MzScheme %d.%d, scripts from [%s]\n",MZSCHEME_VERSION_MAJOR,MZSCHEME_VERSION_MINOR,mzschemepath);

	//It is important for numeric format
	setlocale(LC_ALL, "POSIX");

	/* required for OS X, only */
	scheme_set_stack_base(NULL, 1); 
	
	//important, for case insensitive symbols
#if MZSCHEME_VERSION_MAJOR>208
	scheme_set_case_sensitive(0);
#endif

	//initialize environment
	Scheme_Env *e = scheme_basic_env();

#ifndef _DEBUG
	if (scheme_setjmp(scheme_error_buf)) 
	{	
		printf("an error occurred in PLaSM. Please check the \"startup.log\" file.\n");
		return -1;
	}
#endif
	
#if MZSCHEME_VERSION_MAJOR>208
	Scheme_Config *config = scheme_current_config();
#else
	Scheme_Config *config = scheme_config;;
#endif
	
	Scheme_Object *curout = scheme_get_param(config, MZCONFIG_OUTPUT_PORT);
	
	scheme_eval_string("(display \"Starting PLaSM v.5.1. Please wait....\n\")",e);

	char cmd[1024];
	sprintf(cmd,"(current-library-collection-paths (list (build-path \"%s\" \"collects\")))",mzschemepath);
	scheme_eval_string(cmd, e);

	objscheme_init();
	{
		scheme_Plasm_initialize(e);//parser
		scheme_SimpleXn_initialize(e);//geometry
	}
	objscheme_add_procedures(e);


#if defined(_DEBUG) 
	scheme_eval_string("(time (load (build-path \"scheme\" \"init.scm\")))", e);
	scheme_eval_string("(hpc-benchmark-log #t)", e);
#else
	scheme_eval_string("(time (with-output-to-file \"startup.log\" (lambda() (load (build-path \"scheme\" \"init.scm\"))) 'replace))", e);
	scheme_eval_string("(hpc-benchmark-log #t)", e);
#endif

	scheme_eval_string("(welcome (current-output-port))",e);

	//load files and execute them
	if (argc>=2)
	{
		for (int i = 1; i < argc; i++) 
		{	
			sprintf(cmd,"(load \"%s\")",argv[i]);
			Scheme_Object *v = scheme_eval_string(cmd, e);
			scheme_display(v, curout);
			scheme_display(scheme_make_character('\n'), curout);
		}
		return 0;
	}
	
	//interactive loops
	if (scheme_setjmp(scheme_error_buf)) 
		{;}

	scheme_apply(scheme_builtin_value("read-eval-print-loop"), 0, NULL);
#endif

	return 0;

}
