#include "generaldef.h"

#include <assert.h>
#include <time.h>

#include "svgsogei.h"
#include "gviewer.h"
#include "xml2dpath.h"
#include "rayshooting.h"
#include "xml2dpath.convert.h"


#include "pipeline3d.h"

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

class MyDraw: public Draw
{

	bsp* b2;
	ggraph* gsrc;
	ggraph* gsol;
	float* points;
	bool bFrontToBack;

	void getOpenGLMatrix(real* Tv);

public:

	MyDraw(bsp* b2,ggraph* gsrc,ggraph* gsol,bool bFrontToBack);

	~MyDraw();

	virtual void display();
};




//-----------------------------------------------------------------
MyDraw::MyDraw(bsp* b2,ggraph* gsrc,ggraph* gsol,bool bFrontToBack)
{
	this->b2=b2;
	this->gsrc=gsrc;
	this->gsol=gsol;
	this->bFrontToBack=bFrontToBack;
	this->points=new float[64*1024];
}


//-----------------------------------------------------------------
MyDraw::~MyDraw()
{
	delete [] points;
}



//-----------------------------------------------------------------
void MyDraw::getOpenGLMatrix(real* Tv)
{
	GLdouble modelview[16], proj[16];
	GLint viewport[4];
	glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	viewport[0] = viewport[1] = (int)0;
	viewport[2] = (int)getPanel()->getW();
	viewport[3] = (int)getPanel()->getH();

	real M[16]=
	{
		modelview[0],modelview[4],modelview[ 8],modelview[12],
		modelview[1],modelview[5],modelview[ 9],modelview[13],
		modelview[2],modelview[6],modelview[10],modelview[14],
		modelview[3],modelview[7],modelview[11],modelview[15]
	};

	real P[16]=
	{
		proj[0],proj[4],proj[ 8],proj[12],
		proj[1],proj[5],proj[ 9],proj[13],
		proj[2],proj[6],proj[10],proj[14],
		proj[3],proj[7],proj[11],proj[15]
	};

	real V[16]=
	{
		0.5*viewport[2],0              ,0   ,viewport[0]+0.5*viewport[2],
		0              ,0.5*viewport[3],0   ,viewport[1]+0.5*viewport[3],
		0              ,0              ,0.5 ,0.5,
		0              ,0              ,0   ,1
	};

	;
	tmat<real>::product(4,Tv,P,M);
	tmat<real>::product(4,Tv,V,Tv);

}


//-----------------------------------------------------------------
void MyDraw::display()
{
	float vx,vy,vz;
	this->getPanel()->getViewpoint(vx,vy,vz);

	real Tv[16];
	getOpenGLMatrix(Tv);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(panel->getX(), panel->getX()+panel->getW(), panel->getY(), panel->getY()+panel->getH(), -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	{
		glEnable(GL_COLOR_MATERIAL);
		glDisable(GL_LIGHTING);

		glDisable(GL_DEPTH_TEST);

		{
			//------------------------------------------------
			//SVG part
			//------------------------------------------------
			clock_t t1=clock();
			int width=(int)(getPanel()->getW());
			int height=(int)(getPanel()->getH());
			SvgSogei* svg=new SvgSogei(b2,gsrc,gsol,vx,vy,vz,Tv,width,height,bFrontToBack);

			svg->dump();

			//opengl-related part
			int npols=svg->getNumPolygons();
			for (int i=0;i<npols;i++)
			{
				float R,G,B;
				int numpoints;
				svg->get2dPolygon(i,R,G,B,numpoints,points);

				int j;
				glColor3f(R,G,B);
				glBegin(GL_POLYGON);
				for (j=0;j<numpoints*2;j+=2)
					glVertex2f(points[j],points[j+1]);
				glEnd();

				glColor3f(0,0,0);
				glLineWidth(2);
				glBegin(GL_LINES);
				for (j=0;j<numpoints*2-2;j+=2)
				{
					glVertex2f(points[j+0],points[j+1]);
					glVertex2f(points[j+2],points[j+3]);
				}
				glVertex2f(points[numpoints*2-2],points[numpoints*2-1]);
				glVertex2f(points[0],points[1]);
				glEnd();
				glLineWidth(1);
			}

			delete svg;
		}

		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);
		getPanel()->getWindow()->redisplay();
	}

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}



//______________________________________________________________________
int main(int argn,char** argv)
{
	char* filename=0;
	int ilimit=0;
	bool bFrontToBack=true;
	bool bPipeline=false;
	float Tv[16];
	int width=800;
	int height=800;
	float observer[3];
	char* outfilename="temp.svg";
	bool bHttp=false;
	int i;

	for (i=1;i<argn;)
	{
		if (!strcmp(argv[i],"filename"))
		{
			filename=argv[i+1];
			i+=2;
		}
		else if (!strcmp(argv[i],"outfilename"))
		{
			outfilename=argv[i+1];
			i+=2;
		}
		else if (!strcmp(argv[i],"http"))
		{
			i+=1;
			bHttp=true;

			printf("Content-type: text/html%c%c\n",13,10);fflush(stdout);
			printf("<html><head><title></title></head><body>\n");fflush(stdout);
			printf("<!--\n");fflush(stdout);
		}
		else if (!strcmp(argv[i],"limit"))
		{
			ilimit=atoi(argv[i+1]);
			i+=2;
		}
		else if (!strcmp(argv[i],"width"))
		{
			width=atoi(argv[i+1]);
			i+=2;
		}
		else if (!strcmp(argv[i],"height"))
		{
			height=atoi(argv[i+1]);
			i+=2;
		}
		else if (!strcmp(argv[i],"fronttoback"))
		{
			bFrontToBack=true;
			i+=1;
		}

		else if (!strcmp(argv[i],"backtofront"))
		{
			bFrontToBack=false;
			i+=1;
		}


		else if (!strcmp(argv[i],"pipeline"))
		{
			printf("*** Using pipeline3d library ***\n");fflush(stdout);
			
			bPipeline=true;
			i+=1;

			char* type=argv[i++];
			printf("Type of projection: %s\n",type);fflush(stdout);

			/* Lettura del VRP */
			double vrp[3];
			vrp[0]=strtod(argv[i++],NULL);vrp[1]=strtod(argv[i++],NULL);vrp[2]=strtod(argv[i++],NULL);
			printf("VRP: [%f, %f, %f]\n",vrp[0], vrp[1], vrp[2]);fflush(stdout);
			
			/* Lettura del VUV */
			double vuv[3];
			vuv[0]=strtod(argv[i++],NULL);vuv[1]=strtod(argv[i++],NULL);vuv[2]=strtod(argv[i++],NULL);
			printf("VUP: [%f, %f, %f]\n",vuv[0], vuv[1], vuv[2]);fflush(stdout);
			
			/* Lettura del VPN */
			double vpn[3];
			vpn[0]=strtod(argv[i++],NULL);vpn[1]=strtod(argv[i++],NULL);vpn[2]=strtod(argv[i++],NULL);
			printf("VPN: [%f, %f, %f]\n",vpn[0], vpn[1], vpn[2]);fflush(stdout);
			
			/* Lettura del COP */
			double cop[3];
			cop[0]=strtod(argv[i++],NULL);cop[1]=strtod(argv[i++],NULL);cop[2]=strtod(argv[i++],NULL);
			printf("COP: [%f, %f, %f]\n",cop[0], cop[1], cop[2]);fflush(stdout);
			
			/* Lettura dei piani front e back */
			double front=strtod(argv[i++],NULL);
			double back =strtod(argv[i++],NULL);
			printf("Front: %f\nBack: %f\n",front,back);fflush(stdout);

			/* Lettura della p3d_box */
			double range_u[2],range_v[2];
			range_u[0]=strtod(argv[i++],NULL);range_u[1]=strtod(argv[i++],NULL);
			range_v[0]=strtod(argv[i++],NULL);range_v[1]=strtod(argv[i++],NULL);
			printf("BOX: [(%f, %f); (%f, %f)]\n",range_u[0], range_u[1], range_v[0], range_v[1]);fflush(stdout);

			/* Lettura del viewport 3D (la coord. z viene cablata nel codice */
			double range_x[2],range_y[2];
			range_x[0]=strtod(argv[i++],NULL);range_x[1]=strtod(argv[i++],NULL);
			range_y[0]=strtod(argv[i++],NULL);range_y[1]=strtod(argv[i++],NULL);
			printf("VIEWPORT: [(%f, %f); (%f, %f)]\n",range_x[0], range_x[1], range_y[0], range_y[1]);fflush(stdout);
		
			makePipeline3d
			(
				Tv,observer,
				type,
				vrp,vuv,vpn,cop,
				front,back,
				range_u,range_v,range_x,range_y
			);
		}
		else
		{
SYNTAX_ERROR:
			fprintf(stdout,"Syntax error: use %s\n");fflush(stdout);
			fprintf(stdout,"    filename <char*::filename>\n");fflush(stdout);
			fprintf(stdout,"    [outfilename <char*:temp.svg>]");fflush(stdout);
			fprintf(stdout,"    [width <int:800>] \n");fflush(stdout);
			fprintf(stdout,"    [height <int:800>]\n"); fflush(stdout);
			fprintf(stdout,"    [limit <int>] \n");fflush(stdout);
			fprintf(stdout,"    [fronttoback | backtofront] \n");fflush(stdout);
			fprintf(stdout,"    [pipeline ...]\n");fflush(stdout);
			fprintf(stdout,"    [http]\n");fflush(stdout);
			return -1;
		}
	}

	if (!filename)
		goto SYNTAX_ERROR;


	fprintf(stdout,"Actual configuration\n");
	fprintf(stdout,"    filename    %s\n",filename);fflush(stdout);
	fprintf(stdout,"    outfilename %s\n",outfilename);fflush(stdout);
	fprintf(stdout,"    width       %d\n",width);fflush(stdout);
	fprintf(stdout,"    height      %d\n",height); fflush(stdout);
	fprintf(stdout,"    limit       %d\n",ilimit);fflush(stdout);
	fprintf(stdout,"    fronttoback %s\n",bFrontToBack?"yes":"no");fflush(stdout);
	fprintf(stdout,"    http        %s\n",bHttp?"yes":"no");fflush(stdout);
	fprintf(stdout,"\n");
	fflush(stdout);


	// parsing
	clock_t t1=clock();
	printf("Parsing xml-sogei file %s...",filename);fflush(stdout);
	Xml2DPath* x=new Xml2DPath(filename,ilimit);
	x->parse();
	ggraph* g=x->getGraph();
	printf("Done in %.2f seconds\n",(clock()-t1)/(float)CLOCKS_PER_SEC);fflush(stdout);
	printf("%d points %d edges %d faces\n",(int)g->getNCells(0),(int)g->getNCells(1),(int)g->getNCells(2));fflush(stdout);

	//build ray-shooting structure
	t1=clock();
	printf("building ray shooting structure (for point location)...");fflush(stdout);
	RayShooting2d* rs=new RayShooting2d(0,1,g);
	printf("Done in %.2f seconds\n",(clock()-t1)/(float)CLOCKS_PER_SEC);fflush(stdout);

	// create only one 2-cell which contains all edges
	t1=clock();
	printf("building main 2D ggraph with one cell...");fflush(stdout);
	ggraph* gone=g->clone();
	key Fone=gone->addNode(2);
	ggraph::list_iterator it,itnext;
	for (it=gone->each(1);!it.end();it++) gone->addArch(*it,Fone);
	for (it=gone->each(0);!it.end();it++) gone->addArch(Fone,*it);
	for (it=gone->each(2);!it.end();it=itnext) {itnext=it.next();if (*it!=Fone) gone->remNode(*it);}
	printf("Done in %.2f seconds\n",(clock()-t1)/(float)CLOCKS_PER_SEC);fflush(stdout);

	// conversion
	t1=clock();
	printf("Doing brep to full conversion...");fflush(stdout);
	ggraph* gfull=ggraph::cuboid(2);
	real vt[2]={-0.5,-0.5},vs[2]={2.1f,2.1f};
	gfull->translate(vt);
	gfull->scale(vs);
	bsp* b2=new bsp(2);
	b2->block_expand=true; //I can use More to store informations!
	brep_convert(rs,b2,b2->getRoot(),gone,Fone,gfull,*gfull->each(2));
	printf("Done in %.2f seconds\n",(clock()-t1)/(float)CLOCKS_PER_SEC);fflush(stdout);

	// viewer-interactive
	if (!bPipeline)
	{
		printf("Finally starting the OpenGL viewer\n");fflush(stdout);
		XgeViewer* win=new XgeViewer;
		win->push(new MyDraw(b2,g,gfull,bFrontToBack));
		win->run();
		delete win;
	}
	else
	{
		SvgSogei* svg=new SvgSogei(b2,g,gfull,observer[0],observer[1],observer[2],Tv,width,height,bFrontToBack);
		svg->dump(outfilename);
		delete svg;
	}

	//deallocating
	delete gone;
	delete gfull;
	delete rs;
	delete x;

	if (bHttp)
	{
		printf("-->\n");fflush(stdout);
		printf("<embed src=\"%s\" width=\"800\" height=\"800\" type=\"image/svg+xml\" pluginspage=\"http://www.adobe.com/svg/viewer/install/\" />\n",outfilename);fflush(stdout);
		printf("</body><html>\n");fflush(stdout);
	}

#if defined(WIN32) && defined(_DEBUG)
	printf("press a char to finish\n");
	getchar();
#endif


	return 0;
}
