#include "generaldef.h"
#include <time.h>

#include "svgsogei.h"
#include "scanline.h"
#include "xml2dpath.h"
#include "xml2dpath.convert.h"


//-------------------------------------------------------------------------
SvgSogei::SvgSogei(bsp* b2,ggraph* gsrc,ggraph* gsol,
			float vx,float vy,float vz,float* Tv,int width,int height,bool bFrontToBack)
		:npolygon(0),nbackculling(0),noutofscene(0),nnotvisible(0),cbuffer(0)
{
	this->b2=b2;
	this->gsrc=gsrc;
	this->gsol=gsol;

	this->Tv=Tv;
	this->vx=vx;
	this->vy=vy;
	this->vz=vz;
	this->width=width;
	this->height=height;
	this->bFrontToBack=bFrontToBack;

	this->cx=0;
	this->cy=0;
	this->cz=0;

	if (bFrontToBack)
		scan_init(width,height,1,1,1);
}

//-------------------------------------------------------------------------
void SvgSogei::dump(char* filename)
{
	clock_t t1=clock();
	visit(b2,b2->getRoot(),vx,vy,vz);

	FILE* file;
	
	if (!strcmp(filename,"stdout"))
		file=stdout;
	else
		file=fopen(filename,"wt");

	fprintf(file,"<svg width='%d' height='%d' viewBox='0 0 %d %d'>\n",width,height,width,height);
	fprintf(file,"<g stroke='black' stroke-width='0'>\n",min2(width,height)/1000.0f); 

	std::deque< std::pair<std::list<Point> , Color> >::iterator it;
	for (it=list_to_export.begin();it!=list_to_export.end();it++)
	{
		std::pair< std::list<Point> , Color > item=*it;
		std::list<Point> path=item.first;
		Color clr=item.second;
		dump(file,path,clr.R,clr.G,clr.B);
	}

	fprintf(file,"</g>\n");
	fprintf(file,"</svg>\n");
	
	if (!strcmp(filename,"stdout"))
		;
	else
		fclose(file);

	printf("Done SVG Export in %.2f seconds\n",(clock()-t1)/(float)CLOCKS_PER_SEC);
	printf("  output file       %s\n",filename);
	printf("  export_method     %s\n",bFrontToBack?"front-to-back":"back-to-front");
	printf("  viewpoint         %f %f %f\n",vx,vy,vz);
	printf("  dimension         W=%d H=%d\n",width,height);
	printf("  total_polygon     %d\n",(int)npolygon);
	printf("  exported_faces    %d (%d perc)\n",(int)list_to_export.size(),(int)(100.0f*list_to_export.size()/(float)npolygon));
	printf("  back_face_culled  %d (%d perc)\n",(int)nbackculling         ,(int)(100.0f*nbackculling         /(float)npolygon));
	printf("  out_of_scene      %d (%d perc)\n",(int)noutofscene          ,(int)(100.0f*noutofscene          /(float)npolygon));
	printf("  not_visible       %d (%d perc)\n",(int)nnotvisible          ,(int)(100.0f*nnotvisible          /(float)npolygon));
	printf("\n");
}

//-------------------------------------------------------------------------
void SvgSogei::get2dPolygon(int idx,float& R,float& G,float &B,int& numpoints,float* points)
{
	std::list<Point> path=list_to_export[idx].first;
	Color clr=list_to_export[idx].second;
	
	R=clr.R;G=clr.G;B=clr.B;
	float* t=points;
	std::list<Point>::iterator it,itnext;
	for (it=path.begin();it!=path.end();it++)
	{
		Point p=*it;transform(p);
		*t++=p.x;
		*t++=p.y;
	}
	
	numpoints=path.size();

}

//-------------------------------------------------------------------------
void SvgSogei::dump(FILE* file,std::list<Point> path,float R,float G,float B)
{

#if 0
	if (bFrontToBack)
	{
		glRasterPos2i(20,50);
		glDrawPixels(width,height,GL_RGB,GL_FLOAT,cbuffer); 
		return;
	}
#endif

	/*
	glPolygonMode(GL_FRONT_AND_BACK ,GL_FILL);
	glColor3f(0.5,0.5,0.5);
	drawFaces(gT,0,0,0,false);
	glLineWidth(3);
	glPolygonMode(GL_FRONT_AND_BACK ,GL_LINE);
	glColor3f(0,0,0);
	drawFaces(gT,0,0,0,false);
	glLineWidth(1);
	*/

	fprintf(file,"<path fill='rgb(%d,%d,%d)' d='m ",(int)(R*255),(int)(G*255),(int)(B*255));
	float curx=0,cury=0;
	for (std::list<Point>::iterator it=path.begin();it!=path.end();it++)
	{
		Point p=*it;transform(p);
		p.y=(-1*p.y+height);
		fprintf(file,"%f %f ",p.x-curx,p.y-cury);
		curx=p.x;cury=p.y;
	}

	Point p=*path.begin();transform(p);
	p.y=(-1*p.y+height);
	fprintf(file,"%f %f ",p.x-curx,p.y-cury);
	fprintf(file,"z' />\n");
}


//-------------------------------------------------------------------------
void SvgSogei::flush()
{
	++npolygon;
	cx/=path.size();
	cy/=path.size();
	cz/=path.size();

	float dx=vx-cx; float nx=(bReverse?-1:+1)*this->nx;
	float dy=vy-cy; float ny=(bReverse?-1:+1)*this->ny;
	float dz=vz-cz; float nz=(bReverse?-1:+1)*this->nz;
	float mdir=sqrt(dx*dx+dy*dy+dz*dz);
	dx/=mdir;dy/=mdir;dz/=mdir;
	float cangle=dx*nx+dy*ny+dz*nz;

	int MyColor=colorfield[which];
	float R=((MyColor    ) & 0xFF)/255.0f;
	float G=((MyColor>>8 ) & 0xFF)/255.0f;
	float B=((MyColor>>16) & 0xFF)/255.0f;

#if 0
	R=0.7*cangle+0.3*colors[C][0];
	G=0.7*cangle+0.3*colors[C][1];
	B=0.7*cangle+0.3*colors[C][2];
#endif

	float hue,saturation,lightness;
	RGB2HSL(hue,saturation,lightness,R,G,B);
	lightness*=cangle;
	HSL2RGB(R,G,B,hue,saturation,lightness);

	//cull back faces
	if (cangle<0)
	{
		++nbackculling;
		path.clear();
		return;
	}

	//out of scene
	if (true)
	{
		bool bInScene=false;
		for (std::list<Point>::iterator it=path.begin();it!=path.end();it++)
		{
			Point p=*it;transform(p);

			if (p.x>=0 && p.x<width && p.y>=0  && p.y<height)
				{bInScene=true;break;}
		}

		if (!bInScene)
		{
			noutofscene++;
			path.clear();
			return;
		}
	}

	if (bFrontToBack)
	{
		float* v=new float[path.size()*3];
		float *t=v;

		for (std::list<Point>::iterator it=path.begin();it!=path.end();it++)
		{
			Point p=*it;transform(p);
			*t++=p.x;*t++=p.y;*t++=p.z;
		}

		if (!scan_add_polygon(path.size(),v,R,G,B))
		{
			delete [] v;
			++nnotvisible;
			path.clear();
			return;
		}

		delete [] v;
	}

	Color clr={R,G,B};
	std::pair< std::list<Point> ,Color > item(path,clr);

	if (bFrontToBack)
		list_to_export.push_front( item) ;
	else
		list_to_export.push_back( item) ;

	path.clear();
}



//---------------------------------------------------------------------------
void SvgSogei::visit(bsp* b2,key N,float vx,float vy,float vz)
{
	if (b2->isIn(N))
	{
		key Cell=cut_to_full[N];
		key which=full_to_which[Cell];

		normal(which,false,0,0,1);

		//TOP-full
		key prev0=gsol->getFirstDwNode(gsol->getLastDwNode(Cell)),p1;
		key prev1=gsol->getLastDwNode (gsol->getLastDwNode(Cell)),p2;
		for (graph::iterator i=gsol->goDw(Cell);!i.end();i++,prev0=p1,prev1=p2)
		{
			p1=gsol->getFirstDwNode(*i);
			p2=gsol->getLastDwNode (*i);
			if (p1!=prev0 && p1!=prev1) Swap(key,p1,p2);
			real* g1=gsol->getGeom(p1);
			add(g1[1],g1[2],scalarfield[which]);
		}
		flush();

		//TOP-lines
		/*
		glColor3f(rand()/(float)RAND_MAX,rand()/(float)RAND_MAX,rand()/(float)RAND_MAX);
		glLineWidth(3);
		glBegin(GL_LINES);
		key prev0=gsrc->getFirstDwNode(gsrc->getLastDwNode(which)),p1;
		key prev1=gsrc->getLastDwNode (gsrc->getLastDwNode(which)),p2;			
		for (graph::iterator i=gsrc->goDw(which);!i.end();i++,prev0=p1,prev1=p2)
		{
			p1=gsrc->getFirstDwNode(*i);real* g1=gsrc->getGeom(p1);
			p2=gsrc->getLastDwNode (*i);real* g2=gsrc->getGeom(p2);
			glVertex3f(g1[1],g1[2],scalarfield[which]);
			glVertex3f(g2[1],g2[2],scalarfield[which]);
		}
		glEnd();
		glLineWidth(1);
		*/

		return;
	}

	if (b2->isOut(N))
		return;

	real* h=b2->getPlane(N);
	real acc=h[0]+h[1]*vx+h[2]*vy+0*vz;

	key below=b2->getBelow(N);
	key above=b2->getAbove(N);
	if (acc<0) {Swap(key,below,above);}

	if (bFrontToBack) {Swap(key,below,above);}
	visit(b2,below,vx,vy,vz);

	//extrusion
	for (ggraph::list_iterator eit=gsol->each(1);!eit.end();eit++)
	{
		key Edge=*eit;
		int nup=gsol->getNUp(Edge);
		key Id=gsol->getNode(Edge)->Id;

		//only for boundary edge,only boundary face which correspond to current node
		if (Id!=N)
			continue;

		checkpoint(nup==1 || nup==2);
		key Up=gsol->getFirstUpNode(Edge);
		key which=full_to_which[Up];
		float from=0,to=scalarfield[which];

		if (nup==2)
		{
			key Up2=gsol->getLastUpNode(Edge);
			key which2=full_to_which[Up2];
			if (which2==which) continue;
			if (scalarfield[which2]>scalarfield[which]) {Swap(key,Up,Up2);Swap(key,which,which2);}
			from=scalarfield[which2];
			to  =scalarfield[which];
		}
		
		//fix vertex order
		real* h=b2->getPlane(N);
		real* bball=gsol->boundingBall(Up); //i know there is only one up-cell
		bool bReverse=bball[0]>0?((h[0]+h[1]*bball[1]+h[2]*bball[2])>0):false;
		real* g0=gsol->getGeom(gsol->getFirstDwNode(Edge));
		real* g1=gsol->getGeom(gsol->getLastDwNode(Edge));

		normal(which,bReverse,h[1],h[2],0);
		add(g0[1],g0[2],from);
		add(g1[1],g1[2],from);
		add(g1[1],g1[2],to  );
		add(g0[1],g0[2],to  );
		flush();
	}

	visit(b2,above,vx,vy,vz);
}
