#ifndef _SVG_SOGEI_H__
#define _SVG_SOGEI_H__

#include "generaldef.h"

#include <map>
#include <deque>
#include <list>

#include "bsp.h"
#include "ggraph.h"

class SvgSogei
{
	//////////////////////
	// typedefs
	//////////////////////

	typedef struct{float x,y,z;}  Point;

	typedef struct{float R,G,B;} Color;

	//bsp
	bsp* b2;

	//source and solution graph 
	ggraph *gsrc,*gsol;

	//actual path
	std::list<Point> path; 

	//svg screen viewport 
	int width,height; 

	//actual normal
	float nx,ny,nz; 

	//face id (source cell)
	key which;

	//reverse normal
	bool bReverse; 

	//centroid
	float cx,cy,cz; 

	//viewpoint
	float vx,vy,vz;
	
	//global transformation
	float* Tv;

	//actual color
	float  R,G,B;

	//front to back or back to front
	bool  bFrontToBack;

	//color buffer for front-to-back visit
	float* cbuffer;

	//polygons total,exported, back face culling, out of screen, not visibile (front-to-back)
	int   npolygon,nbackculling,noutofscene,nnotvisible;

	//list of path to export
	std::deque< std::pair<std::list<Point> , Color> > list_to_export;

	//FUNCTION
	void transform(Point& p);

	void visit(bsp* b2,key N,float vx,float vy,float vz);

	void dump(FILE* file,std::list<Point> path,float R,float G,float B);

public:

	//constructor
	SvgSogei(
			bsp* b2,                     // converted bsp
			ggraph* gsrc,                //source 2d path
			ggraph* gsol,                //solution (full) 2d path
			float vx,float vy,float vz,  // view point for bsp traversal
			float* Tv,                   // global transformation for projection
			int width,int height,        //final svg resolution in pixel
			bool bFrontToBack           //how to export (front to back or back to front)
	);

	//export to file
	void dump(char* filename="temp.svg");

	/* to add projected face */
	void normal(key which,bool bReverse,float nx,float ny,float nz);

	void add(float x,float y,float z);

	void flush();

	//return the number of polygon to export
	int getNumPolygons();

	//return the idx-th polygon
	void get2dPolygon(int idx,float& R,float& G,float &B,int& numpoints,float* points);

};



/////////////////////////
// INLINES
/////////////////////////

inline void SvgSogei::add(float x,float y,float z)
{
	Point p;
	p.x=x;cx+=x;
	p.y=y;cy+=y;
	p.z=z;cz+=z;
	path.push_back(p);
}

inline void SvgSogei::normal(key which,bool bReverse,float nx,float ny,float nz)
{
	this->nx=nx;
	this->ny=ny;
	this->nz=nz;
	this->which=which;
	this->bReverse=bReverse;
}


inline void SvgSogei::transform(Point& p)
{
	real tmp=Tv[12]*p.x+Tv[13]*p.y+Tv[14]*p.z+Tv[15];
	real X=(Tv[ 0]*p.x+Tv[ 1]*p.y+Tv[ 2]*p.z+Tv[ 3])/tmp;
	real Y=(Tv[ 4]*p.x+Tv[ 5]*p.y+Tv[ 6]*p.z+Tv[ 7])/tmp;
	real Z=(Tv[ 8]*p.x+Tv[ 9]*p.y+Tv[10]*p.z+Tv[11])/tmp;
	p.x=X;
	p.y=Y;
	p.z=Z;
}

inline int SvgSogei::getNumPolygons()
{
	return list_to_export.size();
}



#endif //_SVG_SOGEI_H__



