#include "generaldef.h"

#include "xml2dpath.convert.h"


std::map<key,key> full_to_which;
std::map<key,key> cut_to_full;
std::map<key,key> cut_to_src;

void brep_convert(RayShooting2d* rs,
						bsp* b,key N,
						ggraph* grep,key Frep,
						ggraph* gfull,key Cfull,
						float x1,float y1,float x2,float y2,
						int nrec,
						float tolerance)
{
	real h[3];

	float xm=x1+(x2-x1)/2;
	float ym=y1+(y2-y1)/2;

	//quadtree
	if (grep->getNDw(Frep)>NUM_FACES_USE_QUAD)
	{
		h[0]=(nrec & 1)?-ym:-xm;
		h[1]=(nrec & 1)?0:1;
		h[2]=(nrec & 1)?1:0;
	}
	else
	{
		key CrepDown=grep->getFirstDwNode(Frep);
		grep->remArch(CrepDown,Frep);
		real* hsrc=grep->getGeom(CrepDown);
		h[0]=hsrc[0];
		h[1]=hsrc[1];
		h[2]=hsrc[2];
	}

	b->setCutn(N,h,b->out(),b->out());

	key Frepminus=0,Frepplus=0;

	while (grep->getNDw(Frep))
	{
		key CDown=grep->getFirstDwNode(Frep);
		key CDownMinus,CDownPlus,CDownEqual;

		int retcode=grep->split(CDown,h,tolerance,CDownMinus,CDownPlus,CDownEqual);

		switch(retcode)
		{
		case (ggraph::OK):
			{
				if (CDownMinus) 
				{
					if (!Frepminus) Frepminus =grep->addNode(2);
					grep->addArch(CDownMinus,Frepminus);
					grep->remArch(CDownMinus,Frep);
				}

				if (CDownPlus) 
				{
					if (!Frepplus) Frepplus =grep->addNode(2);
					grep->addArch(CDownPlus ,Frepplus);
					grep->remArch(CDownPlus, Frep);
				}
				break;
			} 
		case (ggraph::FLAT):
			{
				grep->remArch(CDown,Frep);
				break;
			}
		default:
			{
				fprintf(stderr,"internal error. impossible to split in file %s line %d!\n",__FILE__,__LINE__);
				abort();
				break;
			}
		}
	}

	int nbelow=grep->getNDw(Frepminus);
	int nabove=grep->getNDw(Frepplus);

	grep->remNode(Frep);

	//split full cell
	key Cfullminus,Cfullplus,Cfullequal;

	{
		key Id=N;
		int retcode=gfull->split(Cfull,h,tolerance,Cfullminus,Cfullplus,Cfullequal,Id);

		int maxnumtry=10;
		for (int numtry=0;retcode==ggraph::IMPOSSIBLE && numtry<maxnumtry;numtry++)
		{
			real perplane[3];
			printf(" perturbation ");fflush(stdout);

			perplane[0]=(real)(h[0]+numtry*1e-4*(rand()/(real)RAND_MAX-0.5));
			perplane[1]=(real)(h[1]+numtry*1e-4*(rand()/(real)RAND_MAX-0.5));
			perplane[2]=(real)(h[2]+numtry*1e-4*(rand()/(real)RAND_MAX-0.5));

			retcode=gfull->split(Cfull,perplane,tolerance,Cfullminus,Cfullplus,Cfullequal,Id);
		}

		 //assign to above
		if (retcode==ggraph::FLAT)
		{
			Cfullplus		=Cfull;
			Cfullminus	   =Cfullequal=0;
			retcode	=ggraph::OK;
		}

		checkpoint(retcode==ggraph::OK);
	}

	if (Frepminus && Cfullminus) 
	{
		brep_convert(rs,b,b->getBelow(N),grep,Frepminus,gfull,Cfullminus,x1,y1,(nrec & 1)?x2:xm,(nrec & 1)?ym:y2,nrec+1,tolerance);
	}
	else if (Cfullminus)
	{
		real* center=gfull->boundingBall(Cfullminus);
		key which=rs->getCell(center[1],center[2]);

		if (which) 
		{
			key below=b->getBelow(N);
			b->setIn(below);

			cut_to_full[below]=Cfullminus;
			cut_to_src[below]=which;
			full_to_which[Cfullminus]=which;
		}
		else
			gfull->remNode(Cfullminus);
	}

	if (Frepplus && Cfullplus)
	{
		brep_convert(rs,b,b->getAbove(N),grep,Frepplus ,gfull,Cfullplus,(nrec & 1)?x1:xm,(nrec & 1)?ym:y1,x2,y2,nrec+1,tolerance);
	}
	else if (Cfullplus)
	{
		real* center=gfull->boundingBall(Cfullplus);
		key which=rs->getCell(center[1],center[2]);
		if (which) 
		{
			key above=b->getAbove(N);
			b->setIn(above);

			cut_to_full[above]=Cfullplus;
			cut_to_src[above]=which;
			full_to_which[Cfullplus]=which;
		}
		else
			gfull->remNode(Cfullplus);
	}
}