
#include "generaldef.h"

#include <libxml/xmlerror.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h>
#include <libxml/tree.h>
#include <libxml/debugXML.h>
#include <libxml/xmlmemory.h>

#include <time.h>


#include "xml2dpath.h"


std::map<key,float> scalarfield;
std::map<key,int>   colorfield;

//______________________________________________________________________
Xml2DPath::Xml2DPath(char* filename,int limit)
{
	this->limit=limit;
	strcpy(this->filename,filename);
	this->g2=new ggraph(2);
}


//______________________________________________________________________
Xml2DPath::~Xml2DPath()
{
	delete g2;
}

//______________________________________________________________________
ggraph* Xml2DPath::getGraph()
{
	return g2;
}

//______________________________________________________________________
void Xml2DPath::startElement(void *ctx, const xmlChar *name, const xmlChar **atts) 
{
	Xml2DPath* xp=(Xml2DPath*)ctx;
	ggraph* g=xp->g2;
	int R=255,G=255,B=0; //default color=yellow

	if (atts == NULL)
		return;	

	if (xp->limit && g->getNCells(2)==xp->limit)
		return;


	bool bExistScalarfield=false;
	float cur_scalarfield_value=1;

	for (int a = 0;(atts[a] != NULL);a+=2) 
	{
		const xmlChar *label	= atts[a+0];
		const xmlChar *value	= atts[a+1];
		
		if (!strcmp((char*)label,"id"))
		{
			;
		}
		else if (!strcmp((char*)label,"fill"))
		{

			const char* _p=(const char*)value;
			_p=strchr( _p,'(')+1;
			R=atoi(_p);_p=strchr(_p,',')+1;
			G=atoi(_p);_p=strchr(_p,',')+1;
			B=atoi(_p);

		}
		else if (!strcmp((char*)label,"value"))
		{
			bExistScalarfield=true;
			cur_scalarfield_value=atof((char*)value);
			cur_scalarfield_value=0.6*cur_scalarfield_value/(float)9206.874858;
		}
		else if (!strcmp((char*)label,"d"))
		{
			double cur[2]={0,0};
			int c=0;
			int nloop=0;
			const xmlChar* start=0;
			
			key F=g->addNode(2);

			if (bExistScalarfield)
			{
				scalarfield[F]=cur_scalarfield_value;
			}
			else
			{
				scalarfield[F]=0.1+0.1*rand()/(float)RAND_MAX;
				scalarfield[F]*=0.5;
			}

			colorfield[F]=(R & 0xff) | ((G & 0xff)<<8) | ((B & 0xff)<<16);

			key N0=0;
			key NFirst=0;
			key loopedges[8192];
			key nloopedges=0;
			const xmlChar* p=value;

			while (true)
			{
				switch(*p)
				{
				case  0 ://special case when there isn't the "z" closure
				case 'm'://open new loop inside the current path
				case 'z'://closure
					
					if (nloopedges && nloopedges>=3)
					{
						++nloop;
						
						for (int i=0;i<nloopedges;i++)
						{
							key N0=g->getFirstDwNode(loopedges[i]);
							key N1=g->getLastDwNode (loopedges[i]);
							g->addArch(loopedges[i],F);

							if (!g->findArch(F,N0,DIRECTION_UP)) g->addArch(F,N0);
							if (!g->findArch(F,N1,DIRECTION_UP)) g->addArch(F,N1);
						}
						
						{
							key E=loopedges[nloopedges-1];				
							key V0=g->getFirstDwNode(E);
							key V1=g->getLastDwNode(E);
							key V=(g->getNUp(V0)==1)?V0:V1;
							g->addArch(NFirst,E);
							g->remNode(V);
						}
					}

					nloopedges=0;

					if (*p=='m')
					{
						++p;//jump space
						assert(*p==(unsigned char)' ');
						start=p+1;
						N0=0;
						NFirst=0;
					}
					
					break;
					
				case ' ': //separator of numbers
					
					if (start)
					{
						cur[c]+=atof((const char*)start);
						c=1-c;
						
						if (!c) 
						{
							real v[]={1,cur[0],cur[1]};
							key N1=g->addVertex(v);
							
							if (!NFirst) 
								NFirst=N1;
							
							if (N0)
							{
								key E=g->addEdge(N0,N1);
								real* h=g->setGeom(E);
								loopedges[nloopedges++]=E;
							}
							
							N0=N1;
						}
					}
					
					start=p+1;
					break;
					
				default:
					break;
				}

				if (!*p || *p=='z') 
				{
					break;
				}
				p++;
			}	
		}
	}
}


//______________________________________________________________________
void Xml2DPath::parse()
{
	xmlSAXHandler callbacks={0,0,0,0,0,0,0,0,0,0,0,0,0,0,startElement,0,0,0,0,0,0,0,0,0,0,0,0};
	
	int res = xmlSAXUserParseFile(&callbacks, this, this->filename);
	
	if (res != 0)
	{
		fprintf(stderr,"error in parsing XML file %s\n",filename);
		fflush(stderr);
		throw "error parsing XML file";
	}

	//remove topology errors
	removeWrongCells();

	setNormals();
	g2->scaleToUnitBox();

	real vs1[]={1,-1};
	g2->scale(vs1);
	setNormals(); //more precision
}


//______________________________________________________________________
void Xml2DPath::setNormals()
{
	for (ggraph::list_iterator it=g2->each(1);!it.end();it++)
	{
		key E=*it;
		key N0=g2->getFirstDwNode(E);
		key N1=g2->getLastDwNode(E);
	
		real* h=g2->setGeom(E);
		real* g0=g2->getGeom(N0);
		real* g1=g2->getGeom(N1);

		real x0=g0[1],y0=g0[2];
		real x1=g1[1],y1=g1[2];
		
		h[1]=y0-y1;
		h[2]=x1-x0;
		real n=(real)sqrt(h[1]*h[1]+h[2]*h[2]);
		h[1]/=n;
		h[2]/=n;
		h[0]=-h[1]*x0-h[2]*y0;
	}
}



//______________________________________________________________________
void Xml2DPath::removeWrongCells()
{
	//remove all vertices isolated 
	ggraph::list_iterator it,next;

	for (it=g2->each(0);!it.end();it=next)
	{
		next=it.next();
		if (!g2->getNDw(*it)) g2->remNode(*it);
	}

	//remove not good edges
	for (it=g2->each(1);!it.end();it=next)
	{
		next=it.next();
		if (g2->getNDw(*it)!=2) g2->remNode(*it);
	}

	//remove not good loops
	for (it=g2->each(2);!it.end();it=next)
	{
		next=it.next();
		if (g2->getNDw(*it)<3) g2->remNode(*it);
	}
}

//______________________________________________________________________

#if 0

void Xml2DPath::snap(ggraph* g,int istepx,int istepy)
{
	assert(istepx && istepy);

	graph::list_iterator it;

	float fstepx=2/(float)istepx;
	float fstepy=2/(float)istepy;

	for (it=g2->each(0);!it.end();it=it.next())
	{		
		key N=*it;
		real* coords=g2->getGeom(N);
		real X=coords[1];
		real Y=coords[2];
		assert(X>=-1 && X<=1 && Y>=-1 && Y<=1);

		X=-1+fstepx*((int)(0.5+(0.5*istepx*(X+1)))); \
		Y=-1+fstepy*((int)(0.5+(0.5*istepy*(Y+1))));

		coords[1]=X;
		coords[2]=Y;
	}
}

//______________________________________________________________________
void Xml2DPath::simplify(ggraph* g,int istepx,int istepy,real tolerance)
{	
	assert(istepx && istepy);

	graph::list_iterator it;

	float fstepx=2/(float)istepx;
	float fstepy=2/(float)istepy;

	key* nav=new key[1024*245];

	ggraph::list_iterator itnext;
	for (it=g2->each(2);!it.end();it=itnext)
	{		
		key F=*it;
		itnext=it.next();

		bool bRedo=true;
		while (bRedo)
		{
			bRedo=false;
			int npoints=(int)g2->findCells(0,F,nav);
				
			for (int i=0;i<npoints;i++)
			{
				key N=nav[i];
				assert(g2->getNUp(N)==2);
				key E0=g2->getFirstUpNode(N);
				key E1=g2->getLastUpNode(N);
				key N0=g2->getFirstDwNode(E0);if (N0==N) {N0=g2->getLastDwNode(E0);assert(N0!=N);}
				key N1=g2->getFirstDwNode(E1);if (N1==N) {N1=g2->getLastDwNode(E1);assert(N1!=N);}

				if (N0==N1) //should remove all the face
				{
					g2->remNode(F,true);
					bRedo=false;
					break;
				}
				else
				{
					real* C0=g2->getGeom(N0);real X0=C0[1],Y0=C0[2]; 
					real* C =g2->getGeom(N );real X =C [1],Y =C [2];
					real* C1=g2->getGeom(N1);real X1=C1[1],Y1=C1[2];

					if (fabs((X-X0)*(Y-Y1)-(Y-Y0)*(X-X1))<tolerance) 
					{
						g2->remNode(E0,false);
						g2->remNode(E1,false);
						g2->remNode(N ,false);
						key E=g2->addNode(1);
						g2->addArch(E,F);
						g2->addArch(N0,E);
						g2->addArch(N1,E);
						bRedo=true;
					}
				}
			}
		}
	}

	delete [] nav;	

#if 0
	for (it=g2->each(0);!it.end();it=it.next())
	{		
		key N=*it;
		real* coords=g2->getGeom(N);
		coords[1]+=2;
		coords[2]+=0;
	}
#endif
}


#endif
