#include <vector>
#include <assert.h>
#include <GL/glut.h>


class BspExt
{
	//example: import function from file
	void import(FILE* file);

	//default constructor
	BspExt(){};

	//spacing for print
	inline void tab(int nrec)
		{for (int i=0;i<nrec;i++) printf("   ");}

public:

	// spacedim/pointdim of the bsp
	int dim;

	//0==out 1==in 2==cut
	int type; 

	//coordinates of the splitting hyperplane (valid only for 2==cut)
	std::vector<float> h; 
	
	//below,above childs (valid only for 2==cut)
	BspExt* below;
	BspExt* above;
	
	//list of polygons (valid only for 2==cut)
	std::vector< std::vector<float> > polygons;

	//example of import from file
	BspExt(const char* filename);

	//example of print function
	void print(int nrec=0);

	//example of display function
	void display(int nrec=0,int maxrec=0);
};



//___________________________________________________________________
BspExt::BspExt(const char* filename)
{
	FILE* file=fopen(filename,"rt");
	fscanf(file,"%d",&this->dim);
	import(file);
	fclose(file);
}

//___________________________________________________________________
void BspExt::print(int nrec)
{
	tab(nrec);

	if (type==0)
	{
		printf("out\n");
		return;
	}

	if (type==1)
	{
		printf("in\n");
		return;
	}

	printf("cut npolygons(%d) h(",polygons.size());
	for (int i=0;i<=dim;i++) printf("%f ",h[i]);
	printf(")\n");

	for (i=0;i<polygons.size();i++)
	{
		tab(nrec);printf("polygon ");
		std::vector<float> points=polygons[i];

		for (int j=0;j<points.size()/dim;j++)
		{
			printf("(");
			for (int k=0;k<dim;k++) printf("%f ",points[j*dim+k]);
			printf(") ");
		}
		printf("\n");
	}

	below->print(nrec+1);
	above->print(nrec+1);

}

//___________________________________________________________________
static int getChar(FILE* file)
{
	int ch=fgetc(file);
	
	for (;ch==10 || ch==13 || ch=='\t' || ch==' ';)
		ch=fgetc(file);

	return ch;
}

//___________________________________________________________________
void BspExt::import(FILE* file)
{
	int ch=getChar(file);

	// base case
	if (ch=='i')
		{type=1;return;}

	// base case
	if (ch=='o')
		{type=0;return;}

	// recursive case
	assert(ch=='h');

	int npols;
	fscanf(file,"%d",&npols);
	type=2;
	h.resize(dim+1);
	polygons.resize(npols);

	for (int i=0;i<=dim;i++)
		fscanf(file,"%e",&h[i]);

	for (int np=0;np<npols;np++)
	{
		assert(getChar(file)=='f');
	
		int npoints;
		fscanf(file,"%d",&npoints);

		bool bReverse=npoints<0;
		npoints=bReverse?(-1*npoints):(npoints);

		polygons[np].resize(npoints*dim);

		for (int i=0;i<npoints;i++)
		{
			assert(getChar(file)=='v');

			for (int j=0;j<dim;j++)
				fscanf(file,"%e",&polygons[np][i*dim+j]);
		}

		// TODO

	}

	below=new BspExt;below->dim=dim;below->import(file);
	above=new BspExt;above->dim=dim;above->import(file);
}


//___________________________________________________________________
void BspExt::display(int nrec,int maxrec)
{
	if (type==0 || type==1)
		return;

	for (int i=0;i<polygons.size();i++)
	{
		std::vector<float> points=polygons[i];

		if (dim==2)
		{
			assert(points.size()==4);
			
			glBegin(GL_LINES);
			glVertex2f(points[0],points[1]);
			glVertex2f(points[2],points[3]);
			glEnd();
		}		
		else if (dim==3)
		{
			glBegin(GL_POLYGON);
			for (int j=0;j<points.size()/3;j++)
				glVertex3f(points[j*3+0],points[j*3+1],points[j*3+2]);
			glEnd();
		}
	}

	if (!maxrec || nrec<=maxrec)
	{
		below->display(nrec+1,maxrec);
		above->display(nrec+1,maxrec);
	}
}



/* ---------------------------------------------------------- */

static int maxrec=0;
static int moving, startx, starty;
static GLfloat angley = 0;   /* in degrees */
static GLfloat anglex = 0;   /* in degrees */

void mouse(int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON) 
  {
    if (state == GLUT_DOWN) 
		{moving = 1;startx = x;starty = y;}

    if (state == GLUT_UP) 
		{moving = 0;}
  }
}

void motion(int x, int y)
{
  if (moving) 
  {
    angley = angley + (x - startx);
    anglex = anglex + (y - starty);
    startx = x;starty = y;
    glutPostRedisplay();
  }
}

void keyboard(unsigned char key, int x, int y)
{
	if (key=='+')
		maxrec++;
	else if (key=='-' && maxrec>0)
		maxrec--;

	glutPostRedisplay();
}

void redraw(void)
{
	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glPushMatrix();
	
		glRotatef(anglex, 1.0, 0.0, 0.0);
		glRotatef(angley, 0.0, 1.0, 0.0);

		glBegin(GL_LINES);
			glColor3f(1,0,0);glVertex3f(0,0,0);glVertex3f(1,0,0);
			glColor3f(0,1,0);glVertex3f(0,0,0);glVertex3f(0,1,0);
			glColor3f(0,0,1);glVertex3f(0,0,0);glVertex3f(0,0,1);
		glEnd();

		//************ MORE INTERESTING PART ****************
		static BspExt* ext=0;
		if (!ext) ext=new BspExt("cylinder.txt");

		if (ext->dim==2)
		{
			glColor3f(0,0,0);
			ext->display(0,maxrec);
		}
		else if (ext->dim==3)
		{
			glEnable( GL_POLYGON_OFFSET_FILL );
			glPolygonOffset(1,1);
			glColor3f(0.5,0.5,0.5);
			ext->display(0,maxrec);
			glDisable(GL_POLYGON_OFFSET_FILL);
			glColor3f(0,0,0);
			glPolygonMode( GL_FRONT_AND_BACK , GL_LINE );
			ext->display(0,maxrec);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL );
		}

		//**************************************************


	glPopMatrix();
	
	glutSwapBuffers();
}


int main(int argc, char **argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
  glutInitWindowSize(800,800);
  glutCreateWindow("OpenGl application");
  glutDisplayFunc(redraw);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);

  glEnable(GL_DEPTH_TEST);
  

  glMatrixMode(GL_PROJECTION);
  gluPerspective
  ( 
		40.0,/* field of view in degree */ 
		1.0,/* aspect ratio */
		1.0, /* Z near */
		100.0/* Z far */ 
  );

  glMatrixMode(GL_MODELVIEW);
  gluLookAt(
		0.0, 0.0, 8.0,  /* eye  */
		0.0, 0.0, 0.0,  /* center  */
		0.0, 1.0, 0.);  /* up is in positive Y direction */

  glutMainLoop();
  return 0; /* ANSI C requires main to return int. */
}

