/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#include "generaldef.h"
#include "gbool.h"
#include "gviewer.h"




//-------------------------------------------------------
//-------------------------------------------------------

void GBool::print(key N,char* space,FILE* file)
{
	fprintf(file,"%s[%4d] ",space,(int)N);

	fprintf(file,"down(");

	graph::iterator i;

	for (i=maing.goDw(N);!i.end();i++)
	{
		fprintf(file,"%d ",(int)*i);
		fprintf(file,"[%d] ",(int)maing.nodeData(*i));
	}
	fprintf(file,") ");

	fprintf(file," up(");
	for (i=maing.goUp(N);!i.end();i++)
	{
		fprintf(file,"%d ",(int)*i);
		fprintf(file,"[%d] ",(int)maing.nodeData(*i));
	}
	fprintf(file,") Level(%d) ",(int)maing.getNode(N)->Level);


		
	int lvl=(int)maing.getNode(N)->Level;
	if (lvl==0)
	{
		fprintf(file,"data(%d) ",(int)maing.nodeData(N));
	}
	else if (lvl==1)
	{
		fprintf(file,"data(%0x) ",(int)maing.nodeData(N));
	}	


	char* stype=0;
	switch(nodeType(N))
	{
	case NODE_CELL      :stype="CELL";break;
	case NODE_CELL_MINUS:stype="CELL_MINUS";break;
	case NODE_G         :stype="G";break;
	case NODE_OUT       :stype="OUT";break;
	case NODE_IN        :stype="IN";break;
	case NODE_OR        :stype="OR";break;
	case NODE_AND       :stype="AND";break;
	case NODE_DIFF      :stype="DIFF";break;
	case NODE_XOR       :stype="XOR";break;
	}

	fprintf(file,"type(%s) ",stype);

}




void GBool::print(char* spaces,FILE* file)
{
	int L;

	fprintf(file,"%s-------------------------------------------------\n",spaces);
	for ( L=255;L>=0;L--)
	{
		int cont=0;
		for (graph::list_iterator it=maing.each(L);!it.end();it++)
		{
			key N=*it;
			print(N,spaces,file);
			fprintf(file,")\n");
			++cont;
		}
		if (cont) fprintf(file,"\n");
	}
	fprintf(file,"%s-------------------------------------------------\n",spaces);
}



/////////////////////////////////////////////////////////////////////////////////
//INLINES
/////////////////////////////////////////////////////////////////////////////////

inline key& GBool::nodeData(key Node)
	{return maing.nodeData(Node);}

inline uint8& GBool::nodeType(key Node)
	{return maing.getSign(Node);}

inline void GBool::setNodeIn(key Node,key NodeData)
{
	assert(NodeData);

	while (maing.getNDw(Node)) 
		remNode(maing.getFirstDwNode(Node));

	nodeType(Node)=NODE_IN;
	nodeData(Node)=NodeData;
}

inline void GBool::setNodeOut(key Node)
{
	while (maing.getNDw(Node)) 
		remNode(maing.getFirstDwNode(Node));
	
	nodeType(Node)=NODE_OUT;
}


/////////////////////////////////////////////////////////////////////////////////
GBool::GBool(int pointdim,real Tolerance,bool use_internal_faces,bool use_partition_planes,int MaxNumTry,int MinNumFace,int UseOrthoForDeep)
	:pb(pointdim),tensor(pointdim),bsp(pointdim),args(sizeof(Info))
{
	checkpoint(pointdim>=1);

	this->pointdim=pointdim;
	this->Tolerance=Tolerance;
	this->MaxNumTry=MaxNumTry;
	this->MinNumFace=MinNumFace;
	this->UseOrthoForDeep=UseOrthoForDeep;
	this->use_internal_faces=use_internal_faces;
	this->use_partition_planes=use_partition_planes;
}

GBool::~GBool()
{

}

/////////////////////////////////////////////////////////////////////////////////

void GBool::useInternalFaces(bool value)
{
	this->use_internal_faces=value;
}

/////////////////////////////////////////////////////////////////////////////////
void GBool::usePartitionPlanes(bool value)
{
	this->use_partition_planes=value;
}

/////////////////////////////////////////////////////////////////////////////////
graph* GBool::getGraph()
{
	return &maing;
}

/////////////////////////////////////////////////////////////////////////////////
void GBool::setRoot(key Node)
{
	key I=args.block();
	int numfacetodo=countFaceTodo(Node);
	*((Info*)args[I])=Info(Node,numfacetodo,true,0);
	setMore(getRoot(),I);
}


/////////////////////////////////////////////////////////////////////////////////
void GBool::addBalls(key Node)
{	
	if (nodeType(Node)==NODE_CELL)
	{
		ggraph* g=(ggraph*)nodeData(maing.getFirstUpNode(Node));
		key Cell=nodeData(Node);
		real* bball=g->boundingBall(Cell);
		pb.addBBall(bball);
		tensor.addBall((real)sqrt(bball[0]),&bball[1]);
		return;
	}
	
	for (graph::iterator it=maing.goDw(Node);!it.end();it++)
		addBalls(*it);
}


/////////////////////////////////////////////////////////////////////////////////
key GBool::add(ggraph* g,uint8 nodetype)
{
	checkpoint(g->pointdim>=1);
	
	key Bop=maing.addNode(2);
	nodeType(Bop)=nodetype;

	if (g->pointdim==this->pointdim)
	{
		ggraph::list_iterator it;
		
		//reset the information about if a cell has been reached
		for (it=g->each(g->pointdim);!it.end();it++)
			g->nodeTmp(*it)=0;//not visited
		
		//reset the information about use of a face as splitting plane
		for (ggraph::list_iterator IT=g->each(g->pointdim);!IT.end();IT++)
		{
			for (ggraph::iterator JT=g->goDw(*IT);!JT.end();JT++)
				g->archData(JT.getArch())=0; // do not use
		}

		for (it=g->each(g->pointdim);!it.end();it++)
		{
			if (!g->nodeTmp(*it))
			{
				key Dst=maing.addNode(1); //start a new connected component from here
				nodeType(Dst)=NODE_G;
				nodeData(Dst)=(key)g;
				maing.addArch(Dst,Bop);

				addConnectedCells(Dst,g,*it);
			}
		}
	}
	
	return Bop;
}

/////////////////////////////////////////////////////////////////////////////////
key GBool::Intersection(int lvl)			
{
	key ret=maing.addNode(lvl);
	nodeType(ret)=NODE_AND	;
	return ret;
}

/////////////////////////////////////////////////////////////////////////////////
key GBool::Union(int lvl)			
{
	key ret=maing.addNode(lvl);
	nodeType(ret)=NODE_OR;
	return ret;
}

/////////////////////////////////////////////////////////////////////////////////
key GBool::Difference(int lvl)
{
	key ret=maing.addNode(lvl);
	nodeType(ret)=NODE_DIFF;
	return ret;
}

/////////////////////////////////////////////////////////////////////////////////
key GBool::Xor(int lvl)			
{
	key ret=maing.addNode(lvl);
	nodeType(ret)=NODE_XOR ;
	return ret;
}


/////////////////////////////////////////////////////////////////////////////////
key GBool::BoolOp(int code,int lvl) 
{
	switch(code)
	{
	case bsp_bool::BOOL_AND	 :return Intersection(lvl);
	case bsp_bool::BOOL_OR	 :return Union(lvl);
	case bsp_bool::BOOL_DIFF :return Difference(lvl);
	case bsp_bool::BOOL_XOR  :return Xor(lvl);
	}
	return 0;
}

/////////////////////////////////////////////////////////////////////////////////
key GBool::addArg(key Src,key Dst)	
{
	return maing.addArch(Src,Dst);
}


/////////////////////////////////////////////////////////////////////////////////
void GBool::addConnectedCells(key Dst,ggraph* g,key Cell)
{
	//already associated to a component
	if  (g->nodeTmp(Cell)) 
		return;
	
	g->nodeTmp(Cell)=1; 
	
	key Node=maing.addNode(0);
	nodeType(Node)=NODE_CELL;
	nodeData(Node)=Cell;
	
	maing.addArch(Node,Dst);
	
	for (ggraph::iterator it=g->goDw(Cell);!it.end();it++)
	{
		for (ggraph::iterator jt=g->goUp(*it);!jt.end();jt++)
			addConnectedCells(Dst,g,*jt);
	}
}


/////////////////////////////////////////////////////////////////////////////////
int GBool::countFaceTodo(key Node)
{
	int facetodo=0;
	
	if (nodeType(Node)==NODE_CELL)
	{
		ggraph* g=(ggraph*)nodeData(maing.getFirstUpNode(Node));
		key Cell=nodeData(Node);
		
		for (ggraph::iterator it=g->goDw(Cell);!it.end();it++)
		{
			key Face=*it;

			//if face is not used
			if (!g->archData(it.getArch()) && (use_internal_faces || g->getNUp(Face)!=2)) 
			{
				g->archData(it.getArch())=1; 
				++facetodo;					
			}
		}
		return facetodo;
	}
	
	for (graph::iterator it=maing.goDw(Node);!it.end();it++)
		facetodo+=countFaceTodo(*it);
	
	return facetodo;
}



/////////////////////////////////////////////////////////////////////////////////
void GBool::remNode(key Node)
{
	if (maing.getNUp(Node))
		maing.remArch(maing.getFirstUpArch(Node));
	
	for (key A;A=maing.getFirstDwArch(Node);) 
		remNode(maing.getN0(A));
	
	maing.remNode(Node);
}


/////////////////////////////////////////////////////////////////////////////////
bool GBool::findBoundaryCell(key Node,real* hsplit)
{
	static key _nav[MAX_CELLS_PER_LEVEL];

	//base case
	if (nodeType(Node)==NODE_CELL)
	{
		ggraph* g=(ggraph*)nodeData(maing.getFirstUpNode(Node));
		key Cell=nodeData(Node);

		if (pointdim==1)
		{
			key A1=g->getFirstDwArch(Cell),A2=g->getLastDwArch(Cell);

			if (g->archData(A1) || g->archData(A2))
			{
				if (g->archData(A2) && !g->archData(A1)) Swap(key,A1,A2);
				checkpoint(g->archData(A1));

				g->archData(A1)=0; //do not use anymore

				key V1=g->getN0(A1);real* V1g=g->getGeom(V1);
				key V2=g->getN0(A2);real* V2g=g->getGeom(V2);
				
				nodeType(Node)=NODE_CELL_MINUS; 

				real sign=(V1g[1]<=V2g[1])?(+1.0f):(-1.0f);
				hsplit[0]=sign*V1g[1];
				hsplit[1]=sign*-1;
				return true;
			}
		}

		else
		{
			for (ggraph::iterator it=g->goDw(Cell);!it.end();it++)
			{
				key TMP=*it;
				//a boundary face still to use
				if (g->archData(it.getArch())) 
				{
					//do not use anymore
					g->archData(it.getArch())=0; 

					//force (symbolically) in the below space! see split function
					nodeType(Node)=NODE_CELL_MINUS; 

					//robust find hyperplane
					// there was a bug using bounding ball because the hyplerplane of a face <F> (of a very small cell <C>)
					// was wrongly classified by h*BBall and caused a wrong reversion of the plane.
					g->robustFaceOutside(hsplit,Cell,TMP);
					return true;
				}
			}
		}
		return false;
	}
	
	//iterative case
	for (graph::iterator it=maing.goDw(Node);!it.end();it++) 
	{
		if (findBoundaryCell(*it,hsplit)) 
			return true;
	}
	
	return false;
}



/////////////////////////////////////////////////////////////////////////////////
void GBool::more(key N,key I)
{
	static int CONT=0;
	CONT++;
	key Below,Above;
	key NodeMinus,NodePlus;
	int facetodominus,facetodoplus;
	real hdir[MAX_GEOMETRIC_DIMENSION+1];
	real hsplit[MAX_GEOMETRIC_DIMENSION+1];
	
	Info info=*((Info*)args[I]);args.release(I);
	
	key  Node				= info.Node;
	int  facetodo			= info.facetodo;
	bool bUsePartition	= use_partition_planes && info.usepartition;
	int  deep            = info.deep;
	
	if (nodeType(Node)==NODE_OUT)
	{
		remNode(Node);
		setOut(N);
		return;
	}
	
	if (nodeType(Node)==NODE_IN)
	{
		key C=nodeData(Node);
		assert(C);
		nodeInfo(N)=C;
		remNode(Node);
		setIn(N);
		return;
	}
	
	bool bFoundPlane=false;
	
	//try to use a separating plane (and not a boundary face of the g-objects)
	if (pointdim>=2 && bUsePartition && facetodo>=MinNumFace)
	{
		bool bUseOrtho=(UseOrthoForDeep) && (deep<=UseOrthoForDeep); 
		
		pb.reset();
		tensor.reset();
		
		//add all the balls of the node 
		addBalls(Node);
		
		//find inertia vectors
		if (!bUseOrtho) 
			tensor.solve();
		
		for (int i=0;i<pointdim;i++)
		{
			memset(hdir,0,sizeof(real)*(pointdim+1));
			
			if (bUseOrtho)
				hdir[(i+1)]=1;
			else
				memcpy(&hdir[1],tensor.getEigenVector(i),sizeof(real)*(pointdim));
			
			pb.addDirection(hdir);
		}
		
		bFoundPlane=(pb.getScore()>0);
		
		if (bFoundPlane) 
			memcpy(hsplit,pb.getSolution(),sizeof(real)*(pointdim+1));
		
	}
	
	//use faces
	if (!bFoundPlane) 
	{
		bUsePartition=false;
		bFoundPlane=findBoundaryCell(Node,hsplit);
	}
	
	checkpoint(bFoundPlane);
	
	//set the BSP and split the current node of the maing
	setCutn(N,hsplit,Below=in(),Above=out());

	split(hsplit,Node,NodeMinus,NodePlus,facetodominus,facetodoplus);

	//below and above space
	key Iminus=args.block();setMore(Below,Iminus);
	key Iplus =args.block();setMore(Above,Iplus );
	*((Info*)args[Iminus])=Info(NodeMinus,facetodominus,bUsePartition && facetodominus<facetodo,deep+1);
	*((Info*)args[Iplus ])=Info(NodePlus ,facetodoplus ,bUsePartition && facetodoplus <facetodo,deep+1);
	

//	maing.print();

	//anticipate the simplification of final nodes (in order to deallocate as soon as possible)
	if (nodeType(NodeMinus)==NODE_IN || nodeType(NodeMinus)==NODE_OUT) visit(Below);
	if (nodeType(NodePlus )==NODE_IN || nodeType(NodePlus )==NODE_OUT) visit(Above);

//	maing.print();
}


/////////////////////////////////////////////////////////////////////////////////
void GBool::simplify(key Node,int facetodo)
{
	uint8& Type=nodeType(Node);
	
	//already final
	if (Type==NODE_CELL || Type==NODE_OUT || Type==NODE_IN)
		return;
	
	//no down-arguments!
	if (!maing.getNDw(Node))
	{
		setNodeOut(Node);
		return;
	}
	
	//special case for connected component
	if (Type==NODE_G)
	{
		if (!facetodo) 
		{
			static int nactivation=0;
			nactivation++;
			setNodeIn(Node,nodeData(Node)); //all full
		}
		return;
	}
	
	//simplification
	key first_in_data=0;

	int cont=0,nin=0,nout=0;
	for (graph::iterator it=maing.goDw(Node),itnext;!it.end();it=itnext,cont++)
	{
		itnext=it.next();
		
		key NodeDw=*it;
		uint8& TypeDown=nodeType(NodeDw);

		if (TypeDown==NODE_IN && !first_in_data)
			first_in_data=nodeData(NodeDw);

		//final case
		if (Type==NODE_OR   && TypeDown==NODE_IN)  // OR (...IN... )	-> IN
		{
			setNodeIn(Node,nodeData(NodeDw));
			return;
		}
		
		//final case
		if ((Type==NODE_AND  && TypeDown==NODE_OUT          ) || // AND(...OUT.. )	-> OUT
			 (Type==NODE_DIFF && TypeDown==NODE_OUT && !cont ) || // DIFF(OUT,... )	-> OUT
			 (Type==NODE_DIFF && TypeDown==NODE_IN  &&  cont ))   // DIFF(f,.,IN..)	-> OUT
		{
			setNodeOut(Node);
			return;
		}
		
		//simplification
		if ((Type==NODE_OR   && TypeDown==NODE_OUT         ) || // OR  (...OUT...)		-> OR(...)
			 (Type==NODE_AND  && TypeDown==NODE_IN          ) || // AND (...IN....)		-> IN(...)
			 (Type==NODE_XOR  && TypeDown==NODE_OUT         ) || // XOR (...OUT...)		-> XOR(...)
			 (Type==NODE_DIFF && TypeDown==NODE_OUT && cont ))   // DIFF(f,.OUT...	)	-> DIFF(f,...)
		{
			remNode(NodeDw);
			continue;
		}
		
		if (TypeDown==NODE_IN) 
		{
			++nin; 
		}
		else if (TypeDown==NODE_OUT) 
			++nout; 
	}
	
	if ((Type==NODE_OR  && !maing.getNDw(Node)) ||	//OR(<empty>)					-> OUT
	  	 (Type==NODE_XOR && !maing.getNDw(Node)) ||	//XOR(<empty>)					-> OUT
		 (Type==NODE_XOR && nin>=2))						//XOR (...,IN,...,IN...)	-> OUT
	{
		setNodeOut(Node);
		return;
	}

	if (Type==NODE_AND  && !maing.getNDw(Node)) // AND(<empty>) -> IN
	{
		setNodeIn(Node,first_in_data);
		return;
	}
	
	if ((Type==NODE_DIFF && nin==1 && maing.getNDw(Node)==1) ||		// DIFF(IN)		-> IN
		 (Type==NODE_XOR  && nin==1 && maing.getNDw(Node)==1))		// XOR(IN)		-> IN
	{
		setNodeIn(Node,nodeData(maing.getFirstDwNode(Node)));
		return;
	}
}

/////////////////////////////////////////////////////////////////////////////////
void GBool::split(ggraph* g,real* h,key Cell,key& Cminus,key& Cplus,key& Cequal)
{
	int numtry=0;	
	int retcode=g->split(Cell,h,Tolerance,Cminus,Cplus,Cequal);

	
	if (retcode==ggraph::IMPOSSIBLE)
	{
		do
		{
			printf(" perturbation ");fflush(stdout);
			real perplane[MAX_GEOMETRIC_DIMENSION+1];
			
			for (int k=0;k<=g->pointdim;k++) 
				perplane[k]=(real)(h[k]+numtry*Tolerance*(rand()/(real)RAND_MAX-0.5));
			
			retcode=g->split(Cell,perplane,Tolerance,Cminus,Cplus,Cequal);
			numtry++;
		}
		while (retcode==ggraph::IMPOSSIBLE && numtry<MaxNumTry);
	}
	
	//if flat assign to above. Not sure it is correct (but I always can perturbate the plane...)
	if (retcode==ggraph::FLAT) 
	{
		Cminus=Cequal=0;
		Cplus =Cell;
		retcode=ggraph::OK;
	}

	checkpoint(retcode==ggraph::OK);
}


/////////////////////////////////////////////////////////////////////////////////
void GBool::split(real* h,key Node,key& NodeMinus,key& NodePlus,int& facetodominus,int& facetodoplus)
{
	NodeMinus=0;facetodominus=0;
	NodePlus =0;facetodoplus =0;
	
	if (nodeType(Node)==NODE_CELL || nodeType(Node)==NODE_CELL_MINUS)
	{
		key Cell		= nodeData(Node);
		key UpNode	= maing.getFirstUpNode(Node);
		ggraph* g	= (ggraph*)nodeData(UpNode);
		
		key Cminus=0,Cplus=0,Cequal=0;
		
		if (nodeType(Node)==NODE_CELL_MINUS) //force cell in the below space
			Cminus=Cell;
		else
			split(g,h,Cell,Cminus,Cplus,Cequal);

		if (Cequal)
		{
			checkpoint(g->getNUp(Cequal)==2);
			g->archData(g->getFirstUpArch(Cequal))=0; //do not use this face
			g->archData(g->getLastUpArch (Cequal))=0; 
		}
		
		if (Cminus)
		{
			NodeMinus=maing.addNode(0);
			nodeData(NodeMinus)=Cminus;
			nodeType(NodeMinus)=NODE_CELL;
			
			for (ggraph::iterator it=g->goDw(Cminus);!it.end();it++)
				if (g->archData(it.getArch())) ++facetodominus;
		}
		
		if (Cplus)
		{
			NodePlus =maing.addNode(0);
			nodeData(NodePlus )=Cplus ;
			nodeType(NodePlus )=NODE_CELL;
			
			for (ggraph::iterator it=g->goDw(Cplus);!it.end();it++)
				if (g->archData(it.getArch())) ++facetodoplus;
		}
		
		remNode(Node);
		return;
	}
	
	//split the current node
	NodeMinus=maing.addNode(maing.getLevel(Node));nodeData(NodeMinus)=nodeData(Node);nodeType(NodeMinus)=nodeType(Node);
	NodePlus =maing.addNode(maing.getLevel(Node));nodeData(NodePlus )=nodeData(Node);nodeType(NodePlus )=nodeType(Node);
	
	while (maing.getNDw(Node))
	{
		key NodeDw=maing.getFirstDwNode(Node);
		key NodeDwMinus,NodeDwPlus;
		int Dwfacetodominus,Dwfacetodoplus;
		split(h,NodeDw,NodeDwMinus,NodeDwPlus,Dwfacetodominus,Dwfacetodoplus);
		
		if (NodeDwMinus) 
		{
			maing.addArch(NodeDwMinus,NodeMinus);
			facetodominus+=Dwfacetodominus;
		}
		
		if (NodeDwPlus ) 
		{
			maing.addArch(NodeDwPlus ,NodePlus );
			facetodoplus+=Dwfacetodoplus;
		} 
	}
	
	remNode(Node);
	
	simplify(NodeMinus ,facetodominus);
	simplify(NodePlus  ,facetodoplus );	
}

