/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#ifndef _G_BOOL_H__
#define _G_BOOL_H__

#include "generaldef.h"

#include "ggraph.h"
#include "bsp.h"
#include "tensor.h"
#include "partball.h"

class GBool: public bsp
{
protected:

	// the space where the boolean operation will be performed
	int pointdim;

	//graph to store the boolean operations 
	graph maing;

	//tolerance in the splitting of cells
	real Tolerance;

	//try to split a cell a <maxnumtry>
	int  MaxNumTry;
	
	//use cell-boundary-faces (and not generic separating planes) when a Bsp-Cell contains a numberoffaces <= <minnumface>
	int MinNumFace;

	//use orthogonal planes for the first <UseOrthoForDeep> levels
	int UseOrthoForDeep;

	//to use or not internal faces
	bool use_internal_faces;
	bool use_partition_planes;

	//to perform separating axes method
	PartitionBBall pb;
	InertiaTensor tensor;

	typedef struct Info
	{
		key   Node;
		int	facetodo;
		int   deep;
		bool  usepartition;

		//constructor
		Info(key _Node,int _facetodo,bool _usepartition,int _deep)
			:Node(_Node),facetodo(_facetodo),usepartition(_usepartition),deep(_deep){}
	}
	Info;
	
	memcache args;

	//add all pointdim-cells to <DST>. the cells forms a connected components
	void addConnectedCells(key Dst,ggraph* g,key Cell);

	//count the number of boundary faces in a node (recursive way)
	int countFaceTodo(key Node);

	//transform a node (ex or,and,diff etc) to a full IN node (empty OUT node)
	void setNodeIn (key Node,key NodeData);
	void setNodeOut(key Node);

	// remove a particular node by first (a) removing up connection
	// (b) recursively removing down nodes/arches (c) removing phisically
	// the node from the graph
	void remNode(key Node);

	// find the first cell which has a boundary face (to use for splitting)
	// if found return true. The cell which has been used (==owns the face)
	// is "marked" for a symbolic split (==won't be splitted since it is 
	// in the below subspace)
	bool findBoundaryCell(key Node,real* hsplit);

	// build dinamically the bsp
	virtual void more(key N,key I);

	//add all the balls of a node (recursive way)
	void addBalls(key Node);

	//run-time simplification of a maing node
	void simplify(key Node,int facetodo);

	//split a point-dim cell
	void split(ggraph* g,real* h,key Cell,key& Cminus,key& Cplus,key& Cequal);

	//split a maing node
	void split(real* h,key Node,key& NodeMinus,key& NodePlus,int& facetodominus,int& facetodoplus);



public:

	//constructor
	GBool(int pointdim,real Tolerance=1e-4,bool use_internal_faces=false,bool use_partition_planes=false,
		int MaxNumTry=100,int MinNumFace=20,int UseOrthoForDeep=3);

	//destructor
	~GBool();

	//use or not internal faces
	void useInternalFaces(bool value);

	//use or not partition planes
	void usePartitionPlanes(bool value);

	//function to call when the maing has been built
	void setRoot(key Node);

	//create a new node at a specific level
	key Intersection	(int lvl=3)	;
	key Union			(int lvl=3)	;
	key Difference		(int lvl=3)	;
	key Xor				(int lvl=3)	;
	key BoolOp        (int code,int lvl=3);

	//add an argument (subnode) to a node
	key addArg(key Src,key Dst);

	//add all the connected components of a ggraph
	key add(ggraph* g,uint8 nodetype=NODE_OR);

	//access the boolean graph
	graph* getGraph();

	//information stored in the maing graph (ex. number of cell for Level=0 and ggraph* for Level=1)
	key& nodeData(key Node);

	//the type of the node of the maing (see enum)
	uint8& nodeType(key Node);

	//type of nodes in the maing graph
	enum 
	{
		NODE_CELL,NODE_CELL_MINUS,NODE_G,
		NODE_OUT,NODE_IN,
		NODE_OR,NODE_AND,NODE_DIFF,NODE_XOR
	};

	void print(key N,char* space,FILE* file);
	void print(char* spaces,FILE* file);


};



#endif //_G_BOOL_H__


