/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */

#include "generaldef.h"

#include <math.h>
#include <float.h>
#include <stdio.h>



#include "plane.h"
#include "ggraph.h"
#include "bsp.h"
#include "mat.h"
#include "gbool.h"


////////////////////////////////////////////////////////////////////////

void ggraph::print(char* space,FILE* file) 
{
	if (vmat) tmat<real>::print(spacedim+1,vmat,file);
	if (hmat) tmat<real>::print(spacedim+1,hmat,file);


	graph::print(space,file);
}


////////////////////////////////////////////////////////////////////////

void ggraph::remNode(key Node,bool recursive)
{
	if  (nodeData(Node))
		db.release(nodeData(Node));

	key A;

	for (A;A=getFirstUpArch(Node);) 
		remArch(A);
	
	for (A;A=getFirstDwArch(Node);) 
	{
		key Down=getN0(A);

		remArch(A);

		if (recursive && !getNUp(Down)) 
			remNode(Down);
	}

	graph::remNode(Node);
}


////////////////////////////////////////////////////////////////////////

key ggraph::addEdge(key N0,key N1)
{
	key E=addNode(1);
	addArch(N0,E);
	addArch(N1,E);
	return E;
}

key ggraph::addEdge(real* v0,real* v1)
{
	return addEdge(addVertex(v0),addVertex(v1));
}

////////////////////////////////////////////////////////////////////////

key ggraph::addFace(key n,const key* N)
{
	key i;
	key F=addNode(2);

	if (pointdim==2)
	{
		for (i=0;i<n;i++)
			addArch(F,N[i]);
	}

	for (i=0;i<n;i++)
	{	
		key p1=N[i      ];
		key p2=N[(i+1)%n];

		key E;

		if (!(E=findCommonNode(p1,p2,DIRECTION_UP)))
		{
			E=addNode(1);
			addArch(p1,E);
			addArch(p2,E);
		}

		addArch(E,F);
	}

	nodeData(F)=db.block();

	real* v0=(real*)getGeom(nodeData(N[0]));
	real* v1=(real*)getGeom(nodeData(N[1]));
	real* v2=(real*)getGeom(nodeData(N[2]));

	real x0 = v2[1] - v0[1]    ;real y0 = v2[2] - v0[2]   ;real z0 = v2[3] - v0[3];
	real x1 = v1[1] - v0[1]    ;real y1 = v1[2] - v0[2]   ;real z1 = v1[3] - v0[3];
	real x  = y0 * z1 - z0 * y1;real y = z0 * x1 - x0 * z1;real z = x0 * y1 - y0 * x1;
	real len = x*x + y*y + z*z;
	
	if (!len) 
		{x=y=z=0;}
	else 
		{real recip=1/(real)sqrt (len);x*=recip;y*=recip;z*=recip;}

	real* g=(real*)db[nodeData(F)];
	g[0]=+x*(v0[1])+y*v0[2]+z*v0[3];
	g[1]=-x;g[2]=-y;g[3]=-z;

	return F;
}


////////////////////////////////////////////////////////////////////////

key ggraph::addTet(key N0,key N1,key N2,key N3)
{
	key E01,E02,E03,E12,E13,E23;
	key F012,F023,F013,F123;
	key TET;

	TET=addNode(3);

#define findFace(F,N0,N1,N2) ( \
	 (E01 =findCommonNode(N0 ,N1 ,DIRECTION_UP)) && \
	 (E12 =findCommonNode(N1 ,N2 ,DIRECTION_UP)) && \
	 (F   =findCommonNode(E01,E12,DIRECTION_UP)))

	if (!findFace(F012,N0,N1,N2)) F012=addNode(2);
	if (!findFace(F023,N0,N2,N3)) F023=addNode(2);
	if (!findFace(F013,N0,N1,N3)) F013=addNode(2);
	if (!findFace(F123,N1,N2,N3)) F123=addNode(2);

	if (!(E01=findCommonNode(N0,N1,DIRECTION_UP))) E01=addNode(1);
	if (!(E02=findCommonNode(N0,N2,DIRECTION_UP))) E02=addNode(1);
	if (!(E03=findCommonNode(N0,N3,DIRECTION_UP))) E03=addNode(1);
	if (!(E12=findCommonNode(N1,N2,DIRECTION_UP))) E12=addNode(1);
	if (!(E13=findCommonNode(N1,N3,DIRECTION_UP))) E13=addNode(1);
	if (!(E23=findCommonNode(N2,N3,DIRECTION_UP))) E23=addNode(1);

	key links[]=
	{
		N0,E01,N1,E01,   N0,E02,N2,E02,   N0,E03,N3,E03,   N1,E12,N2,E12,   N1,E13,N3,E13,  N2,E23,N3,E23,
		E02,F012,E12,F012,E01,F012,   E01,F013,E13,F013,E03,F013,   E03,F023,E23,F023,E02,F023,   E12,F123,E23,F123,E13,F123,
		F012,TET,F023,TET,F013,TET,F123,TET,
		0
	};

	for (int i=0;links[i];i+=2)
	{
		if (!findArch(links[i],links[i+1]))
			addArch(links[i],links[i+1]);
	}

	key N[4]={N0,N1,N2,N3};
	addArch(TET,N[0]);
	addArch(TET,N[1]);
	addArch(TET,N[2]);
	addArch(TET,N[3]);

	return TET;
}


//////////////////////////////////////////////////////////////////////////
key ggraph::findCells(int level,key Cell,key* dst)
{
	//fast retrieve going up
	if (Cell && level==pointdim && getNUp(Cell))
	{
		for (ggraph::iterator it=goUp(Cell);!it.end();it++) *dst++=*it;
		return getNUp(Cell);
	}

	return graph::findCells(level,Cell,dst);
}

//////////////////////////////////////////////////////////////////////////

void ggraph::findCells(int level,key Cell,key* nnav,key** nav)
{
	//fast retrieve going up
	if (Cell && level==pointdim && getNUp(Cell))
	{
		key* dst=nav[0];
		for (ggraph::iterator it=goUp(Cell);!it.end();it++) *dst++=*it;
		nnav[0]=getNUp(Cell);
		return;
	}

	graph::findCells(level,Cell,nnav,nav);
}



/////////////////////////////////////////////////////////////////////////////////
void ggraph::fixPlaneOrientation()
{
	key* Points=new key[MAX_CELLS_PER_LEVEL];

	//no planes
	if (pointdim<2)
		return;

	int J;
	for (ggraph::list_iterator it=each(pointdim-1);!it.end();it++)
	{
		key Face=*it;

		//only for boundary face
		if (getNUp(Face)!=1) 
			continue; 
		
		if (!nodeData(Face)) 
		{
			key NPoints=findCells(0,Face,Points);
			findPlane(setGeom(Face),NPoints,Points,pointdim);
		}

		real* bball=boundingBall(getFirstUpNode(Face));

		if (bball[0]>0)
		{
			real* h=getGeom(Face);
			real acc=h[0];
			for (J=1;J<=pointdim;J++) acc+=h[J]*bball[J];
			if (acc>0) 
			{
				for (J=0;J<=pointdim;J++)  
					h[J]*=-1;
			}
		}
	}

	delete [] Points;
}

//////////////////////////////////////////////////////////////////////////

void ggraph::findPlane(real* h,key npoints,key* points,int pointdim,const real module_too_small)
{
	checkpoint(pointdim>=1 && this->pointdim>=pointdim);

	switch(pointdim)
	{
	case 1:
		{
			real* p0=getGeom(points[0]);
			h[0]=-p0[1];
			h[1]=1; //x>=value
			break;
		}

	case 2:
		{
			if (npoints<2) goto GENERIC_CASE;

			real* p0=getGeom(points[0]);
			real* p1=getGeom(points[1]);

			const real& x1=p0[1],y1=p0[2];
			const real& x2=p1[1],y2=p1[2];

			h[0]=-x1*y2+y1*x2; //(x-x1)/(x2-x1)=y-y1)/(y2-y1)
			h[1]=(y2-y1);
			h[2]=(x1-x2); 

			real module=(real)sqrt(h[1]*h[1]+h[2]*h[2]);
			if (module<module_too_small) goto GENERIC_CASE;
			break;
		}
	case 3:
		{
			if (npoints<3) goto GENERIC_CASE;

			real* p0=getGeom(points[0]);
			real* p1=getGeom(points[1]);
			real* p2=getGeom(points[2]);

			const real& x1=p0[1],y1=p0[2],z1=p0[3]; //cross product i j k | x1 y1 z1 | x2 y2 z2
			const real& x2=p1[1],y2=p1[2],z2=p1[3];
			const real& x3=p2[1],y3=p2[2],z3=p2[3];

			h[0] = -x1*(y2*z3 - y3*z2) -x2*(y3*z1 - y1*z3) -x3*(y1*z2 - y2*z1);
			h[1] = y1*(z2 - z3) + y2*(z3 - z1) + y3*(z1 - z2) ;
			h[2] = z1*(x2 - x3) + z2*(x3 - x1) + z3*(x1 - x2) ;
			h[3] = x1*(y2 - y3) + x2*(y3 - y1) + x3*(y1 - y2) ;

			real module=(real)sqrt(h[1]*h[1]+h[2]*h[2]+h[3]*h[3]);
			if (module<module_too_small) goto GENERIC_CASE;
			break;
		}
	default:
		{
GENERIC_CASE:

			const int N=pointdim+1;
			const int M=max2((int)npoints,N); //must be M>=N, otherwise GSL doesn't support it

			gsl_matrix *A=gsl_matrix_alloc(M,N),*V=gsl_matrix_alloc(N,N);
			gsl_vector *S=gsl_vector_alloc(N)  ,*W=gsl_vector_alloc(N);;

			for (int I=0;I<M;I++) 
			{
				const real* p=getGeom(points[I % npoints]);
				for (int J=0;J<N;J++) gsl_matrix_set(A,I,J,p[J]);
			}

			//svd decomposition
			gsl_linalg_SV_decomp(A,V,S,W);
			for (int J=0;J<N;J++) h[J]=(real)gsl_matrix_get(V,J,N-1);

			gsl_matrix_free(A);gsl_matrix_free(V);
			gsl_vector_free(S);gsl_vector_free(W);
			break;
		}
	}

	//plane where the points belong and translation matrix
	normalize_plane<real>(pointdim,h);

#if 0
#ifdef _DEBUG
	for (int i=0;i<(int)npoints;i++)
	{
		real* vg=getGeom(points[i]),acc=0;
		for (int j=0;j<=pointdim;j++) acc+=h[j]*vg[j];
		checkpoint(fabs(acc)<1e-3);
	}
#endif
#endif
}


////////////////////////////////////////////////////////////////////////

void ggraph::print(key N,char* space,FILE* file)
{
	graph::print(N,space,file);

	if (nodeData(N))
	{
		real* g=getGeom(N);
		fprintf(file," geom(");
		for (int i=0;i<=pointdim;i++) fprintf(file,"%e ",(float)g[i]);
		fprintf(file,")");
	}
}


////////////////////////////////////////////////////////////////////////

void ggraph::printStatistics(FILE* file)
{
	int L;
	fprintf(file,"Ggraph(%d)------------------\n",(int)pointdim);
	fprintf(file,"ncells\t");

	for ( L=255;L>=0;L--)
	{
		for (list_iterator it=each(L);!it.end();it++)
		{
			fprintf(file,"%d(%d) ",(int)L,nodelists.n[L]);
			break;
		}
	}
	fprintf(file,"\n");

	size_t t=sizeof(ggraph)+sizeof(nodelists)+nodes.totalmem()+arches.totalmem()+db.totalmem();
	fprintf(file,"Total memory %dkb \n",(int) t/1024);

	fprintf(file,"Node ");
	nodes.printStatistics(file);

	fprintf(file,"Arch ");
	arches.printStatistics(file);

	fprintf(file,"Db ");
	db.printStatistics();
}


////////////////////////////////////////////////////////////////////////

real* ggraph::boundingBall(key C)
{
	iterator it;
	int lvl=getNode(C)->Level;
	checkpoint(lvl==pointdim);

	real* dst=0;

	if (!nodeData(C))
	{
		dst=setGeom(C);
		dst[0]=-1;//invalid radius
	}
	else
		dst=getGeom(C);

	//radius not valid
	if (dst[0]<0)
	{
		memset(dst,0,sizeof(real)*(pointdim+1));

		real coeff=(real)(1.0/getNUp(C));

		//compute the centroid and put it in [1...pointdim]
		for (it=goUp(C);!it.end();it++)
		{
			real* gtmp=getGeom(*it);

			for (int n=1;n<=pointdim;n++)
				dst[n]+=coeff*gtmp[n];
		}

		//compute the max distance for bounding ball radius
		//compute the centroid and put it in [1...pointdim]

		real maxradius=0;
		for (it=goUp(C);!it.end();it++)
		{
			real* gtmp=getGeom(*it);

			real radius=0;
			for (int n=1;n<=pointdim;n++)
				radius+=(dst[n]-gtmp[n])*(dst[n]-gtmp[n]);

			if (radius>maxradius)
				maxradius=radius;
		}

		dst[0]=maxradius;
	}

	return dst;
}



//--------------------------------------------------------------------
void ggraph::robustFaceOutside(real* hsplit,key C,key F)
{
	//not thread safe
	static key _nav[MAX_CELLS_PER_LEVEL];

	checkpoint(pointdim>=2 && getLevel(C)==(pointdim) && getLevel(F)==(pointdim-1));

	key NPoints=findCells(0,F,_nav);

	//I need the plane, if it does not exists find it!
	if (!nodeData(F))
		findPlane(setGeom(F),NPoints,_nav,pointdim);

	//verify if the face of the cells points out the centroid of the ball
	memcpy(hsplit,getGeom(F),sizeof(real)*(pointdim+1));

	//find points of face <F> for which the plane is maximized
	real maxdistance=0;
	for (key I=0;I<NPoints;I++)
	{
		real* vg=getGeom(_nav[I]);
		real acc=0;
		for (int n=1;n<=pointdim;n++) acc+=hsplit[n]*vg[n];
		if (fabs(acc)>fabs(maxdistance)) maxdistance=acc;
	}

	real hsplit_0_old=hsplit[0];
	hsplit[0]=-maxdistance;

	//now, with the current orientation, find the farthest point
	key  V=0;
	maxdistance=0;

	for (iterator it=goUp(C);!it.end();it++)
	{
		key _V=*it;
		real* gtmp=getGeom(_V);

		real acc=0;
		for (int n=0;n<=pointdim;n++) 
			acc+=hsplit[n]*gtmp[n];

		if (fabs(acc)>fabs(maxdistance))
		{
			maxdistance=acc;
			V=*it;
		}
	}

	checkpoint(fabs(maxdistance)>0);

#if 1
	//don't want the perturbation to affect the original plane
	//seems more robust this way
	hsplit[0]=hsplit_0_old;
#endif

	//need to reverse
	if (maxdistance>0)
	{
		for (int i=0;i<=pointdim;i++)
			hsplit[i]*=-1;
	}
}




//////////////////////////////////////////////////////////////////////////

ggraph& ggraph::embed(int new_pointdim)
{
	if (new_pointdim<=pointdim)
		return *this;

	list_iterator it;

	//remove old planes and eventually bounding balls
	if (pointdim>=2)
	{
		for (it=each(pointdim);!it.end();it++)
		{
			key& Info=nodeData(*it);
			if (Info) db.release(Info);
			Info=0;
		}

		for (it=each(pointdim-1);!it.end();it++)
		{
			key& Info=nodeData(*it);
			if (Info) db.release(Info);
			Info=0;
		}
	}

	//remove links from pointdim-cells to vertices
	for (it=each(pointdim);!it.end();it++)
	{
		key A;
		while (A=getFirstUpArch(*it))
			remArch(A);
	}

	this->pointdim=new_pointdim;
	db.resize(sizeof(real)*(this->pointdim+1));
	
	//set up new planes due to the embedding (xn>=0)
	if (pointdim>=2)
	{
		for (it=each(pointdim-1);!it.end();it++)
		{
			real* gtmp=setGeom(*it);
			gtmp[0]=0;
			gtmp[pointdim]=1; //xn>=0
		}
	}

	return *this;
}


//////////////////////////////////////////////////////////////////////////////////////////////


ggraph* ggraph::embed(real From,real To)
{
	ggraph* dcom=new ggraph(this->pointdim+1);

	if (From>To) 
		Swap(real,From,To);

	graph::list_iterator i;

	//find all zero-cells
	key **nav,*nnav=getNavigation(nav);

	findCells(0,0,nnav,nav); 

	int u;
	for (u=0;u<this->pointdim;++u)
		next(true,nnav[u+1],nav[u+1],nnav[u],nav[u]);

	key* fullcells=nav[pointdim];
	const key& nfullcells=nnav[pointdim];

	for (u=0;u<(int)nfullcells;u++)
	{
		key scell=fullcells[u];
		key NewUpCell=dcom->addNode(dcom->pointdim); 
		this->nodeTmp(scell)=NewUpCell; 
	}

	for (int k=0;k<=this->pointdim;++k)
	{
		key* kcells=nav[k];
		const key& nkcells=nnav[k];

		for (key i=0;i<nkcells;++i)
		{
			key scell = kcells[i];
			key dcell1=dcom->addNode(k);
			key dcell2=dcom->addNode(k);
			key dlinkcell=(k==pointdim)? this->nodeTmp(scell) : dcom->addNode(k+1);

			/* link cell->vertices */

			if (!k)
			{
				if (!this->pointdim)
				{
					dcom->addArch(dlinkcell,dcell1);
					dcom->addArch(dlinkcell,dcell2);
				}
				else
				{
					for (ggraph::iterator u=goDw(scell);!u.end();u++)
					{
						key _scell=*u;
						key NewUpCell=this->nodeTmp(_scell);
						dcom->addArch(NewUpCell,dcell1);
						dcom->addArch(NewUpCell,dcell2);
					}
				}
			}

			/* save map  scell->dcell1->dcell2->dlinkcell */

			this->nodeTmp(scell )=dcell1;
			dcom->nodeTmp(dcell1)=dcell2;
			dcom->nodeTmp(dcell2)=dlinkcell;

			/*  geometry  */
			
			if (!k) /* vertices */
			{	
				real* g1=dcom->setGeom(dcell1,this,scell);g1[this->pointdim+1]=From;
				real* g2=dcom->setGeom(dcell2,this,scell);g2[this->pointdim+1]=To;
			}
			else if (dcom->pointdim>=2 && (k)==(dcom->pointdim-1)) /* dcell1, dcell2 */
			{
				real* g1=dcom->setGeom(dcell1);g1[0]=+From;g1[dcom->pointdim]=-1; /* xn<=from */
				real* g2=dcom->setGeom(dcell2);g2[0]=  -To;g2[dcom->pointdim]=+1; /* xn>=to */		
			}
			
			/* linked cell. embed original plane */
			if (dcom->pointdim>=2 && (k+1)==(dcom->pointdim-1) && nodeData(scell)) 
			{
				/* scell is a vertex */
				if (pointdim==1) 
				{
					real* s1= getGeom(scell);
					real  sign=(real)((getNUp(scell)==2) || (getGeom(getNextUpBrother(getFirstUpArch(scell)))[1]<s1[1])?+1:-1);

					real* g1=dcom->setGeom(dlinkcell);
					g1[0]=-sign*s1[1];
					g1[1]=+sign;	
					g1[2]=0;
				}

				/* scell is a plane */
				else 
				{
					real* g1=dcom->setGeom(dlinkcell,this,scell);
					g1[dcom->pointdim]=0;
				}
			}

			/* connectivity */

			key Before=0;
			dcom->addArch(dcell1,dlinkcell,DIRECTION_UP_AND_DOWN,0,Before); //mantain order for 2-faces

			if (k)
			{
				for (graph::iterator i=this->goDw(scell);!i.end();i++)
				{
					key ssubface = (*i);

					key dc1      = this->nodeTmp(ssubface);
					key dc2      = dcom->nodeTmp(dc1);
					key cnew     = dcom->nodeTmp(dc2);

					dcom->addArch(dc1,dcell1);
					dcom->addArch(dc2,dcell2);
					Before=dcom->addArch(cnew,dlinkcell);
				}
			}

			dcom->addArch(dcell2,dlinkcell,DIRECTION_UP_AND_DOWN,0,Before); //mantain order for 2-faces
		}
	}


	return dcom;
}


///////////////////////////////////////////////////////////////////

ggraph* ggraph::cuboid(int pointdim)
{
	static ggraph** cached=0;

	if (!cached)
	{
		cached=new ggraph*[MAX_GEOMETRIC_DIMENSION];
		memset(cached,0,sizeof(ggraph*)*MAX_GEOMETRIC_DIMENSION);
	}

	if (cached[pointdim])
		return cached[pointdim]->clone();


	ggraph* ret=0;

	if (!pointdim)
	{
		ret=new ggraph(pointdim);
		ret->setGeom(ret->addNode(0));
	}
	else
	{
		ggraph* cub=cuboid(pointdim-1);
		ret=cub->embed(0,1);
		delete cub;
	}

	cached[pointdim]=ret->clone();
	
	return ret;

}


ggraph* ggraph::simplex(int pointdim)
{
	static ggraph** cached=0;

	if (!cached)
	{
		cached=new ggraph*[MAX_GEOMETRIC_DIMENSION];
		memset(cached,0,sizeof(ggraph*)*MAX_GEOMETRIC_DIMENSION);
	}

	if (cached[pointdim])
		return (cached[pointdim]->clone());


	ggraph* ret=0;

	if (!pointdim)
	{
		ret=new ggraph(0);
		ret->setGeom(ret->addNode(0));
	}
	else
	{
		ggraph* tmp=simplex(pointdim-1);
		key N=*(tmp->each(pointdim-1));
		ret=tmp->join(N,1);
		delete tmp;
	}

	cached[pointdim]=ret->clone();

	return ret;
}



///////////////////////////////////////////////////////////////////

ggraph* ggraph::grid(int pointdim,uint8* v)
{
	bsp_grid b(pointdim,v); 
	bsp_solver	s(&b);
	s.setSimplify(false);
	s.visit();
	return s.getGraph()->clone();
}


//////////////////////////////////////////////////////////////////////////

void ggraph::check(key N)
{
	iterator i,j;
	graph::check(N);

	int lvl=getLevel(N);

	if (lvl>0)
	{
		checkpoint(getNDw(N)>getLevel(N));
	}

	if (lvl==1)
	{
		checkpoint(getNDw(N)==2);
	}

	if (lvl==2)
	{
		key prev0=getFirstDwNode(getLastDwNode(N)),p1;
		key prev1=getLastDwNode (getLastDwNode(N)),p2;

		for (i=goDw(N);!i.end();i++,prev0=p1,prev1=p2)
		{
			p1=getFirstDwNode(*i);
			p2=getLastDwNode (*i);

			if (p1!=prev0 && p1!=prev1) 
				Swap(key,p1,p2);
			
			checkpoint(p1==prev0 || p1==prev1) ;

			if (pointdim==3 && nodeData(N))
			{	
				real* vertex = getGeom(p1);
				real* plane  = getGeom(N);

				real acc=0;
				for (int ii=0;ii<=pointdim;ii++)
				{
					checkpoint(  !isnan(plane[ii]) && !isnan(vertex[ii])  );
					acc+=plane[ii]*vertex[ii];
				}
				checkpoint(fabs(acc)<1e-1);
			}
		}
	}

	if (getLevel(N)==(key)(pointdim-1))
	{
		checkpoint(getNUp(N)<=2);
	}
}



///////////////////////////////////////////////////////////////////////////

ggraph* ggraph::join(key N,real coord)
{
	graph::iterator i;

	//find all zero-cells

	key **nav,*nnav=getNavigation(nav);
	findCells(0,N,nnav,nav);

	for (int u=0;u<this->pointdim;++u)
		this->next(true,nnav[u+1],nav[u+1],nnav[u],nav[u]);

	ggraph* ret=new ggraph(pointdim+1);
	
	key Njoin=ret->addNode(0);
	ret->setGeom(Njoin);
	ret->getGeom(Njoin)[pointdim+1]=coord;

	checkpoint(getNode(N)->Level==pointdim);

	key Cjoin=ret->addNode(pointdim+1);
	ret->addArch(Cjoin,Njoin);

	for (int k=0;k<=pointdim;k++)
	{
		key* kcells=nav[k];
		const key& nkcells=nnav[k];
		for (key cont=0;cont<nkcells;cont++)
		{
			key S   =kcells[cont];

			key Ck  =ret->addNode(k ) ;
			key Ckup=(k==pointdim)?Cjoin:ret->addNode(k+1);

			/*  geometry  */
			
#if 1
			if (!k) /* vertices */
			{	
				ret->setGeom(Ck,this,S);
			}
			else if (k==ret->pointdim-1) /* plane for (k-cell) */
			{
				real* gtmp=ret->setGeom(Ck);	
				gtmp[pointdim+1]=-1; //Zn<=0
				gtmp[0]=0;
			}
			
			if ((k+1)==(ret->pointdim-1)) /* plane for (k+1)-cell */
			{
				if (pointdim==1)  
				{

					real X=getGeom(S)[1];
					real  sign=(real)((getNUp(S)==2) || (getGeom(getNextUpBrother(getFirstUpArch(S)))[1]<X)?+1:-1);

					real* gtmp=ret->setGeom(Ckup);
					gtmp[0]=sign*(-coord*X);
					gtmp[1]=sign*(+coord);
					gtmp[2]=sign*(+X);
				}
				else 
				{
					real* gtmp=ret->setGeom(Ckup,this,S);
					gtmp[pointdim+1]=-gtmp[0]/coord;
				}
			}
#endif

			/* connectivity */

			nodeTmp(S)=Ck;
			ret->nodeTmp(Ck)=Ckup;

			ret->addArch(Ck,Ckup);

			if (!k)
			{
				ret->addArch(Njoin,Ckup);
				ret->addArch(Cjoin,Ck  );
			}
			else
			{
				for (i=goDw(S);!i.end();i++)
				{
					key map_ck  =nodeTmp(*i);
					key map_ckup=ret->nodeTmp(map_ck);
					ret->addArch(map_ck,Ck);
					ret->addArch(map_ckup,Ckup);
				}
			}	
		}
	}
		
	return ret;
}



/////////////////////////////////////////////////////////////

ggraph& ggraph::permutate(int* perm)
{
	list_iterator it;
	real tmp[MAX_GEOMETRIC_DIMENSION+1];

	for (it=each(0);!it.end();it++)
	{
		real* vg=getGeom(*it);

		for (int n=0;n<=this->pointdim;n++)
			tmp[n]=vg[perm[n]];
		
		memcpy(vg,tmp,sizeof(real)*(pointdim+1));
	}

	if (pointdim>=2)
	{
		for (it=each(pointdim-1);!it.end();it++)
		{
			if (nodeData(*it))
			{
				real* h=getGeom(*it);

				for (int n=0;n<=this->pointdim;n++)
					tmp[n]=h[perm[n]];
		
				memcpy(h,tmp,sizeof(real)*(pointdim+1));
			}
		}
	}

	for (it=each(pointdim);!it.end();it++)
	{
		if (nodeData(*it))
		{
			real* bball=getGeom(*it);
			bball[0]=-1;
		}
	}

	return *this;
}


///////////////////////////////////////////////////////////////////////////////////////////////

void ggraph::getBoundingBox(key npoints,key* points,real* m,real* M,real& limit,real& maxdelta,bool bStartup,bool bApplySpaceDim)
{
  int i;

  limit=maxdelta=0;
  int Dim=bApplySpaceDim?spacedim:pointdim;

  if (bStartup)
  {
		for (i=0;i<=Dim;i++)
		{
			m[i]=(real)+1e18;
			M[i]=(real)-1e18;
		}
	}

	real Tmp[MAX_GEOMETRIC_DIMENSION+1];
	
	for (int it=0;it<(int)npoints;it++)
	{
		real* v=getGeom(points[it]);

		if (bApplySpaceDim)
		{
			memset(Tmp,0,sizeof(Tmp));
			memcpy(Tmp,v,sizeof(real)*(pointdim+1));
			if (vmat) productv<real,real>(spacedim+1,Tmp,vmat,Tmp);
			v=Tmp;
		}

		for (i=0;i<=Dim;i++)
		{
			m[i]=min2(m[i],v[i]);
			M[i]=max2(M[i],v[i]);
			limit=(real)max3(limit,fabs(m[i]),fabs(M[i]));
			maxdelta=max2(maxdelta,M[i]-m[i]);
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
void ggraph::scaleToUnitBox()
{
	real m[MAX_GEOMETRIC_DIMENSION],M[MAX_GEOMETRIC_DIMENSION];
	real t[MAX_GEOMETRIC_DIMENSION+1],s[MAX_GEOMETRIC_DIMENSION+1],limit,maxdelta;

	key* Points=new key[MAX_CELLS_PER_LEVEL];
	key NPoints=findCells(0,0,Points);

	getBoundingBox(NPoints,Points,m,M,limit,maxdelta);
	
	for (int i=1;i<=pointdim;i++)
	{
		t[i-1]=(real)(-0.5*(m[i]+M[i]));
		s[i-1]=(real)(+2.0f/maxdelta); 
	}


	translate(t);
	scale(s);
}






