/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#include "generaldef.h"

#include <stdio.h>
#include <stdlib.h>

#include "ggraph.h"
#include "ply.h"

//----------------------------------------------------------------------
//----------------------------------------------------------------------

static  ggraph* readObj(char* filename)
{
	ggraph* g=new ggraph(3);

	char buf[2048];
	FILE* file=fopen(filename,"rt");
	
	if (!file)
		throw "cannot open file";
	
	int nline=-1;

	int numvertices=0;

	while (fscanf(file, "%s", buf) != EOF) 
	{
		switch (buf[0]) 
		{		
		case 'v': /* vertex! */
			{
				float X,Y,Z;
				fscanf(file, "%e %e %e", &X, &Y, &Z);
				real v[]={1,X,Y,Z};
				g->addVertex(v);
				break;
			}
			
		case 'f': /* face */
			{	
				key   p[256];
				key   id[256];
				key   n=0;
				
				int iv,nv;
				
				for (int i=0;fscanf(file,"%d//%d",&iv,&nv)==1;i++,++n)
				{
					id[n]=iv;

					if (iv<0) 
						iv=numvertices+iv;
					iv--;

					p[n]=iv+1;
				}

				g->addFace(n,p);
				break;
			}
			
		default: /* I don't care about the rest */
			fgets(buf, sizeof(buf), file);
			break;
		}
	}
	
	fclose(file);
	//setNormals(true);

	return g;
}



//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

static ggraph* readPly(char* filename,int PLY_VERTEX_ONE_REF=0)
{
	ggraph* g=new ggraph(3);
	

	PlyProperty px    ={    "x", Float32, Float32,  0, 0, 0, 0, 0};
	PlyProperty py    ={    "y", Float32, Float32,  4, 0, 0, 0, 0};
	PlyProperty pz    ={    "z", Float32, Float32,  8, 0, 0, 0, 0};
	PlyProperty pnx   ={   "nx", Float32, Float32, 12, 0, 0, 0, 0};
	PlyProperty pny   ={   "ny", Float32, Float32, 16, 0, 0, 0, 0};
	PlyProperty pnz   ={   "nz", Float32, Float32, 20, 0, 0, 0, 0};
	PlyProperty pfield={"field", Float32, Float32, 24, 0, 0, 0, 0};
	
	PlyProperty face_prop = {"vertex_indices", Int32, Int32, 4,PLY_LIST, Int32, Int32, 0};

	FILE* file = fopen( filename, "rb" );

	if (!file)
		perror("Error opening file ");

	checkpoint(file);

	PlyFile* ply  = read_ply( file );
	checkpoint(ply);
	
	int esize,nprops;
	PlyProperty** props = get_element_description_ply( ply, "vertex", &esize, &nprops);

	int numvertices=esize;
	
	PlyProperty* properties=new PlyProperty[7];
	
	bool has_normals=false;
	bool has_field  =false;

	unsigned int k,cont;
	for (k=0,cont=0;k<(unsigned int)nprops;k++)
	{
		if      (!strcmp(props[k]->name,"x"))     {                 properties[cont++]=px    ;}
		else if (!strcmp(props[k]->name,"y"))     {                 properties[cont++]=py    ;}
		else if (!strcmp(props[k]->name,"z"))     {                 properties[cont++]=pz    ;}
		else if (!strcmp(props[k]->name,"nx"))    {has_normals=true;properties[cont++]=pnx   ;}
		else if (!strcmp(props[k]->name,"ny"))    {has_normals=true;properties[cont++]=pny   ;}
		else if (!strcmp(props[k]->name,"nz"))    {has_normals=true;properties[cont++]=pnz   ;}
		else if (!strcmp(props[k]->name,"field")) {has_field  =true;properties[cont++]=pfield;}
	}	
	
	get_element_setup_ply( ply, "vertex", cont,properties);


	for(k=0;k<(unsigned int)numvertices; ++k) 
	{
		float coords[]={0,0,0,0,0,0,0};
		get_element_ply( ply, (void*)(coords));

		if (!has_field)
			coords[6]=coords[2];

		real v[]={1,coords[0],coords[1],coords[2]};
		key N=g->addVertex(v);
	}

	delete [] properties;
	
	struct 
	{
		key nverts;
		key *verts;
	} 
	face;

	key C=g->addNode(3);

#if 1
	for(k=0;k<(unsigned int)numvertices; ++k) 
		g->addArch(C,k+1);
#endif

	
	
	if (props=get_element_description_ply( ply, "face", &esize, &nprops))
	{
		get_element_setup_ply( ply, "face", 1, &face_prop);
		
		for(k=0; k<(unsigned int)esize; ++k ) 
		{
			get_element_ply( ply, (void*)&face);

			if (!PLY_VERTEX_ONE_REF)
			{
				for(key j=0;j<face.nverts; ++j) 
					++face.verts[j];
			}

			key F=g->addFace(face.nverts,face.verts);	
			g->addArch(F,C);
			free(face.verts);
		}
	}
	else
	{
		if (!(props = get_element_description_ply( ply, "tristrips", &esize, &nprops)))
		{
			checkpoint(0);
			exit(-1);
		}
		
		get_element_setup_ply(ply,"tristrips",1,&face_prop);
	
		for(k=0; k<(unsigned int)esize;++k ) 
		{
			get_element_ply( ply, (void*)&face);

			if (PLY_VERTEX_ONE_REF)
			{
				for(key j=0;j<face.nverts; ++j) 
					--face.verts[j];
			}
			
			for (key I=2,clockwise=1;I< face.nverts; I++) 
			{
				if (face.verts[I] == -1) 
					{I += 2;clockwise = 1;}
				else 
				{	
					if (clockwise)
					{
						key v[]={1+face.verts[I-2],1+face.verts[I-1],1+face.verts[I]};
						key F=g->addFace(3,v);
						g->addArch(F,C);
					}
					else
					{
						key v[]={1+face.verts[I-2],1+face.verts[I],1+face.verts[I-1]};
						key F=g->addFace(3,v);
						g->addArch(F,C);
					}
					
					clockwise = 1-clockwise;
				}
			}

		
			
			free(face.verts);
		}
	}

	close_ply( ply );
	//setNormals(false);

	

	return g;

}


//---------------------------------------------------------
//---------------------------------------------------------

static  ggraph* readTxt(char* filename)
{
	ggraph* g=new ggraph(3);

	FILE* file=fopen(filename,"rb");
	checkpoint(file);

	int header[2];
	fscanf(file,"%d %d",&header[0],&header[1]);

	int numvertices = header[0];
	int numtets     = header[1];

	real minf=+1e18f;
	real maxf=-1e18f;

	int id;
	for (id=0;id<numvertices;id++)
	{
		float X,Y,Z,F;
		fscanf(file,"%f %f %f %f",&X,&Y,&Z,&F);
		float v[]={1,X,Y,Z};
		g->addVertex(v);
		minf=min2(minf,F);
		maxf=max2(maxf,F);
	}


	for (id=0;id<numtets;id++)
	{
		key N0,N1,N2,N3;
		fscanf(file,"%d %d %d %d",&N0,&N1,&N2,&N3);
		checkpoint(N0!=N1 && N0!=N2 && N0!=N3 && N1!=N2 && N1!=N3 && N2!=N3);
		checkpoint(N0>=0 && N0<(key)numvertices && N1>=0 && N1<(key)numvertices && N2>=0 && N2<(key)numvertices && N3>=0 && N3<(key)numvertices);
		N0++;N1++;N2++;N3++;
		key Tet=g->addTet(N0,N1,N2,N3);
	}
	fclose(file);

	//calculate normals
	//setGrad();
	return g;
}



//---------------------------------------------------------
//---------------------------------------------------------
static ggraph* readNode(char* filename)
{
	ggraph* g=new ggraph(3);

	FILE* file=fopen(filename,"rb");
	checkpoint(file);

	int header[4];
	fscanf(file,"%d %d %d %d",&header[0],&header[1],&header[2],&header[3]);

	int numvertices=header[0];
	checkpoint(header[1]==3);

	real minf=+1e18f;
	real maxf=-1e18f;

	int i,id;
	for (i=0;i<numvertices;i++)
	{
		float X,Y,Z;
		fscanf(file,"%d %f %f %f",&id,&X,&Y,&Z);
		real v[]={1,X,Y,Z};
		g->addVertex(v);
	}

	fclose(file);

	char _filename[1024];
	strcpy(_filename,filename);
	strcpy(fileExtension(_filename),"ele");
	file=fopen(_filename,"rb");
	checkpoint(file);

	fscanf(file,"%d %d %d",&header[0],&header[1],&header[2]);

	int nt=header[0];
	checkpoint(header[1]==4);

	for (i=0;i<nt;i++)
	{
		int n0,n1,n2,n3;
		fscanf(file,"%d %d %d %d %d",&id,&n0,&n1,&n2,&n3);
		checkpoint(id>=0 && id<nt);
		checkpoint(n0!=n1 && n0!=n2 && n0!=n3 && n1!=n2 && n1!=n3 && n2!=n3);
		checkpoint(n0>=0 && n0<numvertices && n1>=0 && n1<numvertices && n2>=0 && n2<numvertices && n3>=0 && n3<numvertices);
		g->addTet(n0+1,n1+1,n2+1,n3+1);
	}

	fclose(file);

	//calculate normals
	//setGrad();

	return g;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

ggraph* ggraph::read(char* filename)
{
	char* ext=fileExtension(filename);

	if (!ext)
		fprintf(stderr,"error.cannot determinate extension for file %s\n",filename);
	
	if (!strcmp("obj",ext)) //2D
		return readObj(filename);

	if (!strcmp("ply",ext)) //2D
		return readPly(filename);

	if (!strcmp("txt",ext)) //3D
		return readTxt(filename);

	if (!strcmp("node",ext)) //3D
		return readNode(filename);

	throw "unknow file extension";
}


