/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#include "generaldef.h"

#include <stdio.h>
#include <stdlib.h>

#include "ggraph.h"

////////////////////////////////////////////////////////////////////////////////////////

#define ZER ((uint8)0x00)
#define POS ((uint8)0x01)
#define NEG ((uint8)0x02)
#define UNK ((uint8)0x03)
#define BOT ((uint8)0x04)


int ggraph::split(const key cell,real* planebsp,const real START_TOLERANCE,key& retcminus,key& retcplus,key& retcequal,key Id)
{
	static int nactivation=0;

	int pointdim  = this->pointdim;
	int celllvl   = getLevel(cell);

	++nactivation;

	key **nav,*nnav=getNavigation(nav);

	int i;
	real acc;
	normalize_plane(pointdim,planebsp);

	/* FIRST STEP: classification                       */

	int ntry=0;
	for (real tolerance=START_TOLERANCE;tolerance && ntry<3;++ntry,tolerance/=100)
	{
		int npos=0,nneg=0,nzer=0,npos_strict=0,nneg_strict=0;
		findCells(0,cell,nnav,nav);

		const key& npoints=nnav[0];
		const key *points=nav[0];

		for (i=0;i<(int)npoints;i++)
		{
			key v=points[i];

			real* gtmp=getGeom(v);

			switch (pointdim)
			{
			case 3:acc=planebsp[0]*gtmp[0]+planebsp[1]*gtmp[1]+planebsp[2]*gtmp[2]+planebsp[3]*gtmp[3];break;
			case 2:acc=planebsp[0]*gtmp[0]+planebsp[1]*gtmp[1]+planebsp[2]*gtmp[2]                    ;break;
			case 1:acc=planebsp[0]*gtmp[0]+planebsp[1]*gtmp[1]                                        ;break;
			default:
				{
					acc=0;
					for (int k=0;k<=pointdim;++k) 
						acc+=planebsp[k]*gtmp[k];
				}
			}

			if (acc>=0) 
				++npos_strict;
			else 
				++nneg_strict;

			if (acc>tolerance)
				{++npos;getNode(v)->Sign=POS;}
			
			else if (acc<-tolerance) 
				{++nneg;getNode(v)->Sign=NEG;}

			else /* mi segno che e' un elemento preesistente */                    
				{++nzer;getNode(v)->Sign=ZER;nodeTmp(v)=0;}
		}
	
		if (nzer && !npos && !nneg)
		{
			retcplus=retcminus=retcequal=0;

#if 1
			//even when if flat try to use the strict classification 
			if (npos_strict==(int)npoints)
			{
				checkpoint(!nneg_strict);
				retcplus=cell;
				return OK;
			}

			if (nneg_strict==(int)npoints)
			{
				checkpoint(!npos_strict);
				retcminus=cell;
				return OK;
			}
#endif
			
			return FLAT;
		}

		else if (npos && !nneg)
		{
			retcplus =cell;
			retcminus=retcequal=0;
			return OK;
		}
		else if (nneg && !npos)
		{
			retcminus=cell;
			retcplus=retcequal=0;
			return OK;
		}

		/* preparo per la navigazione a tutti i livelli */

		if (celllvl==pointdim)
		{
			for (i=0;i<celllvl;++i) 
				next(true,nnav[i+1],nav[i+1],nnav[i],nav[i]);
		}

		/* uso il (celllvl+1)-nav per segnarmi le celle [=] create in caso di abort */

		key& nnewcells=nnav[celllvl+1];
		key* newcells=nav[celllvl+1];
		nnewcells=0;

		/* inizia il loop di verifica */

		for (int d=1;d<=celllvl;++d)
		{
			key& nnavd=nnav[d];
			key  *navd=nav [d];
		   int newnum=0;

			for (int i=0;i<(int)nnavd;++i)
			{
				/* classificazione in senso stretto, senza constraint */
				key face=navd[i];

				checkpoint(getNDw(face)>d && (d!=1 || getNDw(face)==2));

				if (d!=celllvl)
				{
					for (iterator linkj=goUp(face);!linkj.end();linkj++)
						getNode(*linkj)->Sign=UNK;
				}

				int npos=0,nneg=0,nzer=0,ntosplit=0;
				for (iterator linkj=goDw(face);!linkj.end();linkj++)
				{
					switch(getNode(*linkj)->Sign)
					{
						case POS:++npos;break;
						case NEG:++nneg;break;
						case ZER:++nzer;break; 
						case BOT:++npos;++nneg;++ntosplit;break; 
					}
				}

				if (!npos) 
					getNode(face)->Sign=(!nneg)?ZER:NEG;
				else  if (!nneg)
					getNode(face)->Sign=POS;
				else
				{
					if (nzer) /* caso npos, nneg, nzer. non lo posso permettere */
						goto ClassificationError;

					getNode(face)->Sign=BOT;
					navd[newnum++]=face;
				}

				/* simula la creazione dell'intersezione per poi controllarla dopo! */
				checkpoint(getNode(face)->Sign!=UNK);

				if (getNode(face)->Sign!=BOT) 
					continue;
				
				/* ulteriore controllo (le due celle che verranno create devono essere piene) */
				
				checkpoint(npos && nneg && !nzer);

				if ((npos+1)<=d || (nneg+1)<=d)
					goto ClassificationError;

				key mapface=addNode(getLevel(face)-1);
				getNode(mapface)->Id=Id;

				getNode(mapface)->Sign=ZER;
				nodeTmp(face   )= mapface;
				nodeTmp(mapface)=face; /* mi segno che  un elemento new */

				/* mi segno la faccia in caso di abort */
				newcells[nnewcells++]=mapface;

				if (d==1) continue;

				for (iterator linkm=goDw(face);!linkm.end();linkm++)
				{
					key down=*linkm;
					checkpoint(getNode(down)->Sign!=ZER && getNode(down)->Sign!=UNK);

					switch (getNode(down)->Sign)
					{
					case POS:
					case NEG:
						{
							for (iterator linkn=goDw(down);!linkn.end();linkn++)
							{
								key downdown=*linkn;
								checkpoint(getNode(downdown)->Sign==getNode(down)->Sign || getNode(downdown)->Sign==ZER);

								if (getNode(downdown)->Sign==ZER && !findArch(downdown,mapface))
									addArch(downdown,mapface);
							}
							break;
						}

					case BOT: /* deve avere per forza una cella [=] creata nuova */
						addArch(nodeTmp(down),mapface);
						break;
					}
				}
				
				
				if ((getNDw(mapface)<=getLevel(mapface)) /* at least k-dimensional */
					 || (getLevel(mapface)==1 && getNDw(mapface)!=2) /* not good since an edge must have two vertices */
					 || (getLevel(mapface)==2 && !orderFace(mapface))) /* not good since a 2-face must be an ordered polygon*/
					goto ClassificationError;
			}
			
			nnavd=newnum;

			if (d==celllvl && nnavd!=1)
				goto ClassificationError;
		}

		/* classification successfull (both sign) */	
		goto SymbolicSplit;

ClassificationError:

		fputc(!ntry?'F':'.',stdout);
		fflush(stdout);


		/* devo togliere i link che non devono esserci */

		for (i=0;i<(int)nnewcells;++i)
		{
			key mapface=newcells[i];

			key A;
			for (A;A=getFirstUpArch(mapface);) remArch(A);
			for (A;A=getFirstDwArch(mapface);) remArch(A);
			remNode(mapface);
		}

	} /* for (tolerance=1e-4...) */

	
	return IMPOSSIBLE;

SymbolicSplit:

	/* create two new cells and connect them to vertices */
	retcminus=addNode(celllvl); 
	retcplus =addNode(celllvl); 

	if (celllvl==pointdim)
	{
		while (getNUp(cell))
		{
			key v=getFirstUpNode(cell);
			remArch(getFirstUpArch(cell));

			switch (getNode(v)->Sign)
			{
				case POS:addArch(retcplus,v);break;
				case NEG:addArch(retcminus,v);break;
				case ZER:addArch(retcplus ,v);addArch(retcminus,v);break;
				default:checkpoint(0);break;
			}			
		}
	}

	for (int k=1;k<=celllvl;++k)
	{		
		key* navk=nav[k];
		const key& nnavk=nnav[k];

		for (i=0;i<(int)nnavk;++i)
		{
			key face=navk[i];

			checkpoint(getNode(face)->Sign==BOT);

			/* must split in two cells cell+ and cell-. create a k-1 cell which split the cell */
			key cminus = (k!=celllvl)?addNode(k):retcminus ;getNode(cminus)->Sign = NEG;
			key cplus  = (k!=celllvl)?addNode(k):retcplus  ;getNode(cplus )->Sign = POS;

			getNode(cminus)->Id=getNode(face)->Id;
			getNode(cplus )->Id=getNode(face)->Id;

			/* create the new = cell (which is a vertex for k=1) and connect to the new cells */
			key newface=nodeTmp(face);

			if (k==celllvl) 
				retcequal=newface;

			checkpoint(getNode(newface)->Sign==ZER);

			/* plane information */

			if (k==(pointdim-1))
			{
				setGeom(cminus,this,face);
				setGeom(cplus,this,face);
			}

			/* plane information */

			else if (k==pointdim)
			{
				real* gtmp=setGeom(newface);
				memcpy(gtmp,planebsp,(pointdim+1)*sizeof(real));
			}

			/* it's a vertex... create the new coordinates */
			if (k==1)
			{
				real valuev1,valuev2;

				
				real* gtmp = setGeom(newface);
				key p1   = getFirstDwNode(face);real* gv1=getGeom(p1);
				key p2   = getLastDwNode (face);real* gv2=getGeom(p2);

				switch (pointdim)
				{
				case 3:valuev1=planebsp[0]*gv1[0]+planebsp[1]*gv1[1]+planebsp[2]*gv1[2]+planebsp[3]*gv1[3];
						 valuev2=planebsp[0]*gv2[0]+planebsp[1]*gv2[1]+planebsp[2]*gv2[2]+planebsp[3]*gv2[3];
						 break;
				case 2:valuev1=planebsp[0]*gv1[0]+planebsp[1]*gv1[1]+planebsp[2]*gv1[2];
						 valuev2=planebsp[0]*gv2[0]+planebsp[1]*gv2[1]+planebsp[2]*gv2[2];
						 break;
				case 1:valuev1=planebsp[0]*gv1[0]+planebsp[1]*gv1[1];
						 valuev2=planebsp[0]*gv2[0]+planebsp[1]*gv2[1];
						 break;

				default:
					{
						valuev1=valuev2=0;
						for (int j=0;j<=pointdim;++j)
						{
							valuev1+=planebsp[j]*gv1[j];
							valuev2+=planebsp[j]*gv2[j];
						}
						break;
					}
				}

				if (valuev1<0) valuev1*=-1;
				if (valuev2<0) valuev2*=-1;

				real alpha =1/(valuev1+valuev2);
				real beta  =valuev1*alpha;
				alpha *=valuev2;

				gtmp[0]=1;
				switch (pointdim)
				{
				case 3:gtmp[3]=alpha*gv1[3]+beta*gv2[3];
				case 2:gtmp[2]=alpha*gv1[2]+beta*gv2[2];
				case 1:gtmp[1]=alpha*gv1[1]+beta*gv2[1];
						 break;

				default:
					{
						for (int j=1;j<=pointdim;++j)
							gtmp[j]=alpha*gv1[j]+beta*gv2[j];
						break;
					}
				}

				/* add new new vertex to the new faces */ 

				if (celllvl==pointdim)
				{
					addArch(retcplus,newface);
					addArch(retcminus,newface);
				}

				/*
				importante: devo collegare il nuovo vertice a tutte le dim-celle che sono incidenti
				sull'edge.... siccome non vado in alto per stabilirle, passo dai vertici.
				Le dim-celle incidenti sull'edge sono tutte quelle che appartengono a
				INTERSEZIONE(dim-celle che incidono su P1,dim-celle che incidono su P2)
				*/
				for (iterator linkkk=goDw(p1);!linkkk.end();linkkk++)
				for (iterator linkhh=goDw(p2);!linkhh.end();linkhh++)
				{
					if (*linkkk==*linkhh)
					{
						addArch(*linkkk,newface);
						break;
					}	
				}
			}

			/* link to sub faces with the same sign. */

			bool linked=false;
			key Arch;
			while(Arch=getFirstDwArch(face))
			{
				key archinfo=archData(Arch);
				key facedown=getN0(Arch);

				switch(getNode(facedown)->Sign)
				{
					case POS:archData(addArch(facedown,cplus ))=archinfo;break;
					case NEG:archData(addArch(facedown,cminus))=archinfo;break;
					default:checkpoint(0);break;
				}

				/* voglio mantenere l'ordinamento per le 2-facce */

				if (k==2 && !linked)
				{
					key nextedge=getN0(getNextDwArch(Arch)); 

					uint8 s1=getSign(getFirstDwNode(nextedge));
					uint8 s2=getSign(getLastDwNode (nextedge));
					uint8 s3=getSign(getFirstDwNode(facedown));
					uint8 s4=getSign(getLastDwNode (facedown));

					if ((s3==ZER && s1==ZER) || (s3==ZER && s2==ZER) || (s4==ZER && s1==ZER) || (s4==ZER && s2==ZER))
					{
						addArch(newface,cplus );
						addArch(newface,cminus);
						linked=true;
					}
				}

				remArch(Arch);
			}

			checkpoint(k!=2 || linked);

			if (k!=2)
			{
				addArch(newface,cplus );
				addArch(newface,cminus);
			}

			/* link all up faces to cplus and cminus */

			if (celllvl!=pointdim || k<pointdim)
			{
				while (getNUp(face))
				{	
					key archup	 = getFirstUpArch(face);
					key faceup	 = getN1(archup);
					key archinfo = archData(archup);
					
					/* qui devo fare un inserimento opportuno nei links per mantenere ordinamento */

					key aup0,aup1;

					if (getLevel(faceup)==2) 
					{	
						key nextarch=getNextDwArch(getFirstUpArch(face));
						key nextface=getN0(nextarch);
						bool goodorder=findCommonNode(cplus,nextface,DIRECTION_DOWN)!=0;
						aup0=addArch(goodorder?cminus:cplus ,faceup,DIRECTION_UP_AND_DOWN,0,nextarch);
						aup1=addArch(goodorder?cplus :cminus,faceup,DIRECTION_UP_AND_DOWN,0,nextarch);
					}
					else
					{
						aup0=addArch(cplus ,faceup);
						aup1=addArch(cminus,faceup);
					}

					archData(aup0)=archinfo;
					archData(aup1)=archinfo;

					remArch(getFirstUpArch(face));
				}
			}

			checkpoint(getLevel(newface)!=1 || getNDw(newface)==2);
			remNode(face);
		}
	}

	return OK;
}

