/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#include "generaldef.h"

#include <stdio.h>
#include <stdlib.h>

#include "ggraph.h"


///////////////////////////////////////////////////////////////////////////////////////////////
// vt=[vx,vy,vz,...]
/*
with(LinearAlgebra): mdir:=<<1,+vx,+vy,+vz>|<0,1,0,0>|<0,0,1,0>|<0,0,0,1>>;
> MatrixMatrixMultiply(mdir, <<1,x,y,z>>);
> with(LinearAlgebra):minv:=MatrixInverse(mdir);minv;
> MatrixMatrixMultiply(<<d>|<a>|<b>|<c>>, minv);
*/
///////////////////////////////////////////////////////////////////////////////////////////////

ggraph& ggraph::translate(real* vt)
{
	int k;
	list_iterator i;

	switch (pointdim)
	{
	case 0:
		{
			break;
		}

	case 1:
		{
			//vertices
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]+=vt[0];
			}
			break;
		}

	case 2:
		{
			//vertces
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]+=vt[0];
				gtmp[2]+=vt[1];
			}

			//planes
			for (i=each(pointdim-1);!i.end();i++)
			{
				key G=*i;
				real* gtmp=getGeom(G);
				gtmp[0]-=gtmp[1]*vt[0]+gtmp[2]*vt[1];
				normalize_plane(pointdim,gtmp);
			}

			break;
		}

	case 3:
		{
			//vertices
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]+=vt[0];
				gtmp[2]+=vt[1];
				gtmp[3]+=vt[2];
			}

			//planes
			for (i=each(pointdim-1);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[0]-=gtmp[1]*vt[0]+gtmp[2]*vt[1]+gtmp[3]*vt[2];
				normalize_plane(pointdim,gtmp);
			}

			break;
		}

	default:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				for (int k=1;k<=pointdim;k++) gtmp[k]+=vt[k-1];
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				for (k=1;k<=pointdim;k++) gtmp[0]-=gtmp[k]*vt[k-1];
				normalize_plane(pointdim,gtmp);
			}
			
			break;
		}
	}

	//invalid radius for bounding ball
	if (pointdim>=2)
	{
		for (i=each(pointdim);!i.end();i++)
		{
			if (nodeData(*i)) getGeom(*i)[0]=-1;
		}
	}

	return *this;
}


///////////////////////////////////////////////////////////////////////////////////////////////
// vs=[sx,sy,...]
/*
with(LinearAlgebra): mdir:=<<1,0,0,0>|<0,sx,0,0>|<0,0,sy,0>|<0,0,0,sz>>;
> MatrixMatrixMultiply(mdir, <<1,x,y,z>>);
> with(LinearAlgebra):minv:=MatrixInverse(mdir);minv;
> MatrixMatrixMultiply(<<d>|<a>|<b>|<c>>, minv);
*/
///////////////////////////////////////////////////////////////////////////////////////////////

ggraph& ggraph::scale(real* vs)
{	
	list_iterator i;

	switch (pointdim)
	{
	case 0:
		{
			break;
		}

	case 1:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]*=vs[0];
			}
			break;
		}

	case 2:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]*=vs[0];
				gtmp[2]*=vs[1];
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]/=vs[0];
				gtmp[2]/=vs[1];
				normalize_plane(pointdim,gtmp);
			}

			break;
		}

	case 3:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]*=vs[0];
				gtmp[2]*=vs[1];
				gtmp[3]*=vs[2];
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[1]/=vs[0];
				gtmp[2]/=vs[1];
				gtmp[3]/=vs[2];
				normalize_plane(pointdim,gtmp);
			}

			break;
		}

	default:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				for (int k=1;k<=pointdim;k++) gtmp[k]*=vs[k-1];
			}
	
			for (i=each(pointdim-1);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				for (int k=1;k<=pointdim;k++) gtmp[k]/=vs[k-1];
				normalize_plane(pointdim,gtmp);
			}

			break;
		}
	}

	//invalid radius for bounding ball
	if (pointdim>=2)
	{
		for (i=each(pointdim);!i.end();i++)
		{
			if (nodeData(*i)) getGeom(*i)[0]=-1;
		}
	}

	return *this;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// ax1,ax2=[1...pointdim]
/*
with(LinearAlgebra): mdir:=<<1,0,0,0>|<0,Cos,0,-Sin>|<0,0,1,0>|<0,+Sin,0,+Cos>>;
MatrixMatrixMultiply(mdir, <<1,x,y,z>>);
with(LinearAlgebra):minv:=MatrixInverse(mdir);minv;
MatrixMatrixMultiply(<<d>|<a>|<b>|<c>>, minv);
*/
///////////////////////////////////////////////////////////////////////////////////////////////
ggraph& ggraph::rotate(int ax1,int ax2,real alpha)
{
	list_iterator i;
	checkpoint (pointdim>=2 && ax1 && ax2);

	real c=(real)cos(alpha);
	real s=(real)sin(alpha);

	for (i=each(0);!i.end();i++)
	{
		real* gtmp=getGeom(*i);
		real Ai=+c*gtmp[ax1]+s*gtmp[ax2];
		real Aj=-s*gtmp[ax1]+c*gtmp[ax2];
		gtmp[ax1]=Ai;
		gtmp[ax2]=Aj;
	}
	
	for (i=each(pointdim-1);!i.end();i++)
	{
		real* gtmp=getGeom(*i);
		real Ai=+c*gtmp[ax1]+s*gtmp[ax2];
		real Aj=-s*gtmp[ax1]+c*gtmp[ax2];
		gtmp[ax1]=Ai;
		gtmp[ax2]=Aj;

		normalize_plane(pointdim,gtmp);
	}

	//invalid radius for bounding ball
	if (pointdim>=2)
	{
		for (i=each(pointdim);!i.end();i++)
		{
			if (nodeData(*i)) getGeom(*i)[0]=-1;
		}
	}

	return *this;
}


///////////////////////////////////////////////////////////////////////////////////////////////
/*
v=[sx,sy,sz,...] ref=[1,2,...]

wwith(LinearAlgebra): mdir:=<<1,0,0,0>|<0,sx,0,0>|<0,sy,1,0>|<0,sz,0,1>>;
MatrixMatrixMultiply(mdir, <<1,x,y,z>>);
with(LinearAlgebra):minv:=MatrixInverse(mdir);minv;
MatrixMatrixMultiply(<<d>|<a>|<b>|<c>>, minv);
wwith(LinearAlgebra): g:= <<a00,a01,a02,a03>|<a10,a11,a12,a13>|<a20,a21,a22,a23>|<a30,a31,a32,a33>>;
*/
///////////////////////////////////////////////////////////////////////////////////////////////

ggraph& ggraph::shear(int ref,const real* v)
{
	int k;
	list_iterator i;

	switch (pointdim)
	{
	case 0:
		{
			break;
		}

	case 1:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[ref]=gtmp[1]*v[0];
			}
			break;
		}

	case 2:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[ref]=gtmp[1]*v[0]+gtmp[2]*v[1];
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				if (ref!=1) gtmp[1]-=gtmp[ref]*v[0]/v[ref-1];
				if (ref!=2) gtmp[2]-=gtmp[ref]*v[1]/v[ref-1];
				gtmp[ref]/=v[ref-1];
				normalize_plane(pointdim,gtmp);
			}

			break;
		}

	case 3:
		{
			for (i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				gtmp[ref]=gtmp[1]*v[0]+gtmp[2]*v[1]+gtmp[3]*v[2];
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				real* gtmp=getGeom(*i);

				if (ref!=1) gtmp[1]-=gtmp[ref]*v[0]/v[ref-1];
				if (ref!=2) gtmp[2]-=gtmp[ref]*v[1]/v[ref-1];
				if (ref!=3) gtmp[3]-=gtmp[ref]*v[2]/v[ref-1];
				gtmp[ref]/=v[ref-1];
				normalize_plane(pointdim,gtmp);
			}

			break;
		}

	default:
		{
	
			for (list_iterator i=each(0);!i.end();i++)
			{
				real* gtmp=getGeom(*i);
				real acc=0;
				for (k=1;k<=pointdim;k++) acc+=gtmp[k]*v[k-1];
				gtmp[ref]=acc;
			}
			
			if (pointdim>1)
			{
				for (list_iterator i=each(pointdim-1);!i.end();i++)
				{
					real* gtmp=getGeom(*i);
					for (int k=1;k<=pointdim;k++) 
						if (k!=ref) gtmp[k]-=gtmp[ref]*v[k-1]/v[ref-1];
					gtmp[ref]/=v[ref-1];

					normalize_plane(pointdim,gtmp);
				}
			}
			
			break;
		}
	}

	//invalid radius for bounding ball
	if (pointdim>=2)
	{
		for (i=each(pointdim);!i.end();i++)
		{
			if (nodeData(*i)) getGeom(*i)[0]=-1;
		}
	}

	return *this;
}


///////////////////////////////////////////////////////////////////////////////////////////////
// generic transformation
///////////////////////////////////////////////////////////////////////////////////////////////

ggraph& ggraph::transform(const real* vmat,const real *hmat)
{
	list_iterator i;

	switch (pointdim)
	{
	case 0:
		{
			for (i=each(0);!i.end();i++)
			{
				real* point=getGeom(*i);
				real a0=point[0]*vmat[0];
				point[0]=a0;
			}

			break;
		}

	case 1:
		{
			for (i=each(0);!i.end();i++)
			{
				real* point=getGeom(*i);
				real a0=point[0]*vmat[0]+point[1]*vmat[1];
				real a1=point[0]*vmat[2]+point[1]*vmat[3];
				point[0]=1;
				point[1]=a1/a0;
			}

			break;
		}

	case 2:
		{
			for (i=each(0);!i.end();i++)
			{
				real* point=getGeom(*i);
				real a0=point[0]*vmat[0]+point[1]*vmat[1]+point[2]*vmat[2];
				real a1=point[0]*vmat[3]+point[1]*vmat[4]+point[2]*vmat[5];
				real a2=point[0]*vmat[6]+point[1]*vmat[7]+point[2]*vmat[8];
				point[0]=1;
				point[1]=a1/a0;
				point[2]=a2/a0;
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				if (nodeData(*i))
				{
					real* plane=getGeom(*i);
					real a0=plane[0]*hmat[0]+plane[1]*hmat[3]+plane[2]*hmat[6];
					real a1=plane[0]*hmat[1]+plane[1]*hmat[4]+plane[2]*hmat[7];
					real a2=plane[0]*hmat[2]+plane[1]*hmat[5]+plane[2]*hmat[8];
					plane[0]=a0;
					plane[1]=a1;
					plane[2]=a2;
					normalize_plane(pointdim,plane);
				}
			}

			break;
		}

	case 3:
		{
			for (i=each(0);!i.end();i++)
			{
				real* point=getGeom(*i);
				real a0=point[0]*vmat[ 0]+point[1]*vmat[ 1]+point[2]*vmat[ 2]+point[3]*vmat[ 3];
				real a1=point[0]*vmat[ 4]+point[1]*vmat[ 5]+point[2]*vmat[ 6]+point[3]*vmat[ 7];
				real a2=point[0]*vmat[ 8]+point[1]*vmat[ 9]+point[2]*vmat[10]+point[3]*vmat[11];
				real a3=point[0]*vmat[12]+point[1]*vmat[13]+point[2]*vmat[14]+point[3]*vmat[15];
				point[0]=1;
				point[1]=a1/a0;
				point[2]=a2/a0;
				point[3]=a3/a0;
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				if (nodeData(*i))
				{
					real* plane=getGeom(*i);
					real a0=plane[0]*hmat[0]+plane[1]*hmat[4]+plane[2]*hmat[ 8]+plane[3]*hmat[12];
					real a1=plane[0]*hmat[1]+plane[1]*hmat[5]+plane[2]*hmat[ 9]+plane[3]*hmat[13];
					real a2=plane[0]*hmat[2]+plane[1]*hmat[6]+plane[2]*hmat[10]+plane[3]*hmat[14];
					real a3=plane[0]*hmat[3]+plane[1]*hmat[7]+plane[2]*hmat[11]+plane[3]*hmat[15];
					plane[0]=a0;
					plane[1]=a1;
					plane[2]=a2;
					plane[3]=a3;
					normalize_plane(pointdim,plane);
				}
			}

			break;
		}

	default:
		{
			real tmp[MAX_GEOMETRIC_DIMENSION+1];

			for (i=each(0);!i.end();i++)
			{
				real* point=getGeom(*i);
				memset(tmp,0,sizeof(real)*(pointdim+1));
		
				int I,J;
				for (I=0;I<=pointdim;I++)
				for (J=0;J<=pointdim;J++)
					tmp[I]+=vmat[I*(pointdim+1)+J]*point[J];

				for (I=1;I<=pointdim;I++) tmp[I]/=tmp[0];
				tmp[0]=1;

				memcpy(point,tmp,sizeof(real)*(pointdim+1));
			}

			for (i=each(pointdim-1);!i.end();i++)
			{
				if (nodeData(*i))
				{
					real* plane=getGeom(*i);
					memset(tmp,0,sizeof(real)*(pointdim+1));

					for (int I=0;I<=pointdim;I++)
					for (int J=0;J<=pointdim;J++)
						tmp[J]+=plane[I]*hmat[I*(pointdim+1)+J];

					memcpy(plane,tmp,sizeof(real)*(pointdim+1));
					normalize_plane(pointdim,plane);
				}
			}

			break;
		}
	}

	//invalid radius for bounding ball
	if (pointdim>=2)
	{
		for (i=each(pointdim);!i.end();i++)
		{
			if (nodeData(*i)) getGeom(*i)[0]=-1;
		}
	}

	return *this;

}

