/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#include "generaldef.h"

#include <stdio.h>
#include <stdlib.h>

#include "ggraph.h"


///////////////////////////////////////////////////////////////////////////

void ggraph::triangulate_inner(key N)
{
	graph::iterator i;
	
	uint8 celllvl=getNode(N)->Level;

	//not necessary
	if (celllvl<=1)
		return;

	//not necessary (simplex)
	if (getNDw(N)==(celllvl+1))
		return;

	key **nav,*nnav=getNavigation(nav);

	key* newfaces  =nav [celllvl+1];
	key& nnewfaces =nnav[celllvl+1];
	nnewfaces=0;

	//special case (do not add point)
	if (celllvl==2)
	{
		key F=N;
		key T[2],nt=0;

		key Prev0=getFirstDwNode(getLastDwNode(F)),P1;
		key Prev1=getLastDwNode (getLastDwNode(F)),P2;
		key FirstNode=0;

		for (graph::iterator i=goDw(F);!i.end();i++,Prev0=P1,Prev1=P2)
		{
			key E=*i;
			P1=getFirstDwNode(E);
			P2=getLastDwNode (E);
			if (P1!=Prev0 && P1!=Prev1) Swap(key,P1,P2);
			if (!FirstNode) FirstNode=P1;

			T[nt++]=E;

			if (nt==2)
			{
				key NewFace=addNode(2);
				newfaces[nnewfaces++]=NewFace;

				if (pointdim!=2)
				{
					for (graph::iterator jt=goUp(F);!jt.end();jt++)
						addArch(NewFace,*jt);
				}
				
				addArch(T[0],NewFace);
				addArch(T[1],NewFace);

				key ToFirstEdge;
				
				if (i.next().next().end()) 
				{
					ToFirstEdge=*i.next();
					addArch(ToFirstEdge,NewFace);
					break;
				}
				
				ToFirstEdge=addNode(1);
				addArch(FirstNode,ToFirstEdge);
				addArch(P2       ,ToFirstEdge);

				addArch(ToFirstEdge,NewFace);
				nt=0;
				T[nt++]=ToFirstEdge;
			}
		}
	}
	else
	{
		//general case
		findCells(0,N,nnav,nav);

		key* points=nav[0];
		const key& npoints=nnav[0];

		key Centroid=addNode(0);
		real* vg=setGeom(Centroid);

		real coeff=(real)(1.0/npoints);

		for (key cont=0;cont<npoints;cont++)
		{
			real* gtmp=getGeom(points[cont]);
			for (int n=1;n<=pointdim;n++) vg[n]+=coeff*gtmp[n];
		}

		for (int k=0;k<celllvl;k++)
		{
			key* cells=nav[k];
			const key& ncells=nnav[k];

			for (key cont=0;cont<ncells;cont++)
			{
				key S   =cells[cont];
				key Ckup=addNode(k+1);

				//new celllvl-cells
				if (k==(celllvl-1)) 
					newfaces[nnewfaces++]=Ckup;

				if (celllvl!=pointdim && (k+1)==celllvl)
				{
					for (i=goUp(N);!i.end();i++)
						addArch(Ckup,*i);
				}

				nodeTmp(S)=Ckup;
				addArch(S,Ckup);

				if (!k)
				{
					addArch(Centroid,Ckup);
				}
				else
				{
					for (i=goDw(S);!i.end();i++)
						addArch(nodeTmp(*i),Ckup);
				}

				//DO NOT STORE plane equation 
				/*
				if (pointdim>=2 && k==(pointdim-1) && nodeData(S))
				{
					db.release(nodeData(S));
					nodeData(S)=0;
				}
				*/
			}
		}
	}

	remNode(N);

	//link simplex to points
	if (pointdim>=2 && celllvl==pointdim)
	{
		for (key cont=0;cont<nnewfaces;cont++)
		{
			key NewCell = newfaces[cont];
			checkpoint(getLevel(NewCell)==celllvl);

			findCells(0,NewCell,nnav,nav);
			key npoints=nnav[0],*points=nav[0];

			for (int n=0;n<(int)npoints;n++)
				addArch(NewCell,points[n]);

			checkpoint(getNUp(NewCell)==(celllvl+1) && getNDw(NewCell)==(celllvl+1));
		}
	}
}

///////////////////////////////////////////////////////////////////////////

ggraph& ggraph::triangulate()
{

	for (int lvl=2;lvl<=pointdim;lvl++)
	{
		for (list_iterator it=each(lvl),next;!it.end();it=next)
		{
			next=it.next();
			triangulate_inner(*it);
		}
	}

	return *this;
}

