/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#include "generaldef.h"

#include <stdio.h>
#include <time.h>


#include "memcache.h"
#include "graph.h"


int graph::MainNumGGraph=0;


//////////////////////////////////////////////////////////////////////////
graph::graph() 
	:nodes(sizeof(node_t)),arches(sizeof(arch_t)),refcount(0)
{
	NumGGraph=MainNumGGraph++;
	memset(&nodelists,0,sizeof(nodelists));
}

graph::graph(const graph& src)
	:arches(src.arches),nodes(src.nodes),refcount(0)
{
	NumGGraph=MainNumGGraph++;
	memcpy(&(this->nodelists),&src.nodelists,sizeof(this->nodelists));
}

graph::~graph()
{
}



////////////////////////////////////////////////////////////////////////////////////////
key graph::findCells(int level,key Cell,key* dst)
{
	static key _nav[MAX_CELLS_PER_LEVEL];

	//all the cells at a certain level
	if (!Cell)
	{
		int cont=0;
		for (list_iterator it=each(level);!it.end();it++) 
		{
			*dst++=*it;
			cont++;
		}
		return getNCells(level);
	}

	int celllvl=getLevel(Cell);

	//the cell itself
	if (celllvl==level) 
	{
		*dst++=Cell;
		return 1;
	}

	key* nav[2]={dst,_nav};
	key  nnav[2]={0,0};
	int  cur=0;

	nav[cur][nnav[cur]++]=Cell;

	if (level<celllvl)
	{
		for (int i=celllvl;i>=(level+1);i--,cur=1-cur) 
			next(false,nnav[1-cur],nav[1-cur],nnav[cur],nav[cur]);
	}
	else
	{
		for (int i=celllvl;i<=(level-1);i++,cur=1-cur) 
			next(true,nnav[i-cur],nav[1-cur],nnav[cur],nav[cur]);
	}

	if (nav[cur]!=dst)
		memcpy(dst,nav[cur],sizeof(key)*nnav[cur]);
	
	return nnav[cur];
}



////////////////////////////////////////////////////////////////////////////////////////
void graph::findCells(int level,key Cell,key* nnav,key** nav)
{
	//all the cells at a certain level

	if (!Cell)
	{
		key* dst=nav[level];
		for (list_iterator it=each(level);!it.end();it++) *dst++=*it;
		nnav[level]=getNCells(level);
		return;
	}

	int celllvl=getLevel(Cell);
	nav[celllvl][0]=Cell;
	nnav[celllvl]=1;

	//the cell itself
	if (celllvl==level) 
		return;


	if (level<celllvl)
	{
		for (int i=celllvl;i>=(level+1);i--) 
			next(false,nnav[i-1],nav[i-1],nnav[i],nav[i]);
	}
	else
	{
		for (int i=celllvl;i<=(level-1);i++) 
			next(true,nnav[i+1],nav[i+1],nnav[i],nav[i]);
	}
}



//////////////////////////////////////////////////////////////////////////

graph& graph::operator=(const graph& src)
{
	this->arches     =src.arches;
	this->nodes      =src.nodes;
	this->nodelists  =src.nodelists;

	return *this;
}


void graph::clear()
{
	nodes.clear();
	arches.clear();
	memset(&nodelists,0,sizeof(nodelists));
}



//////////////////////////////////////////////////////////////////////////

key graph::addArch(key N0,key N1,Direction dir,key _A0,key _A1)
{
	key K=arches.block();

	arch_t& a=*getArch(K);

	//important when a.N0_prev=A.N0_next=0 means that the arch exists only for dir=DIRECTION_DOWN
	//               a.N1_prev=A.N1_next=0 means that the arch exists only for dir=DIRECTION_UP

	memset(&a,0,sizeof(arch_t)); 
	
	node_t& n0=*getNode(N0);
	node_t& n1=*getNode(N1);

	a.N0=N0;
	a.N1=N1;

	if (dir!=DIRECTION_DOWN)
	{
		if (!(n0.Nup++)) 
			{n0.Up=a.N0_prev=a.N0_next=K;}
		else
		{
			key A0=_A0;
			if (!A0) A0=n0.Up;
			a.N0_next=A0;
			a.N0_prev=getArch(A0)->N0_prev;
			getArch(a.N0_prev)->N0_next=getArch(A0)->N0_prev=K;
			if (_A0 && _A0==n0.Up) n0.Up=K;
		}
	}
	else
	{
		a.N0_prev=a.N0_next=0;
	}

	if (dir!=DIRECTION_UP)
	{
		if (!(n1.Ndw++)) 
			{n1.Dw=a.N1_prev=a.N1_next=K;}
		else
		{
			key A1=_A1;
			if (!A1) A1=n1.Dw;
			a.N1_next=A1;
			a.N1_prev=getArch(A1)->N1_prev;
			getArch(a.N1_prev)->N1_next=getArch(A1)->N1_prev=K;
			if (_A1 && _A1==n1.Dw) n1.Dw=K;
		}
	}
	else
	{
		a.N1_prev=a.N1_next=0;
	}

	return K;
}


void graph::addArchDirection(key A,Direction dir,key _A0,key _A1)
{
	const key& K=A;
	arch_t& a =*getArch(A);
	node_t& n0=*getNode(a.N0);
	node_t& n1=*getNode(a.N1);

	//important when a.N0_prev=A.N0_next=0 means that the arch exists only for dir=DIRECTION_DOWN
	//               a.N1_prev=A.N1_next=0 means that the arch exists only for dir=DIRECTION_UP

	if (dir!=DIRECTION_DOWN && !a.N0_prev && !a.N0_next)
	{
		if (!(n0.Nup++)) 
			{n0.Up=a.N0_prev=a.N0_next=K;}
		else
		{
			key A0=_A0;
			if (!A0) A0=n0.Up;
			a.N0_next=A0;
			a.N0_prev=getArch(A0)->N0_prev;
			getArch(a.N0_prev)->N0_next=getArch(A0)->N0_prev=K;
			if (_A0 && _A0==n0.Up) n0.Up=K;
		}
	}

	if (dir!=DIRECTION_UP && !a.N1_prev && !a.N1_next)
	{
		if (!(n1.Ndw++)) 
			{n1.Dw=a.N1_prev=a.N1_next=K;}
		else
		{
			key A1=_A1;
			if (!A1) A1=n1.Dw;
			a.N1_next=A1;
			a.N1_prev=getArch(A1)->N1_prev;
			getArch(a.N1_prev)->N1_next=getArch(A1)->N1_prev=K;
			if (_A1 && _A1==n1.Dw) n1.Dw=K;
		}
	}
}

//////////////////////////////////////////////////////////////////////////

void graph::remArch(key Arch,Direction dir)
{
	arch_t&  a=*getArch(Arch);
	node_t& n0=*getNode(a.N0);
	node_t& n1=*getNode(a.N1);

	if (dir & DIRECTION_UP)
	{
		if (!(--n0.Nup))
			{n0.Up=0;}
		else
		{
			if (n0.Up==Arch) n0.Up=a.N0_next;
			getArch(a.N0_prev)->N0_next=a.N0_next;
			getArch(a.N0_next)->N0_prev=a.N0_prev;
		}

		a.N0_prev=a.N0_next=0;
	}
	 
	if (dir & DIRECTION_DOWN) 
	{
		if (!(--n1.Ndw))
			{n1.Dw=0;}
		else
		{
			if (n1.Dw==Arch) n1.Dw=a.N1_next;
			getArch(a.N1_prev)->N1_next=a.N1_next;
			getArch(a.N1_next)->N1_prev=a.N1_prev;
		} 

		a.N1_prev=a.N1_next=0;
	}

	//doesn't link in DOWN and UP direction

	if (!a.N0_next && !a.N1_next)
		arches.release(Arch);
}


//////////////////////////////////////////////////////////////////////////

void graph::remNode(key N)
{
	node_t& n=*getNode(N);

	//togli dalla lista <nodelists>
	uint8 L=n.Level;
	if (n.Prev) getNode(n.Prev)->Next=n.Next;
	if (n.Next) getNode(n.Next)->Prev=n.Prev;

	if (nodelists.v[L]==N)
		nodelists.v[L]=n.Next;

	--nodelists.n[L];

	nodes.release(N);
}




////////////////////////////////////////////////////////////////


bool graph::orderFace(key face)
{
	key N;
	iterator it;

	checkpoint(getLevel(face)==2);

	for (it=goDw(face);!it.end();it++)
	{
		getNode(getFirstDwNode(*it))->Level = 255; //to reach
		getNode(getLastDwNode (*it))->Level = 255;
	}

	key first   = getFirstDwArch(face);
	key tomatch = first;
	key actual  = getNextDwArch(tomatch);
	int nmatched= 0;
	
	while(actual!=tomatch)
	{
		if (N=findCommonNode(getN0(tomatch),getN0(actual),DIRECTION_DOWN))
		{
			if (!getLevel(N)) //already reached
				goto ABORT;

			getNode(N)->Level=0; 
			++nmatched;
			
			if (actual==first) //finished
			{
				if (nmatched!=getNDw(face))
					goto ABORT;

				return true;
			}
			
			swapDwOrder(getNextDwArch(tomatch),actual);
			actual=tomatch=getNextDwArch(tomatch);
		}
		actual=getNextDwArch(actual);
	}

ABORT:
	
	for (it=goDw(face);!it.end();it++)
	{
		getNode(getFirstDwNode(*it))->Level=0; 
		getNode(getLastDwNode (*it))->Level=0; 
	}

	return false;
}


//-------------------------------------------------------------
void graph::check(key N)
{
	iterator i,j;


	key last;

	if (!goDw(N).end())
	{
		for (i=goDw(N);!i.end();i++)
		{
			checkpoint(getArch(i.getArch())->N1==N);
			last=*i;
		}

		key _tmp=getLastDwNode(N);
		checkpoint(last==_tmp);
	}

	if (!goUp(N).end())
	{
		for (i=goUp(N);!i.end();i++)
		{
			checkpoint(getArch(i.getArch())->N0==N);
			last=*i;
		}
		checkpoint(last==getLastUpNode(N));
	}

	key cont;
	for (cont=0,i=goDw(N);!i.end();i++,cont++)
	{
		key M=*i;

		for (j=goUp(M);!j.end();j++)
		{
			if (getN1(j.getArch())==N)
			{
				checkpoint(j.getArch()==i.getArch());
				break;
			}
		}

		checkpoint(!j.end());
	}
	checkpoint(cont==getNDw(N));

	for (cont=0,i=goUp(N);!i.end();i++,cont++)
	{
		key M=*i;

		for (j=goDw(M);!j.end();j++)
		{
			if (getN0(j.getArch())==N)
			{
				checkpoint(j.getArch()==i.getArch());
				break;
			}
		}

		checkpoint(!j.end());
	}
	checkpoint(cont==getNUp(N));
}


//////////////////////////////////////////////////////////////////////////

void graph::check()
{
	for (int k=0;k<256;k++)
	{
		for (list_iterator i=each(k);!i.end();i++)
			check(*i);
	}

}


//-------------------------------------------------------
//-------------------------------------------------------

void graph::next(bool bUp,key& ndstcells,key* dstcells,const key nsrccells,const key* srccells)
{
	checkpoint(dstcells!=srccells);
	key i;

	ndstcells=0;

	for (i=0;i<nsrccells;++i)
	{
		for(graph::iterator it=(bUp)?goUp(srccells[i]):goDw(srccells[i]);!it.end();it++)
			nodeTmp(*it)=0;
	}

	for (i=0;i<nsrccells;++i)
	{
		for(graph::iterator it=(bUp)?goUp(srccells[i]):goDw(srccells[i]);!it.end();it++)
		{
			if ((bUp   && ++nodeTmp(*it)==getNDw(*it)) ||
				 (!bUp  && ++nodeTmp(*it)==1         ))
			{
				dstcells[ndstcells++]=*it;
			}
		}	
	}

	checkpoint(ndstcells<=MAX_CELLS_PER_LEVEL);
}


//-------------------------------------------------------
//-------------------------------------------------------
key graph::findArch(key N0,key N1,Direction dir)
{
	node_t& n0=*getNode(N0);
	node_t& n1=*getNode(N1);

	uint16 i;
	key K;
	arch_t* a;

	if (dir==DIRECTION_UP || (dir==DIRECTION_UP_AND_DOWN && n0.Nup<=n1.Ndw))
	{
		for (i=0,a=getArch(K=n0.Up);i!=n0.Nup;i++,a=getArch(K=a->N0_next))
		{
			if (a->N1==N1) 
				return K;
		}
	}
	else
	{
		for (i=0,a=getArch(K=n1.Dw);i!=n1.Ndw;i++,a=getArch(K=a->N1_next))
		{
			if (a->N0==N0) 
				return K;
		}
	}

	return 0;	
}


//-------------------------------------------------------
//-------------------------------------------------------
key graph::findCommonNode(key N,key M,Direction dir)
{
	if (dir!=DIRECTION_DOWN)
	{
		for (graph::iterator it=goUp(N);!it.end();it++)
		{
			key Up=*it;

			for (graph::iterator jt=goDw(Up);!jt.end();jt++)
			{
				if (*jt==M)
					return Up;
			}
		}
	}

	if (dir!=DIRECTION_UP)
	{
		for (graph::iterator it=goDw(N);!it.end();it++)
		{
			key Dw=*it;

			for (graph::iterator jt=goUp(Dw);!jt.end();jt++)
			{
				if (*jt==M)
					return Dw;
			}
		}
	}

	return 0; //not found
}




//-------------------------------------------------------
//-------------------------------------------------------
key graph::findCommonNodes(key N,key M,key* v,Direction dir)
{
	int cont=0;

	if (dir!=DIRECTION_DOWN)
	{
		for (graph::iterator it=goUp(N);!it.end();it++)
		{
			key Up=*it;
			for (graph::iterator jt=goDw(Up);!jt.end();jt++)
				if (*jt==M) v[cont++]=Up;
		}
	}

	if (dir!=DIRECTION_UP)
	{
		for (graph::iterator it=goDw(N);!it.end();it++)
		{
			key Dw=*it;
			for (graph::iterator jt=goUp(Dw);!jt.end();jt++)
				if (*jt==M) v[cont++]=Dw;
		}
	}

	return cont;
}


//-------------------------------------------------------
//-------------------------------------------------------

void graph::print(char* spaces,FILE* file)
{
	int L;

	fprintf(file,"%s-------------------------------------------------\n",spaces);
	for ( L=255;L>=0;L--)
	{
		int cont=0;
		for (list_iterator it=each(L);!it.end();it++)
		{
			if (!cont) fprintf(file,"%sLevel %d ncells %d\n",spaces,(int)L,(int)nodelists.n[L]);
			key N=*it;
			print(N,spaces,file);
			fprintf(file,")\n");
			++cont;
		}
		if (cont) fprintf(file,"\n");
	}
	fprintf(file,"%s-------------------------------------------------\n",spaces);
}

//-------------------------------------------------------
//-------------------------------------------------------

void graph::print(key N,char* space,FILE* file)
{
	fprintf(file,"%s[%4d] ",space,(int)N);

	fprintf(file,"down(");

	iterator i;

	for (i=goDw(N);!i.end();i++)
	{
		fprintf(file,"%d ",(int)*i);
		//fprintf(file,"*%d* ",(int)archData(findArch(*i,N)));
		//if (bNodeData) fprintf(file,"[%d] ",(int)nodeData(*i));
	}
	fprintf(file,") ");

	fprintf(file," up(");
	for (i=goUp(N);!i.end();i++)
	{
		fprintf(file,"%d ",(int)*i);
		//if (bNodeData) fprintf(file,"[%d] ",(int)nodeData(*i));
	}
	fprintf(file,") Level(%d) ",(int)getNode(N)->Level);

	//fprintf(file,"data(%d) sign(%d)",(int)nodeData(N),(int)getSign(N));


}

//-------------------------------------------------------
//-------------------------------------------------------
void graph::printStatistics(FILE* file)
{
	int L;
	fprintf(file,"ncells\t");

	for ( L=255;L>=0;L--)
	{
		for (list_iterator it=each(L);!it.end();it++)
		{
			fprintf(file,"%d(%d) ",(int)L,nodelists.n[L]);
			break;
		}
	}
	fprintf(file,"\n");

	size_t t=sizeof(graph)+sizeof(nodelists)+nodes.totalmem()+arches.totalmem();
	fprintf(file,"Total memory %dkb \n",(int) t/1024);

	fprintf(file,"Node ");
	nodes.printStatistics(file);

	fprintf(file,"Arch ");
	arches.printStatistics(file);
}




