/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _GVIEWER_H__
#define _GVIEWER_H__

#include "generaldef.h"

#include <vector>

#include "ggraph.h"
#include "bsp.h"
#include "viewer.h"


class XgeDraw;
class BspDraw;


class XgeViewer:public Win
{
	friend class HpcViewer;

	Panel* panel;
	Options* opt;

	std::vector<XgeDraw*> glist;
	std::vector<BspDraw*> blist;
	std::vector<Draw*>    dlist;

	void setMaxDeep(int value);

public: 

	//constructor
	XgeViewer();

	//destructor
	~XgeViewer();

	//acess the panel
	inline Panel* getPanel() {return panel;}

	//keybord
	virtual void keydown(unsigned char key, GLint x, GLint y);
	virtual void special(int, GLint x, GLint y);

	//getviewmode
	int   getViewmode();
	float getExpand();
	float getPointSize();
	void  setPointSize(float value);
	float getLineWidth();
	void  setLineWidth(float value);
	bool  getBBall();

	//add a ggraph
	void push(ggraph* g,float R=0.6f,float G=0.6f,float B=0.6f);

	//add a draw
	void push(XgeDraw* d);

	//add a draw
	void push(Draw* d);


	//add a bsp building a new solver
	ggraph* push(bsp* src,real tolerance=1e-4,int maxdeep=0,real limit=1e4);
}; 



class XgeDraw: public Draw
{
	ggraph* g;

public:

	float R,G,B;
	float linewidth;
	bool nonconvex;
	bool view_face_normal;
	int8 view_cell_level;

	//constructor
	XgeDraw(ggraph* g,float R=0.5,float G=0.5,float B=0.5,float linewidth=1);

	//display
	virtual void display();

	static void display(ggraph* g,int8 view_cell_level,float R,float G,float B,
		real expandx=0,real expandy=0,real expandz=0,bool nonconvex=false,
		float pointsize=1,float linewidth=1,bool bball=false);

	inline void setGraph(ggraph* g)
	{
		this->g=g;
	}

};

class XgeDrawDebug: public Draw
{
	typedef struct
	{
		ggraph* g;
		key Cell;
		float R,G,B;
	} 
	TCell;

	typedef struct
	{
		real v0[MAX_GEOMETRIC_DIMENSION+1];
		real v1[MAX_GEOMETRIC_DIMENSION+1];
		float R,G,B;
	}
	TLine;

	typedef struct
	{
		real v0[MAX_GEOMETRIC_DIMENSION+1];
		float R,G,B;
	}
	TPoint;

	std::vector<TCell > cells;
	std::vector<TLine > lines;
	std::vector<TPoint> points;

public:

	//constructor
	XgeDrawDebug();
	
	//add a cells to draw
	void add(ggraph* g,key Cell,float R=0.6,float G=0.6,float B=0.6);

	//add a point
	void add(int dim,real* v0,float R=0.6,float G=0.6,float B=0.6);

	//add a line
	void add(int dim,real* v0,real* v1,float R=0.6,float G=0.6,float B=0.6);

	//display
	virtual void display();
};

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

class BspDraw: public XgeDraw
{
	bsp_solver* solver;

public:

	inline BspDraw(bsp_solver* solver,real tolerance=1e-3,int maxdeep=0)
		:XgeDraw(solver->getGraph())
	{
		this->solver=solver;
	}

	inline bsp_solver* getSolver()
		{return solver;}

};


/////////////////////////////////////////////////////////////////////////////////////////////
//COUNTOUR 2 (working only on 3d-triangles)
/////////////////////////////////////////////////////////////////////////////////////////////

class XgeContour2: public Draw
{
	ggraph* g;
	bool changing_isovalue;
	real isovalue;

public:

	//constructor
	XgeContour2(ggraph* g);

	//mouse callback
	virtual void mouse(GLint button, GLint state, GLint x, GLint y) ;

	void display();

	//motion callback
	virtual void motion(GLint x, GLint y);

	int8 vsign(key N);

	int8 vless(key A,key B);

	ggraph getContour() ;


};


/////////////////////////////////////////////////////////////////////////////////////////////
//COUNTOUR 3 (working only on 3d-tet)
/////////////////////////////////////////////////////////////////////////////////////////////

class XgeContour3: public Draw
{
	ggraph* g;
	bool changing_isovalue;
	real isovalue;

public:

	//constructor
	XgeContour3(ggraph* g);

	//mouse callback
	virtual void mouse(GLint button, GLint state, GLint x, GLint y);

	void display();

	//motion callback
	virtual void motion(GLint x, GLint y);

	int8 vsign(key N);

	int8 vless(key A,key B);

	ggraph getContour();

};




#endif // _GVIEWER_H__

