#include "generaldef.h"

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>


#include "memcache.h"
#include "graph.h"
#include "ggraph.h"
#include "viewer.h"
#include "bsp.h"
#include "gbool.h"
#include "gviewer.h"
#include "mat.h"


//==========================================================================


inline void view(ggraph* g)
{
	XgeViewer win;
	win.setPointSize(10);
	win.setLineWidth(6);
	win.push(g);
	win.setPointSize(5);
	win.run();
}

inline void view(bsp* b)
{
	XgeViewer win;
	win.setPointSize(10);
	win.setLineWidth(6);
	win.push(b);
	win.run();
}

inline ggraph* solve(bsp& b)
{
	bsp_solver solver(&b);
	solver.visit();
	return (solver.getGraph()->clone());
}


////////////////////////////////////////////////////////////////////////////////
static void memcache_test()
{
	memcache d(sizeof(float[4]));
	clock_t t1=clock();

	key nloop		= 150000;
	key nlooses		= 100;
	key nperloop	= 1000;

	for (key i=0;i<nloop;i++)
	{
		key K[1000];

		for (key j=0;j<nperloop;j++)
		{
			K[j]=d.block();
			float* v=(float*)d[K[j]];
			v[0]=0;
			v[1]=1;
			v[2]=2;
			v[3]=3;
		}
		for (key k=nlooses;k<nperloop;k++)
		{
			d.release(K[k]);
		}
	}
	checkpoint(nlooses*nloop==d.getNBlocked());
	d.printStatistics();
	clock_t t2=clock();

	printf("time %.2f\n",(float)(t2 - t1) / CLOCKS_PER_SEC);
	printf("");
}



//-------------------------------------------------------
//-------------------------------------------------------

static key graph_test_speed(graph& g,int maxrec,key& nop,key& narch,key& nnode)
{
	key N=g.addNode(0);
	++nop;
	++nnode;

	if (maxrec--)
	{
		for (int i=0;i<6;i++)
		{
			key Up=graph_test_speed(g,maxrec,nop,narch,nnode);

			g.addArch(N,Up);
			++nop;
			++narch;
			
			if (!(i%2))
			{
				g.remArch(N,Up);++nop;--narch;
				//g.addArch(N,Up);++nop;++narch;
			}
		}
	}	
		
	return N;

}

void testEmbeddingPower(bool bView)
{
	printf("testing power with embedding\n");
	for (int R=0;R<2;R++)
	{
		for (int D=0;D<2;D++)
		{
			real vt2[]={sqrt((float)2)/2,sqrt((float)2)/2};
			ggraph* g=solve(bsp_circle(2,2));
			g->translate(vt2);
			int num;ggraph** v=g->skeleton(D,num);

			XgeViewer* win=0;

			if (bView)
			{
				win=new XgeViewer;
				win->setPointSize(10);
				win->setLineWidth(6);
			}

			real vt1[]={-0.5};
			ggraph* c2=ggraph::cuboid(1);
			c2->translate(vt1);
			ggraph* u[256];
			int i;

			for (i=0;i<num;i++)
			{
				if (!R)
					u[i]=ggraph::power(v[i],v[i]->spacedim,v[i]->vmat,v[i]->hmat,c2,c2->spacedim,c2->vmat,c2->hmat,1e-6f,true,false);
				else
					u[i]=ggraph::power(c2,c2->spacedim,c2->vmat,c2->hmat,v[i],v[i]->spacedim,v[i]->vmat,v[i]->hmat,1e-6f,true,false);

#ifdef _DEBUG
				u[i]->check();
#endif

				if (win) win->push(u[i]);
			}

			if (win)
			{
				win->run();
				delete win;
			}

			for (i=0;i<num;i++) 
			{
				delete v[i];
				delete u[i];
			}
			delete [] v;
			delete c2;
			delete g;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
static void graph_test()
{

	//first test
	{
		graph g;
		key n=g.addNode(0);
		key m=g.addNode(1);
		key p=g.addNode(1);

		checkpoint(g.getNumNode()==3);

		g.addArch(n,m);
		g.addArch(n,p);
		g.remArch(n,m);
		g.remArch(n,p);

		checkpoint(g.getNumArch()==0);

		checkpoint(g.getNUp(n)==0 && g.getNDw(m)==0 && g.getNDw(p)==0);
		g.remNode(n);
		g.remNode(m);
		g.remNode(p);

		checkpoint(g.getNumNode()==0);
	}

	//test memory
	{
		graph  g;

		const int Max=1000000;
		int i;
		key* nodes=new key[Max];

		for (i=0;i<Max;i++)
		{
			nodes[i]=g.addNode(0);

			if (i) 
				g.addArch(nodes[i-1],nodes[i]);
		}

		checkpoint(g.getNumNode()==Max);

		key maxnode=g.getMaxNode();
		key maxarch=g.getMaxArch();
		
		for (i=0;i<Max;i++)
			if (i) g.remArch(nodes[i-1],nodes[i]);

		for (i=0;i<Max;i++)
			g.remNode(nodes[i]);

		checkpoint(g.getNumNode()==0 && g.getNumArch()==0);

		for (i=0;i<Max;i++)
		{
			nodes[i]=g.addNode(0);
			if (i) g.addArch(nodes[i-1],nodes[i]);
		}

		checkpoint(maxnode==g.getMaxNode() && maxarch==g.getMaxArch());
	}

	//speed test
	{
		const int MAXREC=8;
		clock_t start = clock();
		graph g;
		key narch=0,nnode=0,nop=0;
		graph_test_speed(g,MAXREC,nop,narch,nnode);

		checkpoint(g.getNumArch()==narch && g.getNumNode()==nnode);

		clock_t finish = clock();
		double duration = (double)(finish - start) / CLOCKS_PER_SEC;
		printf( "%2.1f seconds narch=%d nnode=%d (%d operation/sec)\n", duration,(int)narch,(int)nnode,(int)(nop/duration));
	}
}


//==================================================================================

#ifdef _DEBUG
#define CHECK(g) g->check()
#else
#define CHECK(g)
#endif

void test_xge(int Cont=1,bool bView=true)
{



REDO:

	uint8 gridv2[]={0,1,0,1,-1};
	uint8 gridv3[]={0,1,2,0,1,2,-1};
	
	printf("testing builders\n");
	{
		ggraph* g=ggraph::cuboid(3);
		CHECK(g);
		if (bView) view(g);
		delete g;
	}
	
	{
		ggraph* g=ggraph::simplex(3);
		CHECK(g);
		if (bView) view(g);
		delete g;
	}
	{
		ggraph* g=ggraph::grid(2,gridv2);
		CHECK(g);
		if (bView) view(g);
		delete g;
	}
	{
		ggraph* g=ggraph::grid(3,gridv3);
		CHECK(g);
		if (bView) view(g);
		delete g;
	}
	
	{
		printf("testing transformations\n");

		real vt3[16],ht3[16],vtt[16],htt[16];
		tmat<real>::identity(4,vtt);
		tmat<real>::identity(4,htt);

		real vshear3[]={1,0.1f,0.2f}   ;
		real vtrans3[]={0.1f,0.2f,0.3f};
		real vscale3[]={1,2,3};
		real vangle3=3.14f/6;

		tmat<real>::shear    (4,1,vshear3,vt3,ht3)   ;tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3);
		tmat<real>::scale    (4,1,vscale3[0],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 
		tmat<real>::scale    (4,2,vscale3[1],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 
		tmat<real>::scale    (4,3,vscale3[2],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 
		//tmat<real>::rotate   (4,1,2,vangle3         );tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 
		tmat<real>::translate(4,1,vtrans3[0],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3);
		tmat<real>::translate(4,2,vtrans3[1],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3);   
		tmat<real>::translate(4,3,vtrans3[2],vt3,ht3);tmat<real>::product(4,vtt,vt3,vtt);tmat<real>::product(4,htt,htt,ht3); 

		ggraph* g=ggraph::cuboid(3);
		g->shear(1,vshear3);
		g->scale(vscale3);
		//g->rotate(1,2,vangle3);
		g->translate(vtrans3);
		g->transform(htt,vtt); //reverse the transformation
		CHECK(g);
		if (bView) view(g);
		delete g;
	}

	printf("testing embed and permutations\n");
	{
		real vt3[]={0,0,1};
		ggraph* g=ggraph::cuboid(2);
		g->embed(3);
		g->translate(vt3);
		CHECK(g);
		if(bView) view(g);
		delete g;

		g=ggraph::cuboid(2);
		g->embed(0,0.5);
		CHECK(g);
		if (bView) view(g);
		delete g;

		int p3[]={0,1,3,2};
		g=ggraph::cuboid(2);
		g->embed(3);
		g->permutate(p3);
		CHECK(g);
		if (bView) view(g);
		delete g;
	}

	printf("testing bsp\n");
	{
		bsp* b=new bsp_circle(16);
		if (bView) view(b);
		delete b;

		b=new bsp_circle(4,4);
		if (bView) view(b);
		delete b;

		b=new bsp_sphere(16);
		if (bView) view(b);
		delete b;

		b=new bsp_sphere(2,4);
		if (bView) view(b);
		delete b;

		b=new bsp_grid(2,gridv2);
		if (bView) view(b);
		delete b;

		b=new bsp_grid(3,gridv3);
		if (bView) view(b);
		delete b;
	}

	{
		printf("testing boolean operations\n");
		key vmax3[]={1,1,1};
		bsp_sphere s1(2,2),s2(2,2);
		real vt3[]={0.3f,0.8f,0.3f};
		bsp_transform l1(TRANSLATE,&s2,vt3);
		bsp* args[3]={&s1,&l1,0};

		bsp* b=new bsp_bool(bsp_bool::BOOL_AND ,args,vmax3);
		if (bView) view(b);
		delete b;

		b=new bsp_bool(bsp_bool::BOOL_OR  ,args,vmax3);
		if (bView) view(b);
		delete b;

		b=new bsp_bool(bsp_bool::BOOL_DIFF,args,vmax3);
		if (bView) view(b);
		delete b;
	}

	{
		printf("testing permutations\n");
		ggraph* g=solve(bsp_circle(2,2));
		g->embed(3);
		int perm3[]={0,1,3,2};
		g->permutate(perm3);
		CHECK(g);
		if (bView) view(g);
		delete g;
	}

	{
		printf("testing import\n");
		ggraph* g=ggraph::read("D:\\demos\\data\\fields\\3d\\engine.txt");
		CHECK(g);
		if (bView) view(g);
		delete g;
	}

	{
		printf("testing triangulation\n");

		ggraph* g=ggraph::grid(2,gridv2);
		g->embed(0,1);
		CHECK(g);
		if (bView) view(g);
		delete g;

		g=solve(bsp_cube(2));
		g->triangulate();
		CHECK(g);
		if (bView) view(g);
		delete g;

		g=solve(bsp_cube(3));
		g->triangulate();
		CHECK(g);
		if (bView) view(g);
		delete g;
	}

	//mkpol
	{
		printf("testing mkpol\n");
		int coeff[4][3]={{0,0,0},{1,0,0},{1,1,0},{1,1,1}};

		for (int i=0;i<4;i++)
		{
			for (int j=0;j<2;j++) //symbolic or not
			{
				ggraph* gpoints=new ggraph(3);
				const int na=20,nb=20,nc=10;			
				for (int r=0;r<nc;r++)
				for (int a=0;a<na;a++)
				for (int b=0;b<nb;b++)
				{
					real radius  =r/(float)(nc-1);
					real angle_a =2*(3.14159)*a/(na-1);
					real angle_b =(3.14159)*b/(nb-1) -(3.14159)/2;

					real vertex[]=
					{
						1,
						coeff[i][0]*radius*cos(angle_a)*cos(angle_b),
						coeff[i][1]*radius*sin(angle_a)*cos(angle_b),
						coeff[i][2]*radius*sin(angle_b)
					};

					gpoints->addVertex(vertex);
				}
				
				real vt3[]={0,0,0.5};
				gpoints->translate(vt3);
				gpoints->rotate(1,3,(real)(3.14/6));

				key tmp[MAX_CELLS_PER_LEVEL];

				int npoints=gpoints->findCells(0,0,tmp);
				ggraph* m=gpoints->mkpol(npoints,tmp,j?true:false) ;
				checkpoint(m->pointdim==i && m->spacedim==3);

				CHECK(m);
				if (bView) view(m);
				delete m;
				delete gpoints;
			}
		}
	}

	{
		printf("testing gbool\n");
		for (int dim=2;dim<=3;dim++)
		{
			XgeViewer* win=0;

			if (bView) win=new XgeViewer;

			const int ngs=10;
			int i;
			ggraph* gs[ngs];
			
			for ( i=0;i<ngs;i++)
			{
				float radius=dim==3?0.8:0.1;
				float rdim  =radius*rand()/(real)RAND_MAX;

				if (rdim<(radius/10)) rdim=radius/10;

				if (dim==2)
					gs[i]=solve(bsp_circle(2,2));
				else
					gs[i]=solve(bsp_sphere(2,2));
				
				real vs[3]={rdim/2,rdim/2,rdim/2};
				real vt[3]={(-0.5+rand()/(real)RAND_MAX),(-0.5+rand()/(real)RAND_MAX),(-0.5+rand()/(real)RAND_MAX)};
				
				gs[i]->scale(vs);
				gs[i]->translate(vt);
				gs[i]->rotate(1,2,3.14*rand()/(real)RAND_MAX);
				
				if (win && dim==2) win->push(gs[i],1,1,0);
			}
			
			GBool* gbool=new GBool(dim);
			key Up=gbool->Union();	
			for (i=0;i<ngs;i++) gbool->addArg(gbool->add(gs[i]),Up);			
			
			gbool->setRoot(Up);
			gbool->visit();

			if (win)  win->push(gbool);
			
			if (dim==2)
			{
				for (i=0;i<ngs;i++)
				{
					real vt[]={0,0,+0.1f};
					gs[i]->embed(3);
					gs[i]->translate(vt);
				}
			}

			if (win) 
			{
				win->run();
				delete win;
			}

			delete gbool;
			for (i=0;i<ngs;i++) delete gs[i];
		}
	}

	{
		printf("testing skeleton\n");
		int i,j,num;
		
		for (i=0;i<=3;i++)
		{
			XgeViewer* win=0;

			ggraph* c2=ggraph::cuboid(3);

			if (bView)
			{
				win=new XgeViewer;
				win->setPointSize(10);
				win->setLineWidth(6);
			}

			ggraph** v=c2->skeleton(i,num);
			for (j=0;j<num;j++) 
			{
				CHECK(v[j]);
				if (win) win->push(v[j]);
			}

			if (win)
			{
				win->run();
				delete win;
			}

			delete c2;
			for (j=0;j<num;j++) delete v[j];
			delete [] v;
		}
	}

	{
		printf("testing power\n");
		ggraph* c1=solve(bsp_circle(8));
		ggraph* c2=ggraph::cuboid(1);
		ggraph* p=ggraph::power(c1,c1->spacedim,c1->vmat,c1->hmat,c2,c2->spacedim,c2->vmat,c2->hmat,1e-6f,true,false);
		CHECK(p);
		if (bView) view(p);
		delete c1;
		delete c2;
		delete p;
	}

	testEmbeddingPower(bView);

	Cont--;
	if (Cont) goto REDO;
}

//=========================================
//=========================================

#ifdef WIN32
#pragma warning(disable:4305)
#endif
int main(int argn,char** argv)
{

#if 0
	memcache_test();
	graph_test();
#endif


	test_xge(1,true);
	

#if defined(WIN32) && defined(_DEBUG)
	printf("press a char to exit\n");
	getchar();
#endif


	return 0;
}

