/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#include <time.h>

#include "memcache.h"


memcache::memcache(int itemsize)
{
	//I need space for next
	checkpoint(itemsize>=sizeof(key));
	this->itemsize=itemsize;
	this->mem=0;

	clear();
}

memcache::memcache(const memcache& src)
{
	this->itemsize=src.itemsize;
	this->max     =src.max;
	this->num     =src.num;
	this->ff      =src.ff;
	this->mem=malloc(itemsize*this->max);
	memcpy(this->mem,src.mem,itemsize*this->max);
}

memcache* memcache::clone()
{
	return new memcache(*this);
}

memcache& memcache::operator=(const memcache& src)
{
	this->itemsize=src.itemsize;
	
	this->max     =src.max;
	this->num     =src.num;
	this->ff      =src.ff;
	this->mem=realloc(this->mem,itemsize*this->max);
	if (!this->mem) {fprintf(stderr,"fatal error in %s line %d. Realloc failed. Requested %dmb\n",__FILE__,__LINE__,(int)(itemsize*max/(1024*1024)));checkpoint(0);exit(-1);}
	memcpy(this->mem,src.mem,itemsize*this->max);
	return *this;
}

memcache::~memcache()
{
	free(this->mem);
	this->mem=0;
}


void memcache::clear()
{
	if (this->mem) free(this->mem);
	this->mem=malloc(itemsize*2);
	max = 2;num = 1;ff  = 1;
	*((key*)(((unsigned char*)mem)+1*itemsize))=0;
}


void memcache::resize(int new_itemsize)
{
	if (new_itemsize==itemsize)
		return;

	checkpoint(new_itemsize>=sizeof(key));

	unsigned char* new_mem=(unsigned char*)malloc(new_itemsize*max);
	memset(new_mem,0,new_itemsize*max);

	unsigned char* p1=(unsigned char*)this->mem;
	unsigned char* p2=(unsigned char*)new_mem;

	int minsize=min2(itemsize,new_itemsize);

	for (key i=0;i<max;i++)
	{
		memcpy(p2,p1,minsize);
		p1+=    itemsize;
		p2+=new_itemsize;
	}

	free(this->mem);

	this->itemsize=new_itemsize;
	this->mem=new_mem;
}

void memcache::printStatistics(FILE* file)
{
	fprintf(file,"n=%d m=%d (%d%%) single=%d bytes tot=%d kb\n",
		(int)num-1,(int)max,(int)(num*100.f)/max,(int)itemsize,(int)(totalmem()/1024.0f));
}



