/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _MEMCACHE_H__
#define _MEMCACHE_H__


#include "generaldef.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//====================================================
// Class for fast memory allocation
/*
Example:

	key Ki,Kj;
	memcache m(sizeof(int));
	Ki=m.block();
	*((int*)m[Ki])=10;
	Kj=m.block();
	*((int*)m[Kj])=30;
	m.release(Ki);
	m.release(Kj);
*/
//====================================================

class memcache
{
	int            itemsize;
	void*          mem;
	key            max,num,ff;

public:

	//constructor
	memcache(int itemsize);

	//copy constructor
	memcache(const memcache& src);

	//destructor
	~memcache();

	//copy
	memcache& operator=(const memcache& src);

	//fast copy
	memcache* clone();

	//clear memory
	void clear();

	//access an item 
	void* operator[](key K);

	//get (and block) a new item
	key block();

	//release a previous blocked item
	void release(key K);

	//make sure there is enough memory
	//<n> is the number of items I need to block
	//if you do not call memory every time you "block" a memory(1) is automatically called
	//the memory(<n>) speed up the allocation for a bunch of items
	void memory(key n);

	//number of elements currenty blocked
	key getNBlocked();

	//number of max elements
	key getMax();

	//size of single item
	size_t size() const;

	//resize 
	void resize(int new_itemsize);

	//total memory 
	inline size_t totalmem() const;

	//print statistics
	void printStatistics(FILE* file=stdout);


}; //end class



inline key		memcache::getMax()				{return max;  }
inline key		memcache::getNBlocked()			{return num-1;} //do not count item 0
inline void*	memcache::operator[](key K)	{return (void*)(((unsigned char*)mem)+K*itemsize);}
inline size_t	memcache::size() const			{return itemsize;}


inline key memcache::block()
{
	if (max==num) memory(1);
	key K=ff;
	ff=*((key*)(((unsigned char*)mem)+K*itemsize));
	++num;
	return K;
}

inline void memcache::release(key K)
{
	*((key*)(((unsigned char*)mem)+K*itemsize))=ff;
	ff=K;
	--num;
}

inline void memcache::memory(key n)
{
	if ((max-num)>=n) return;
	key oldsize=max;
	if (!max) max=n;
	while ((max-num)<n) max<<=1;
	max=max2(32,max);
	this->mem=realloc(this->mem,itemsize*max);
	if (!this->mem) {fprintf(stderr,"fatal error in %s line %d. Realloc failed. Requested %dmb\n",__FILE__,__LINE__,(int)(itemsize*max/(1024*1024)));assert(0);exit(-1);}
	for (key i=oldsize;i<max-1;i++) *((key*)(((unsigned char*)mem)+i*itemsize))=(i+1);
	*((key*)(((unsigned char*)mem)+(max-1)*itemsize))=ff;
	ff=oldsize;
}

inline size_t memcache::totalmem() const
	{return sizeof(size_t)+itemsize*max;}


#endif //_MEMCACHE_H__

