/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#include "generaldef.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <algorithm>
#include <functional>
#include <assert.h>

#include "partball.h"

///////////////////////////////////////////////////////////////////////////

static inline real Fn1(real arg) //Maple: plot((2^(-32*x)),x=0..1) 
{
	return (real)(pow(2,-32*arg));
}

static inline real Fn2(real arg) //Maple: plot(cos(Pi*x+Pi/2)^2 , x=0..1)
{
	return (real)pow(cos(PIGRECO*arg+PIGRECO/2),2);
}


///////////////////////////////////////////////////////////////////////////
PartitionBBall::PartitionBBall(int dim)
{
	this->dim=dim;
	this->bballs=0;
	this->maxbballs=this->nbballs=0;
}

///////////////////////////////////////////////////////////////////////////
PartitionBBall::~PartitionBBall()
{
	if (bballs) 
		delete [] bballs;
}

///////////////////////////////////////////////////////////////////////////
void PartitionBBall::addBBall(real* bball)
{
	if (nbballs==maxbballs)
	{
		maxbballs=(int)max2(maxbballs*1.5,nbballs+1);;
		bballs=(real*)realloc(bballs,maxbballs*sizeof(real)*(dim+1));
		assert(bballs);
	}
	
	memcpy(bballs+nbballs*(dim+1),bball,sizeof(real)*(dim+1));
	nbballs++;
}

///////////////////////////////////////////////////////////////////////////
void PartitionBBall::addDirection(real* dir)
{
	Solution dirsol;
	memset(&dirsol,0,sizeof(Solution));
	memcpy(dirsol.dir,dir,sizeof(real)*(dim+1));
	
	int ntotal=nbballs;
	real m=(real)+1e18;
	real M=(real)-1e18;
	
	std::vector<event_t> W;
	std::less<event_t> comp;
	
	//project all bounding balls onto the current direction (==intertia tensor)
	for (int I=0;I<nbballs;I++)
	{				
		real  *bball=&bballs[I*(dim+1)];
		real  radius=sqrt(bball[0]);
		real  proj=0;
		
		for (int i=1;i<=dim;i++) 
			proj+=bball[i]*dirsol.dir[i];
		
		m=min2(m,proj-radius);
		M=max2(M,proj+radius);
		
		W.push_back(event_t(proj-radius,false));std::push_heap(W.begin(), W.end(), comp); 
		W.push_back(event_t(proj+radius,true ));std::push_heap(W.begin(), W.end(), comp); 
	}
	
	real last= (real)(m - (1e-4));
	
	Solution cursol=dirsol;
	cursol.score=0;
	cursol.nbelow=cursol.nzero=0;
	cursol.nabove=nbballs;
	
	while (true)
	{
		int total=nbballs;
		
		real where=(real)(W.size() ? W.front().where:M+1e-4);
		real middle=(last+where)/2;
		
		int newtotalbelow=cursol.nbelow+cursol.nzero;
		int newtotalabove=cursol.nabove+cursol.nzero;
		
		cursol.score = 
			0 
			//malus if the difference for the new total is too small
			//+(((newtotalabove)==(real)total)?-1e4:0)
			//+(((newtotalbelow)==(real)total)?-1e4:0)
			//+(((newtotalabove+newtotalbelow)/total>=1.5)?-1e4:0)
			
			//how many splits (best score 1.0 at nzero=0)
			+ 1.0 * Fn1(cursol.nzero/(real)ntotal)			

			//balance between below and above (best score 1.0 at nbelow=nabove)
			+ 1.0 * Fn2((cursol.nbelow)/(real)(cursol.nbelow+cursol.nabove))	

			//split the arch possibly in the middle (best score 1.0 at middle=0.5*(m+M) )	
			//+ 1.0 * Fn2((middle-m)/(M-m))

			;					
		
		cursol.dir[0]=-middle;	
		
		if (cursol.score > dirsol.score) dirsol=cursol;
		
		if (!W.size()) break;
		
		if (!(W.front().closure))	{++cursol.nzero;--cursol.nabove;}
		else	       	            {--cursol.nzero;++cursol.nbelow;}
		
		std::pop_heap(W.begin(), W.end(), comp);
		W.pop_back();
		
		last=where;
	}
	
	assert(!cursol.nzero && !cursol.nabove && cursol.nbelow==nbballs);
	
	if (dirsol.score > globsol.score)
		globsol=dirsol;
	
}


