/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _PARTITION_BALL_H__
#define _PARTITION_BALL_HH__

#include "generaldef.h"

#include <string.h>

/* 
find the best way to partition some bounding balls.
Ex

   //first add the balls
	partball->addBall(..);
	partball->addBall(..);
	...

   //then add the direction
	partball->addDirection(...);
	partball->addDirection(...);
	...

   // finally access the solution
	if (partball.getScore()>0)
	{
		real* h=partball.getSolution();
		...
	}
*/

class PartitionBBall
{
	struct event_t
	{
		real where;
		bool closure;
		
		inline event_t(real _where,bool _closure)
			:where(_where),closure(_closure) {}
		
		inline bool operator < (event_t const &src) const  
			{return (this->where>src.where) || (this->where==src.where && this->closure);}
	};

	typedef struct 
	{
		real dir[MAX_GEOMETRIC_DIMENSION+1];
		int nbelow,nzero,nabove;
		real score;
	} 
	Solution;

	Solution globsol;

	int    dim;
	real*  bballs;
	int    nbballs,maxbballs;

public:

	//constructor
	PartitionBBall(int dim);

	//destructor
	~PartitionBBall();

	//restart
	void reset();

	//add a bounding ball
	void addBBall(real* bball);

	//add a direction
	void addDirection(real* dir);

	//get the score and the solution
	real  getScore();
	real* getSolution();

};

// INLINES 


inline void PartitionBBall::reset()
{
	this->nbballs=0;
	memset(&globsol,0,sizeof(Solution));
}

inline real PartitionBBall::getScore()
	{return globsol.score;}

inline real* PartitionBBall::getSolution()
	{return globsol.dir;}


#endif // _PARTITION_BALL_HH__

