#include "generaldef.h"

#include <time.h>

#include "rayshooting.h"


//------------------------------------------------------------------------------
RayShooting2d::RayShooting2d(real vx,real vy,const ggraph* _g)
{
	this->totseconds=0;
	this->vx=vx;
	this->vy=vy;

	assert(_g->pointdim==2 && _g->spacedim==2);

	if (!vy)
	{
		fprintf(stderr,"error, in rayshooting vy must be not equal to zero!");
		assert(vy!=0);
	}

	this->g=_g->clone();

	ggraph::list_iterator it;

	//transform all points
	srand(0);

	for (it=g->each(0);!it.end();it++)
	{
		real* v=g->getGeom(*it);

		//perturbation
		v[1]+=0.001f*(2*((rand()/(float)RAND_MAX)-0.5f));
		v[2]+=0.001f*(2*((rand()/(float)RAND_MAX)-0.5f));

		transformPoint(v[1],v[2]);
	}

	//delete bounding balls
	for (it=g->each(2);!it.end();it++)
		g->releaseGeom(*it);
}

RayShooting2d::~RayShooting2d()
{
	printf("total time in ray shooting %.2f seconds\n",totseconds);
	delete g;
}


//------------------------------------------------------------------------------
key RayShooting2d::getCell(real X,real Y)
{
	clock_t t1=clock();
	transformPoint(X,Y);
	
	int nintersection;
	
	for (ggraph::list_iterator fit=g->each(2);!fit.end();fit++)
	{
		key F=*fit;
		real *bb=g->boundingBall(F);
		real radius2=bb[0];
		real bx=bb[1];
		real by=bb[2];
		
		//outside the bounding ball!
		if (radius2 && (((X-bx)*(X-bx)+(Y-by)*(Y-by))>radius2))
			continue;
		
		nintersection=0;
		
		for (ggraph::iterator it=g->goDw(F);!it.end();it++)
		{
			key _E=*it;
			key _V1=g->getFirstDwNode(_E);real* _v1=g->getGeom(_V1);
			key _V2=g->getLastDwNode( _E);real* _v2=g->getGeom(_V2);
			
			real x1=_v1[1],y1=_v1[2];
			real x2=_v2[1],y2=_v2[2];
			
			//must be x1<x2
			if (x1>x2)
			{
				Swap(real ,x1,x2);
				Swap(real ,y1,y2);
			}
			
			if (!((x1<X) && (x2>=X)))
				continue;
			
			if (y1<Y && y2<Y)
				continue;
			
			if (y1>=Y && y2>=Y)
			{
				nintersection++;
				continue;
			}
			
			if ((y2-y1)*(X-x1)>=(x2-x1)*(Y-y1))
			{
				nintersection++;
				continue;
			}
		}
		
		if (nintersection % 2)
		{
			totseconds+=(clock()-t1)/(float)CLOCKS_PER_SEC;
			return F;
		}
	}
	
	totseconds+=(clock()-t1)/(float)CLOCKS_PER_SEC;
	return 0;
}

