#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <math.h>

static int XMAX=0;
static int YMAX=0;

static int bVisible;

/* max #sides to a polygon; change if needed */
/* note that poly_clip, given an n-gon as input, might output an (n+6)gon */
/* POLY_NMAX=10 is thus appropriate if input polygons are triangles or quads */

#define POLY_NMAX 100

static float* cbuffer=0;
static float* zbuffer=0;


/* A POLYGON VERTEX */
typedef struct 
{		
    double sx, sy, sz, sw;	/* screen space position (sometimes homo.) */
    double x, y, z;		/* world space position */
    double u, v, q;		/* texture position (sometimes homogeneous) */
    double r, g, b;		/* (red,green,blue) color */
    double nx, ny, nz;		/* world space normal vector */
} 
Poly_vert;

/* update poly.c if you change this structure */
/* note: don't put > 32 doubles in Poly_vert struct, or mask will overflow */


/* A POLYGON */
typedef struct 
{		
    int n;			/* number of sides */
    unsigned long mask;		/* interpolation mask for vertex elems */
    Poly_vert vert[POLY_NMAX];	/* vertices */
} 
Poly;

/*
 * mask is an interpolation mask whose kth bit indicates whether the kth
 * double in a Poly_vert is relevant.
 * For example, if the valid attributes are sx, sy, and sz, then set
 *	mask = POLY_MASK(sx) | POLY_MASK(sy) | POLY_MASK(sz);
 */

/* A BOX (TYPICALLY IN SCREEN SPACE) */
typedef struct 
{		
    double x0, x1;		/* left and right */
    double y0, y1;		/* top and bottom */
    double z0, z1;		/* near and far */
} 
Poly_box;


/* WINDOW: A DISCRETE 2-D RECTANGLE */
typedef struct 
{		
    int x0, y0;			/* xmin and ymin */
    int x1, y1;			/* xmax and ymax (inclusive) */
} 
Window;

#define POLY_MASK(elem) (1 << (&poly_dummy->elem - (double *)poly_dummy))

static Poly_vert *poly_dummy;		/* used superficially by POLY_MASK macro */

/*
-----------------------------------------------------------------------------
* incrementalize_y: put intersection of line Y=y+.5 with edge between points
* p1 and p2 in p, put change with respect to y in dp
-----------------------------------------------------------------------------
*/

static incrementalize_y(p1, p2, p, dp, y, mask)
register double *p1, *p2, *p, *dp;
register unsigned long mask;
int y;
{
	double dy, frac;
	
	dy = ((Poly_vert *)p2)->sy - ((Poly_vert *)p1)->sy;
	if (dy==0.) dy = 1.;
	frac = y+.5 - ((Poly_vert *)p1)->sy;
	
	for (; mask!=0; mask>>=1, p1++, p2++, p++, dp++)
	{
		if (mask&1) {
			*dp = (*p2-*p1)/dy;
			*p = *p1+*dp*frac;
		}
	}
}

/*
-----------------------------------------------------------------------------
* incrementalize_x: put intersection of line X=x+.5 with edge between points
* p1 and p2 in p, put change with respect to x in dp
-----------------------------------------------------------------------------
*/

static incrementalize_x(p1, p2, p, dp, x, mask)
register double *p1, *p2, *p, *dp;
register unsigned long mask;
int x;
{
	double dx, frac;
	
	dx = ((Poly_vert *)p2)->sx - ((Poly_vert *)p1)->sx;
	if (dx==0.) dx = 1.;
	frac = x+.5 - ((Poly_vert *)p1)->sx;
	
	for (; mask!=0; mask>>=1, p1++, p2++, p++, dp++)
	{
		if (mask&1) {
			*dp = (*p2-*p1)/dx;
			*p = *p1+*dp*frac;
		}
	}
}

/*
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
*/
static increment(p, dp, mask)
register double *p, *dp;
register unsigned long mask;
{
	for (; mask!=0; mask>>=1, p++, dp++)
	{
		if (mask&1) *p += *dp;
	}
}

/*
-----------------------------------------------------------------------------
scanline: output scanline by sampling polygon at Y=y+.5 
-----------------------------------------------------------------------------
*/
static void scanline(int y, Poly_vert *l, Poly_vert *r, Window *win, unsigned long  mask)
{
	int x, lx, rx;
	Poly_vert p, dp;
	
	mask &= ~POLY_MASK(sx);		/* stop interpolating screen x */
	
	lx = (int)ceil(l->sx-.5);
	
	if (lx<win->x0) 
		lx = win->x0;
	
	rx = (int)floor(r->sx-.5);
	
	if (rx>win->x1) 
		rx = win->x1;

	if (lx>rx) 
		return;
	
	incrementalize_x(l, r, &p, &dp, lx, mask);

	/* scan in x, generating pixels */
	for (x=lx; x<=rx; x++) 
	{		
		Poly_vert *point=&p;

		if (point->sz < zbuffer[y*XMAX+x] ) 
		{
			bVisible=1;
			cbuffer[(y*XMAX+x)*3+0]=(float)point->r;
			cbuffer[(y*XMAX+x)*3+1]=(float)point->g;
			cbuffer[(y*XMAX+x)*3+2]=(float)point->b;

			zbuffer[y*XMAX+x]=(float)point->sz;
		}

		increment(&p, &dp, mask);
	}
}


/*
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
*/
static void poly_scan(register Poly *p, Window *win)
{
	register int i, li, ri, y, ly, ry, top, rem;
	register unsigned long mask;
	double ymin;
	Poly_vert l, r, dl, dr;
	
	if (p->n>POLY_NMAX) 
	{
		fprintf(stderr, "poly_scan: too many vertices: %d\n", p->n);
		return;
	}

	if (sizeof(Poly_vert)/sizeof(double) > 32) 
	{
		fprintf(stderr, "Poly_vert structure too big; must be <=32 doubles\n");
		exit(1);
	}
	
	/* find top vertex (y points down) */
	ymin = 1e18;
	for (i=0; i<p->n; i++)		
	{
		if (p->vert[i].sy < ymin) {
			ymin = p->vert[i].sy;
			top = i;
		}
	}
	
	li = ri = top;			/* left and right vertex indices */
	rem = p->n;				/* number of vertices remaining */
	y = (int)ceil(ymin-.5);			/* current scan line */
	ly = ry = y-1;			/* lower end of left & right edges */
	mask = p->mask & ~POLY_MASK(sy);	/* stop interpolating screen y */
	
	/* scan in y, activating new edges on left & right */
	/* as scan line passes over new vertices */
	while (rem>0) 
	{		
		/* advance left edge? */
		while (ly<=y && rem>0) 
		{	
			rem--;
			i = li-1;			/* step ccw down left side */
			if (i<0) i = p->n-1;
			incrementalize_y(&p->vert[li], &p->vert[i], &l, &dl, y, mask);
			ly = (int)floor(p->vert[i].sy+.5);
			li = i;
		}

		/* advance right edge? */
		while (ry<=y && rem>0) 
		{	
			rem--;
			i = ri+1;			/* step cw down right edge */
			if (i>=p->n) i = 0;
			incrementalize_y(&p->vert[ri], &p->vert[i], &r, &dr, y, mask);
			ry = (int)floor(p->vert[i].sy+.5);
			ri = i;
		}
		
		/* do scanlines till end of l or r edge */
		while (y<ly && y<ry) 
		{	    
			if (y>=win->y0 && y<=win->y1)
				if (l.sx<=r.sx) 
					scanline(y, &l, &r, win, mask);
				else		
					scanline(y, &r, &l, win, mask);

				y++;
				increment(&l, &dl, mask);
				increment(&r, &dr, mask);
		}
	}
}



/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */

static int CUR_DIM=0;


float* scan_init(int winx,int winy,float R,float G,float B)
{
	int i,j;
	/* screen clipping window */

	if ((CUR_DIM)<(winx*winy))
	{
		if (cbuffer) free(cbuffer);
		if (zbuffer) free(zbuffer);
		CUR_DIM=winx*winy;
		cbuffer=(float*)malloc(CUR_DIM*sizeof(float)*3);
		zbuffer=(float*)malloc(CUR_DIM*sizeof(float));
	}

	XMAX=winx;
	YMAX=winy;

	for (j=0;j<winy;j++)
	{
		for (i=0;i<winx;i++)
		{
			cbuffer[(j*winx+i)*3+0]=R;
			cbuffer[(j*winx+i)*3+1]=G;
			cbuffer[(j*winx+i)*3+2]=B;
			zbuffer[(j*winx+i)    ]=+1e18f;
		}
	}


	return cbuffer;
}


/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */
int scan_add_polygon(int npoints,float* v,float R,float G,float B)
{
	Window win;

	Poly p;
	int i;
	p.n=npoints;

	win.x0=0;
	win.y0=0;
	win.x1=XMAX-1;
	win.y1=YMAX-1;

	bVisible=0;

	for (i=0;i<npoints;i++)
	{
		p.vert[i].sx=*v++;
		p.vert[i].sy=*v++;
		p.vert[i].sz=*v++;
		p.vert[i].r=R;p.vert[i].g=G;p.vert[i].b=B;
	}

	p.mask = POLY_MASK(sx) | POLY_MASK(sy) | POLY_MASK(sz) | POLY_MASK(r) | POLY_MASK(g) | POLY_MASK(b);
	poly_scan(&p, &win);	
	return bVisible;
}




/*
int main()
{
	float p1[]={30, 30,0, 100, 30,0 ,  100,100,0, 30,100,0};
	float p2[]={10, 10,1, 200, 10,1,  200,200,1,  10,200,1};
	float* cbuffer;

	cbuffer=scan_init(1280,1024,1,1,1);
	scan_add_polygon(4,p1,0.3,0.7,0.4);
	scan_add_polygon(4,p2,0,0,0);
	scan_end();
}
*/
