/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _INERTIA_TENSOR_H__
#define _INERTIA_TENSOR_H__

#include "generaldef.h"

#include <string.h>

//calculates the inertia tensor from bounding balls (Radius,X0,X1,...)
//
//example:
// InertiaTensor it;
// it.addBall(b1);
// it.addBall(b2);
// ...
// it.solve();
// it.getEigenVector(0|1|...);it.getEigenValue(0|1|...);

class InertiaTensor
{
	real A      [(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)],Bar   [MAX_GEOMETRIC_DIMENSION+1];
	real Eigvec [(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)],Eigval[MAX_GEOMETRIC_DIMENSION+1];
	real G      [(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)],GT    [(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
	
	int  dim;
	
public:

	//constructor
	InertiaTensor(int dim);

	//reset
	void reset();

	//add a ball
	void addBall(real R,real* C);

	//solve the tensor
	void solve();

	//access the baricenter
	real* getBaricenter();

	//access eigenvector
	real* getEigenVector(int ref);

	//access eigen value
	real getEigenValue(int ref);

};

// INLINES

inline void InertiaTensor::reset()
	{memset(A,0,sizeof(A));memset(Bar,0,sizeof(Bar));}

inline real* InertiaTensor::getBaricenter()
	{return this->Bar;}

inline real* InertiaTensor::getEigenVector(int ref)
	{return &Eigvec[ref*dim];}

inline real InertiaTensor::getEigenValue(int ref)
	{return Eigval[ref];}


#endif // _INERTIA_TENSOR_H__

