
/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#ifdef WIN32
#include <windows.h>
#pragma warning(disable: 4305 4244)
#endif 

#ifdef __APPLE__
#include <GLUT/glut.h>
#endif

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "generaldef.h"
#include "viewer.h"
#include "trackball.h"

Win* _win=0;

const int BORDER=3;

static int default_win_sizx=0,default_win_sizy=0;
static int default_win_posx=0,default_win_posy=0;

static int _winhandle=0;

//-----------------------------------------------------------------
 void Win::run()
{
#ifdef __APPLE__
  close_flag=0;
  while(!close_flag) glutCheckLoop();
#else
	glutMainLoop();
#endif
}

#ifdef __APPLE__
void _close()  {
	_win->close();
	}
#endif

void _keydown(unsigned char key,int x, int y)                
	{_win->keydown(key,x,y);if (_win->need_redisplay) glutPostRedisplay();}

void _special(int key,int x,int y)                          
	{_win->special(key,x,y);if (_win->need_redisplay) glutPostRedisplay();}

void _reshape(int winx,int winy)                              
	{_win->reshape(winx,winy);if (_win->need_redisplay) glutPostRedisplay();}

void _mouse(int button,int state,int pickX,int pickY)  
	{_win->mouse(button,state,pickX,pickY);if (_win->need_redisplay) glutPostRedisplay();}

void _motion(int pickX, int pickY)                            
	{_win->motion(pickX,pickY);if (_win->need_redisplay) glutPostRedisplay();}

void _display()                                                   
	{_win->display();if (_win->need_redisplay) glutPostRedisplay();}

void Win::special(int key,int x, int y)           
	{getViewport()->special(key,x,y);}


//===========================================================
//===========================================================

Win::Win()
{
	const int W=1024;
	const int H=768;
	const int L=50  ;
	const int R=50;
	
#ifdef __APPLE__
	if(!_win) 
#else
	if (true)
#endif
	{
		int narg=1;
		char* args[]={"name"};

		glutInit(&narg,args);
	}
		
	glutInitDisplayMode(GLUT_DEPTH|GLUT_RGB|GLUT_DOUBLE | GLUT_STENCIL);
		
	glutInitWindowSize (W,H);
	_winhandle=glutCreateWindow("viewer");
	glutInitWindowPosition(L,R);
		
	glutKeyboardFunc(_keydown);
	glutSpecialFunc (_special);
	glutDisplayFunc (_display);
	glutMouseFunc   (_mouse);
	glutMotionFunc  (_motion);		
	glutReshapeFunc (_reshape);
		
#ifdef __APPLE__
	glutWMCloseFunc(_close);
#else
	glutSetOption ( GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_CONTINUE_EXECUTION ) ; //freeglut
#endif

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glDepthFunc(GL_LESS);
	glClearColor(1,1,1,1);
	
	this->winhandle=_winhandle;
	_win=this;
	this->options=0;
	this->loading=0;
	this->reshaping=-1;
	this->need_redisplay=1;
	this->aview=0;
	this->nview=0;
	this->W=W;
	this->H=H;
} 

#ifdef __APPLE__
void Win::close()
{
	glutDestroyWindow(this->winhandle);
	close_flag=1;
}
#endif

void Win::reshape(GLint winx,GLint winy)
{
	W=winx;
	H=winy;

	if (winx && winy)
		pack(pview);

	redisplay();
}


int Win::repack(float* pview,int& n,int pickX,int pickY,int X,int Y,int W,int H)
{
	
	int X1,X2,Y1,Y2,W1,W2,H1,H2;
	float MinF=0.05;

	if (!pview[n])
	{
		++n;
		return 0;
	}
	else if (pview[n]>0)
	{
		X1=X;
		W1=(int)((+1*pview[n])*W-BORDER);
		Y1=Y;
		H1=H;
		X2=(int)(X+(+1*pview[n])*W+BORDER);
		W2=X+W-X2;
		Y2=Y;
		H2=H;

		if (reshaping==-1 && (X1+W1)<=pickX && pickX<=X2 && pickY>=Y && pickY<=(Y+H))
			reshaping=n;

		if (reshaping==n)
		{
			this->pview[n]=(pickX-X1)/(float)W;
			if (this->pview[n]<MinF) this->pview[n]=MinF;
			if (this->pview[n]>(1-MinF)) this->pview[n]=(1-MinF);
			++n;
			return 1;
		}
	}
	else if (pview[n]<0)
	{
		float Pview=1-(-1*pview[n]);
		X2=X;
		W2=W;
		Y2=Y;
		H2=(int)((Pview)*H-BORDER);
		
		X1=X;
		W1=W;
		Y1=(int)(Y+Pview*H+BORDER);
		H1=(int)(Y+H-Y1);

		if (reshaping==-1 && (Y2+H2)<=pickY && pickY<=Y1 && pickX>=X && pickX<=(X+W))
			reshaping=n;

		if (reshaping==n)
		{
			this->pview[n]=-1*(1-(pickY-Y2)/(float)H);
			if (this->pview[n]>-1*MinF) this->pview[n]=-MinF;
			if (this->pview[n]<-1*(1-MinF)) this->pview[n]=-1*(1-MinF);
			++n;
			return 1;
		}
	}

	++n;

	return 
		repack(pview,n,pickX,pickY,X1,Y1,W1,H1) || 
		repack(pview,n,pickX,pickY,X2,Y2,W2,H2);
}


void Win::mouse(GLint button, GLint state, GLint pickX, GLint pickY)
{
	int n=0;

	if (state==GLUT_UP && reshaping!=-1)
	{
		repack(pview,n,pickX,H-pickY,BORDER<<1,BORDER<<1,W-(BORDER<<2),H-(BORDER<<2));
		pack(pview);
		reshaping=-1;
		return;
	}

	if (state==GLUT_DOWN && repack(pview,n,pickX,H-pickY,BORDER<<1,BORDER<<1,W-(BORDER<<2),H-(BORDER<<2)))
	{
		pack(pview);
		return;
	}

	if (state==GLUT_DOWN)
	{
		for (int i=0;i<nview;i++)
		{
			Panel* v=view[i];

			if (pickX>=v->X && pickX<=(v->X+v->W) && (H-pickY)>=v->Y && (H-pickY)<=(v->Y+v->H))
			{
				aview=i;
				break;
			}
		}
	}

	getViewport()->mouse(button,state,pickX,pickY);
}

void Win::motion(GLint pickX, GLint pickY)                     
{
	if (reshaping!=-1)
	{
		int n=0;
		repack(pview,n,pickX,H-pickY,BORDER<<1,BORDER<<1,W-(BORDER<<2),H-(BORDER<<2));
		pack(pview);
		return;
	}

	getViewport()->motion(pickX,pickY);
}

void Win::display()
{
	clock_t start, finish;
	start = clock();

	int i;
	glViewport(0,0,_win->W,_win->H);
	int W=_win->W;
	int H=_win->H;

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	//background(0,0,(int)(W),(int)(H),
	//	100,100,210,100,100,210,170,170,210,170,170,210
	//	);

	GLfloat white_ambient[]	= { 0.5, 0.5, 0.5, 1.0 };
   GLfloat white_diffuse[]	= { 0.5, 0.5, 0.5, 1.0 };
   GLfloat white_specular[] = { 1, 1.0, 1.0, 1.0 };
   GLfloat position_one[]	= { 1.0, 1.0, 1.0, 0.0 };
	glMaterialfv(GL_FRONT, GL_AMBIENT, white_ambient);
   glMaterialfv(GL_FRONT, GL_DIFFUSE, white_diffuse);
   glMaterialfv(GL_FRONT, GL_SPECULAR, white_specular);
   glMaterialf(GL_FRONT, GL_SHININESS, 20.0);
   glEndList();

   glLightfv(GL_LIGHT0, GL_POSITION, position_one);
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	//glLightModelf(GL_LIGHT_MODEL_TWO_SIDE,);

	need_redisplay=0;
	for (i=0;i<_win->nview;i++)
	{
		//if (i)
		{
			Panel* v=_win->getViewport(i);
			//glViewport((int)(v->X+1),(int)(v->Y+1),(int)(v->W-2),(int)(v->H-2));
			v->display();
		}
		//glDisable(GL_DEPTH_TEST);
	}
	//glEnable(GL_DEPTH_TEST);

	glViewport(0,0,_win->W,_win->H);
	//glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_LIGHTING);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_DEPTH_TEST);

	glColor3f(0,0,0);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glMatrixMode(GL_PROJECTION);glPushMatrix();glLoadIdentity();glOrtho(0, W, 0, H, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW) ;glPushMatrix();glLoadIdentity();

	glBegin(GL_LINE_LOOP);
	glVertex2i(0,0);glVertex2i(W-1,0);
	glVertex2i(W-1,H-1);glVertex2i(0,H-1);
	glEnd();

	/*
	for (i=0;i<_win->nview;i++)
	{
		Panel* v=_win->getViewport(i);
		if (!(v->options & Panel::DRAW_NO_BORDER))
		{
			glBegin(GL_LINE_LOOP);
			glVertex2i((int)(v->X),(int)(v->Y));glVertex2i((int)(v->X+v->W-1),(int)(v->Y));
			glVertex2i((int)(v->X+v->W-1),(int)(v->Y+v->H-1));glVertex2i((int)(v->X),(int)(v->Y+v->H-1));
			glEnd();
		}
	}
	*/

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glutSwapBuffers();

	finish = clock();

	if (options)
		options->set("fps",(int)(1/((finish-start)/(float)CLOCKS_PER_SEC)));
}



//===========================================================
//===========================================================
void Win::add(Panel* view)
{
	view->win=this;
	this->view[nview++]=view;

	if (!strcmp(view->name,"options"))
		this->options=(Options*)view;

}


//===========================================================
//===========================================================

void Win::_pack(float* pview,int& n,int& v,int X,int Y,int W,int H)
{
	int X1,X2,Y1,Y2,W1,W2,H1,H2;

	this->pview[n]=pview[n];

	if (!pview[n])
	{
		view[v]->X=X;view[v]->W=W;
		view[v]->Y=Y;view[v]->H=H;
		view[v]->reshape();
		++n;
		++v;
		return;
	}

	else if (pview[n]>0)
	{
		X1=X;
		W1=(int)((+1*pview[n])*W-BORDER);
		Y1=Y;
		H1=H;

		X2=(int)(X+(+1*pview[n])*W+BORDER);
		W2=X+W-X2;
		Y2=Y;
		H2=H;
	}
	else if (pview[n]<0)
	{
		float Pview=1-(-1*pview[n]);
		X2=X;
		W2=W;
		Y2=Y;
		H2=(int)((Pview)*H-BORDER);
		
		X1=X;
		W1=W;
		Y1=(int)(Y+Pview*H+BORDER);
		H1=(int)(Y+H-Y1);
	}

	++n;
	_pack(pview,n,v,X1,Y1,W1,H1);
	_pack(pview,n,v,X2,Y2,W2,H2);
}



//===========================================================
//===========================================================
void Win::pack(float* pview) 
{
	int n=0,v=0;

	//fix panel options
	pview[0]=-1*(1-20/(float)H);

	_pack(pview,n,v,BORDER<<1,BORDER<<1,W-(BORDER<<2),H-(BORDER<<2));
}


//===========================================================
//===========================================================
void Win::keydown(unsigned char key, GLint x, GLint y) 
{
	switch(key) 
	{
	case 27:
#ifndef __APPLE__
		default_win_sizx=glutGet(GLUT_WINDOW_WIDTH );
		default_win_sizy=glutGet(GLUT_WINDOW_HEIGHT );
		default_win_posx=glutGet(GLUT_WINDOW_X );
		default_win_posy=glutGet(GLUT_WINDOW_Y );
		glutLeaveMainLoop();
#else
		close_flag=1;
#endif
		return;

	case 'F':case 'f':
		//glutFullScreen();
		freeze=1-freeze;
		return;



	default:
		getViewport()->keydown(key,x,y);
		redisplay();
		break;
	}
}


//===========================================================
//===========================================================
void Win::text3d(GLfloat x,GLfloat y,GLfloat z,char* s)
{
	glRasterPos3f(x,y,z);
	for(char* c=s;*c;c++) glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *c);
}


//===========================================================
//===========================================================
void Win::text2d(GLint x, GLint y, char* s, GLfloat r, GLfloat g, GLfloat b)
{
	int lines;
	char* p;

	const int L=15;

	glColor3f(r,g,b);
	glRasterPos3i(x, y,1);

	for(p = s, lines = 0; *p; p++)
	{
		if (*p == 10 || *p==13) 
		{
			while (*p==10 || *p==13)
				p++;

			lines++;
			glRasterPos2i(x, y-(lines*L));
		}

		if (!strncmp(p,"rgb",3))
		{
			p=strchr(p,'(')+1;
			int ncolors=0;
			int colors[3]={0,0,0};

			while (1)
			{
				colors[ncolors++]=atoi(p);
				if (ncolors<3) 
					p=strchr(p,',')+1;
				else
					break;
			}
			p=strchr(p,')')+1;
			
			GLint rp[10];
			glGetIntegerv(GL_CURRENT_RASTER_POSITION,rp);
			glColor3f(colors[0]/255.0,colors[1]/255.0,colors[2]/255.0);
			glRasterPos3iv(rp);
		}

		
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *p);
	}

}

//===========================================================
//===========================================================
void Win::background(int x,int y,int winx,int winy,
							int r1,int g1,int b1,
							int r2,int g2,int b2,
							int r3,int g3,int b3,
							int r4,int g4,int b4)
{
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glMatrixMode(GL_PROJECTION);glPushMatrix();glLoadIdentity();glOrtho(x, x+winx, y, y+winy, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW) ;glPushMatrix();glLoadIdentity();
	
	glEnable(GL_POINT_SMOOTH);
	
	glBegin(GL_QUADS);
	glColor3f(r1/255.0f,g1/255.0f,b1/255.0f);glVertex2i(x			,y);			
	glColor3f(r2/255.0f,g2/255.0f,b2/255.0f);glVertex2i(x+winx	,y);			
	glColor3f(r3/255.0f,g3/255.0f,b3/255.0f);glVertex2i(x+winx	,y+winy);	
	glColor3f(r4/255.0f,g4/255.0f,b4/255.0f);glVertex2i(x			,y+winy);
	glEnd();
	glDisable(GL_POINT_SMOOTH);
	
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

//===========================================================
//===========================================================
void Win::box(GLfloat x1,GLfloat y1,GLfloat z1,GLfloat x2,GLfloat y2,GLfloat z2,GLfloat r,GLfloat g,GLfloat b)
{
	glColor3f(r,g,b);

	glBegin(GL_LINE_LOOP);
	glVertex3f(x1,y1,z1);glVertex3f(x2,y1,z1);glVertex3f(x2,y2,z1);glVertex3f(x1,y2,z1);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3f(x1,y1,z2);glVertex3f(x2,y1,z2);glVertex3f(x2,y2,z2);glVertex3f(x1,y2,z2);
	glEnd();

	glBegin(GL_LINES);
	glVertex3f(x1,y1,z1);glVertex3f(x1,y1,z2);glVertex3f(x2,y1,z1);glVertex3f(x2,y1,z2);
	glVertex3f(x1,y2,z1);glVertex3f(x1,y2,z2);glVertex3f(x2,y2,z1);glVertex3f(x2,y2,z2);
	glEnd();
}

//===========================================================
//===========================================================
void Win::move(int x1,int y1,int x2,int y2)
{
	glutSetWindow(winhandle);
	glutPositionWindow(x1,y1);
	glutReshapeWindow(x2-x1,y2-y1);

}


//-------------------------------------------------------------------------------------------

void Panel::transform(float& X,float& Y,float& Z)
{
	float *bbox=this->bbox;
	float scalefactor=this->scalefactor;
	float (*curmat)[4]=this->curmat;
	Panel::SavedMat  savedmat=this->savedmat;

	float tx=-0.5f*(bbox[0]+bbox[1]);
	float ty=-0.5f*(bbox[2]+bbox[3]);
	float tz=-0.5f*(bbox[4]+bbox[5]);

	float lenx=bbox[1]-bbox[0];
	float leny=bbox[3]-bbox[2];
	float lenz=bbox[5]-bbox[4];

	float f=0.75f/max3(lenx,leny,lenz);

	X=scalefactor*f*(X+tx);
	Y=scalefactor*f*(Y+ty);
	Z=scalefactor*f*(Z+tz);

	float _X=X*curmat[0][0]+Y*curmat[1][0]+Z*curmat[2][0]+1*curmat[3][0];
	float _Y=X*curmat[0][1]+Y*curmat[1][1]+Z*curmat[2][1]+1*curmat[3][1];
	float _Z=X*curmat[0][2]+Y*curmat[1][2]+Z*curmat[2][2]+1*curmat[3][2];
	float _O=X*curmat[0][3]+Y*curmat[1][3]+Z*curmat[2][3]+1*curmat[3][3];

	X=_X/_O;
	Y=_Y/_O;
	Z=_Z/_O;

	const float& left   = savedmat.glconf.ortho[0];
	const float& right  = savedmat.glconf.ortho[1];
	const float& bottom = savedmat.glconf.ortho[2];
	const float& top    = savedmat.glconf.ortho[3];
	const float& Near   = savedmat.glconf.ortho[4];
	const float& Far    = savedmat.glconf.ortho[5];

	float A =+2.0f/(right-left);
	float B =+2.0f/(top-bottom);
	float C =-2.0f/(Far-Near);
	float TX=+(right+left)/(right-left);
	float TY=+(top+bottom)/(top-bottom);
	float TZ=-(Far+Near)/(Far-Near);

	//float ortho[4][4]={ {A,0,0,0},{0,B,0,0},{0,0,C,0},{TX,TY,TZ,-1}};
	float ortho[4][4]={ {A,0,0,0},{0,B,0,0},{0,0,-1,0},{TX,TY,0,-1}};

	_X=X*ortho[0][0]+Y*ortho[1][0]+Z*ortho[2][0]+1*ortho[3][0];
	_Y=X*ortho[0][1]+Y*ortho[1][1]+Z*ortho[2][1]+1*ortho[3][1];
	_Z=X*ortho[0][2]+Y*ortho[1][2]+Z*ortho[2][2]+1*ortho[3][2];
	_O=X*ortho[0][3]+Y*ortho[1][3]+Z*ortho[2][3]+1*ortho[3][3];
	X=_X/_O;Y=_Y/_O;Z=_Z/_O;

	float Scale=100;
	X*=Scale;
	Y*=Scale;
	Z*=Scale;
}


//===========================================================
//===========================================================
void Panel::setNewRotation(int rot)
{
	memset(curmat,0,sizeof(float)*16);

	switch(rot)
	{
	case 0:
		vx[0] = 1.0; vx[1] = 0.0; vx[2] = 0.0;
		vy[0] = 0.0; vy[1] = 1.0; vy[2] = 0.0;
		vz[0] = 0.0; vz[1] = 0.0; vz[2] = 1.0;
		curmat[0][0]=curmat[1][1]=curmat[2][2]=curmat[3][3]=1;
		break;

	case 1:
		vx[0] = 0.0; vx[1] = 1.0; vx[2] = 0.0;
		vy[0] = 0.0; vy[1] = 0.0; vy[2] = 1.0;
		vz[0] = 1.0; vz[1] = 0.0; vz[2] = 0.0;
		curmat[0][1]=curmat[1][2]=curmat[2][0]=curmat[3][3]=1;
		break;

	case 2:
		vx[0] = 0.0; vx[1] = 0.0; vx[2] = 1.0;
		vy[0] = 1.0; vy[1] = 0.0; vy[2] = 0.0;
		vz[0] = 0.0; vz[1] = 1.0; vz[2] = 0.0;
		curmat[0][2]=curmat[1][0]=curmat[2][1]=curmat[3][3]=1;
		break;

	case 3:
		vx[0] = 1; vx[1] = 0; vx[2] = 0.0;
		vy[0] = 0; vy[1] = -1; vy[2] = 0.0;
		vz[0] = 0.0; vz[1] = 0.0; vz[2] = -1;
		curmat[0][0]=1;
		curmat[1][1]=curmat[2][2]=-1;
		curmat[3][3]=1;
		break;
	}


	savedmat.newmat=1;
}


//===========================================================
//===========================================================
void Panel::setbbox(float x1,float x2,float y1,float y2,float z1,float z2) 	
{
	this->bbox[0]=x1;this->bbox[1]=x2;
	this->bbox[2]=y1;this->bbox[3]=y2;
	this->bbox[4]=z1;this->bbox[5]=z2;
}


void Panel::setbbox(float* bbox) 	
{
	memcpy(this->bbox,bbox,sizeof(float)*6);
}

//===========================================================
//===========================================================
void Panel::add(Draw* draw)
{
	draw->panel=this;
	this->draw[ndraw++]=draw;
}

//===========================================================
//===========================================================
void Panel::display()
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();



	glOrtho(
		savedmat.glconf.ortho[0],savedmat.glconf.ortho[1],
		savedmat.glconf.ortho[2],savedmat.glconf.ortho[3],
		savedmat.glconf.ortho[4],savedmat.glconf.ortho[5]);

	glMatrixMode (GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(&(curmat[0][0]));
	glScalef(scalefactor,scalefactor,scalefactor);

	float lenx=bbox[1]-bbox[0];
	float leny=bbox[3]-bbox[2];
	float lenz=bbox[5]-bbox[4];

	float f=0.75f/max3(lenx,leny,lenz);
	glScalef(f,f,f);
	glTranslatef(-0.5f*bbox[0]-0.5f*bbox[1],-0.5f*bbox[2]-0.5f*bbox[3],-0.5f*bbox[4]-0.5f*bbox[5]);

	
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3f(1,0,0);glVertex3f(0,0,0);glVertex3f(1,0,0);
	glColor3f(0,1,0);glVertex3f(0,0,0);glVertex3f(0,1,0);
	glColor3f(0,0,1);glVertex3f(0,0,0);glVertex3f(0,0,1);
	glEnd();
	glEnable(GL_LIGHTING);

	for (int i=0;i<ndraw;i++)
		draw[i]->display();

	grabMatrices();
	
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

//===========================================================
//===========================================================
void Panel::rotate(GLfloat beginx,GLfloat beginy,GLfloat x,GLfloat y)
{
	float lastquat[4];
	float	tmat[4][4];

	beginy=win->getWindowHeight()-beginy;
	     y=win->getWindowHeight()-y;

	float p1x=-1+2*(beginx-X)/(float)W;
	float p2x=-1+2*(     x-X)/(float)W;
	
	float p1y=-1+2*(beginy-Y)/(float)H;
	float p2y=-1+2*(     y-Y)/(float)H;

	trackball(lastquat,p1x,p1y,p2x,p2y);

	build_rotmatrix(tmat,lastquat); 
	float newmat[4][4];
	matxmat(newmat,curmat,tmat);
	for (int i=0;i<4;i++)
	for (int j=0;j<4;j++)
		curmat[i][j]=newmat[i][j];

	savedmat.newmat=1;
	win->redisplay();
}

//===========================================================
//===========================================================
void Panel::ortho(GLfloat dx,GLfloat dy)
{
	dx/=(double)W;
	dy/=(double)H;

	savedmat.glconf.ortho[0]+=dx;
	savedmat.glconf.ortho[1]+=dx;
	savedmat.glconf.ortho[3]-=dy;
	savedmat.glconf.ortho[2]-=dy;

	savedmat.newmat=1;
	win->redisplay();
}

void Panel::getViewpoint(float& x,float &y,float& z)
{
	x=savedmat.glconf.eye[1];
	y=savedmat.glconf.eye[2];
	z=savedmat.glconf.eye[3];
}


//===========================================================
//===========================================================
void Panel::reshape()
{	

	/*
	if (win->loading)
		;
	else if (win->reshaping!=-1)
	*/
	{

		if (W < H) 
		{
			float factor=((float)W/(float)H);
			float other=fabs(savedmat.glconf.ortho[3]-savedmat.glconf.ortho[2]);
			float middle=(savedmat.glconf.ortho[0]+savedmat.glconf.ortho[1])/2.0;
			savedmat.glconf.ortho[0]=middle-0.5*factor*other;
			savedmat.glconf.ortho[1]=middle+0.5*factor*other;
		}
		else 
		{
			float factor=((float)H/(float)W);
			float other=fabs(savedmat.glconf.ortho[1]-savedmat.glconf.ortho[0]);
			float middle=(savedmat.glconf.ortho[3]+savedmat.glconf.ortho[2])/2.0;
			savedmat.glconf.ortho[2]=middle-0.5*factor*other;
			savedmat.glconf.ortho[3]=middle+0.5*factor*other;
		}
	}
	/*
	else
	{

		savedmat.glconf.ortho[0]=-1.f;
		savedmat.glconf.ortho[1]=+1.f;
		savedmat.glconf.ortho[2]=-1.f;
		savedmat.glconf.ortho[3]=+1.f;
		
		if (W > H) 
		{
			savedmat.glconf.ortho[0]*=((float)W/(float)H);
			savedmat.glconf.ortho[1]*=((float)W/(float)H);
		}
		else 
		{
			savedmat.glconf.ortho[3]*=((float)H/(float)W);
			savedmat.glconf.ortho[2]*=((float)H/(float)W);
		}
	}
	*/
	savedmat.newmat=1;
	win->redisplay();
}


//===========================================================
//===========================================================
void Panel::mouse(GLint button, GLint state, GLint x, GLint y)
{
	//left
	if (button==GLUT_LEFT_BUTTON)
	{
		if (state==GLUT_DOWN)
		{
			scaling=(glutGetModifiers() & GLUT_ACTIVE_SHIFT);
			moving =!scaling;
			beginx = x;
			beginy = y;
		}
		else
		{
			moving=scaling=0;
			win->redisplay();
		}
	}

	//middle
	else if (button==GLUT_MIDDLE_BUTTON)
	{
		if (state==GLUT_DOWN){
			panning=1;
			beginx = x;beginy = y;
		}
		else{
			panning=0;
			win->redisplay();
		}
	}
	else
	{
		for (int i=0;i<ndraw;i++)
			draw[i]->mouse(button,state,x,y);
	}
}

//===========================================================
//===========================================================
void Panel::motion(GLint x, GLint y)
{
	if (panning)
	{
		ortho((beginx-x),(beginy-y));
		savedmat.newmat=1;
		beginx = x;beginy = y;
		return;
	}
	else if (scaling) 
	{
		scale(beginx-x);
		savedmat.newmat=1;
		beginx = x;beginy = y;
		return;
	}
	else if (moving) 
	{
		this->rotate(beginx,beginy,x,y);

		//this->rotate(beginx-X,beginy-Y,x-X,y-Y);

		savedmat.newmat=1;
		beginx = x;beginy = y;
		return;
	}
	else
	{
		for (int i=0;i<ndraw;i++)
			draw[i]->motion(x,y);
	}
}


//===========================================================
//===========================================================

int Win::save(int version,FILE* file)
{
	char hwrite[5]="cwin";
	if (fwrite(hwrite,1,5,file)!=5)
		return -1;

	int X      =(int)glutGet(GLUT_WINDOW_X     );
	int Y      =(int)glutGet(GLUT_WINDOW_Y      );
	int screenw=(int)glutGet(GLUT_SCREEN_WIDTH );
	int screenh=(int)glutGet(GLUT_SCREEN_HEIGHT);

	if (fwrite(      &(X),sizeof(int),1,file)!=1 ||
		 fwrite(      &(Y),sizeof(int),1,file)!=1 ||
		 fwrite(&(screenw),sizeof(int),1,file)!=1 ||
		 fwrite(&(screenh),sizeof(int),1,file)!=1 ||
		 fwrite(&(this->W),sizeof(int),1,file)!=1 ||
		 fwrite(&(this->H),sizeof(int),1,file)!=1)
		return -1;

	if (fwrite(this->pview,sizeof(float),this->nview,file)!=(size_t)this->nview)
		return -1;

	for (int i=0;i<nview;i++)
	{
		if (view[i]->save(version,file))
			return -1;
	}

	return 0;
}

//===========================================================
//===========================================================
int Win::open(int version,FILE* file)
{
	char htmp[5],hread[5]="cwin";
	if (fread(htmp,1,5,file)!=5 || strcmp(htmp,hread)) 
		return -1;

	this->loading=1;

	int X,Y,screenw,screenh,W,H;

	if (fread(      &X,sizeof(int),1,file)!=1 ||
		 fread(      &Y,sizeof(int),1,file)!=1 ||
		 fread(&screenw,sizeof(int),1,file)!=1 ||
		 fread(&screenh,sizeof(int),1,file)!=1 ||
		 fread(      &W,sizeof(int),1,file)!=1 ||
		 fread(      &H,sizeof(int),1,file)!=1)
	{
		this->loading=0;
		return -1;
	}
	
	if (X==glutGet(GLUT_WINDOW_X) &&
		 Y==glutGet(GLUT_WINDOW_Y) &&
		 glutGet(GLUT_SCREEN_WIDTH )==screenw && 
		 glutGet(GLUT_SCREEN_HEIGHT)==screenh && 
		 this->W==W && 
		 this->H==H)
	{
		;//nothing to do
	}
	else
	{
		float rx= glutGet(GLUT_SCREEN_WIDTH )/(float)screenw;
		float ry= glutGet(GLUT_SCREEN_HEIGHT)/(float)screenh;
		float r=min2(rx,ry);

		X=(int)(X*r);
		Y=(int)(Y*r);
		W=(int)(W*r);
		H=(int)(H*r);

		glutPositionWindow(X,Y);
		glutReshapeWindow(W,H);

		this->W=W;
		this->H=H;
	}

	if (fread(this->pview,sizeof(float),this->nview,file)!=(size_t)this->nview)
	{
		this->loading=0;
		return -1;
	}

	for (int i=0;i<nview;i++)
	{
		if (view[i]->open(version,file))
		{
			this->loading=0;
			return -1;
		}
	}

	pack(this->pview);
	redisplay();
	this->loading=0;
	return 0;
}


//===========================================================
//===========================================================

int Panel::save(int version,FILE* file)
{
	char hwrite[5]="cpan";
	if (fwrite(hwrite,1,5,file)!=5)
		return -1;

	if (fwrite(this->curmat,sizeof(float),16,file)!=16)
		return -1;
		
	if (fwrite(this->vx,sizeof(float),3,file)!=3)
		return -1;
		
	if (fwrite(this->vy,sizeof(float),3,file)!=3)
		return -1;		
		
	if (fwrite(this->vz,sizeof(float),3,file)!=3)
		return -1;		
		
	if (fwrite(&this->scalefactor,sizeof(float),1,file)!=1)
		return -1;	
		
	if (fwrite(this->savedmat.glconf.ortho,sizeof(float),6,file)!=6)
		return -1;	
	
	for (int i=0;i<ndraw;i++)
	{
		if (draw[i]->save(version,file))
			return -1;
	}

	return 0;
}


//===========================================================
//===========================================================

int Panel::open(int version,FILE* file)
{
	char htmp[5],hread[5]="cpan";
	if (fread(htmp,1,5,file)!=5 || strcmp(htmp,hread)) 
		return -1;


	if (fread(this->curmat,sizeof(float),16,file)!=16)
		return -1;
		
	if (fread(this->vx,sizeof(float),3,file)!=3)
		return -1;
		
	if (fread(this->vy,sizeof(float),3,file)!=3)
		return -1;		
		
	if (fread(this->vz,sizeof(float),3,file)!=3)
		return -1;		
		
	if (fread(&this->scalefactor,sizeof(float),1,file)!=1)
		return -1;
		
	if (fread(this->savedmat.glconf.ortho,sizeof(float),6,file)!=6)
		return -1;	

	for (int i=0;i<ndraw;i++)
	{
		if (draw[i]->open(version,file))
			return -1;
	}

	return 0;
}


//===========================================================
//===========================================================
void Panel::scale(GLfloat d)
{
	scalefactor = scalefactor*(1.0+float(d)/W);
	savedmat.newmat=1;
	win->redisplay();
}


//===========================================================
//===========================================================
void Panel::grabMatrices(bool bForce)
{
	if (bForce)
		savedmat.newmat=1;

	if (savedmat.newmat)
	{
		/* special code for bspheres */
		glGetDoublev (GL_MODELVIEW_MATRIX,  savedmat.glconf.model);
		glGetDoublev (GL_PROJECTION_MATRIX, savedmat.glconf.proj);
		glGetIntegerv(GL_VIEWPORT,          savedmat.glconf.view);

		/* eye position */
		double px=W/2.0,py=H/2.0;
		float reversepy=(float)(H-py);
		GLdouble eyex,eyey,eyez;
		gluUnProject(px,reversepy,0.0,savedmat.glconf.model,savedmat.glconf.proj,savedmat.glconf.view,&eyex,&eyey,&eyez);
		savedmat.glconf.eye[0]=1;
		savedmat.glconf.eye[1]=eyex;
		savedmat.glconf.eye[2]=eyey;
		savedmat.glconf.eye[3]=eyez;
		savedmat.newmat=0;
		++(savedmat.viewpoint);
	}
}



//===========================================================
//===========================================================
void Panel::convertScreenPoint(const GLint& px,const GLint& py,
							 GLdouble& x1,GLdouble& y1, GLdouble& z1,
							 GLdouble& x2,GLdouble& y2, GLdouble& z2)
{
	float reversepy=(float)(H-py);
	gluUnProject(px,reversepy,0.0,savedmat.glconf.model,savedmat.glconf.proj,savedmat.glconf.view,&x1,&y1,&z1);
	gluUnProject(px,reversepy,1.0,savedmat.glconf.model,savedmat.glconf.proj,savedmat.glconf.view,&x2,&y2,&z2);
}


//===========================================================
//===========================================================
Panel::Panel(const char* name,int options)
{
	this->options=options;
	this->ndraw=0;

	this->bbox[0]=-0.5f;
	this->bbox[1]=+0.5f;
	this->bbox[2]=-0.5f;
	this->bbox[3]=+0.5f;
	this->bbox[4]=-0.5f;
	this->bbox[5]=+0.5f;

	strcpy(this->name,name);

	this->win = 0;

	this->scaling  = false;
	this->moving   = false;
	this->panning  = false;
	this->beginx   = 0;
	this->beginy   = 0;

	savedmat.glconf.ortho[0]=-1.f;
	savedmat.glconf.ortho[1]=+1.f;
	savedmat.glconf.ortho[2]=-1.f;
	savedmat.glconf.ortho[3]=+1.f;
	savedmat.glconf.ortho[4] =-100;
	savedmat.glconf.ortho[5] =+100;
	
	//resetOrtho();

	matident(curmat);

	curmat[0][0]=-6.411653e-001f;curmat[0][1]=-4.397923e-001f;curmat[0][2]=6.288798e-001f;curmat[0][3]=0.000000e+000f;
	curmat[1][0]=7.673997e-001f;curmat[1][1]=-3.650807e-001f;curmat[1][2]=5.270808e-001f;curmat[1][3]=0.000000e+000f;
	curmat[2][0]=-2.214402e-003f;curmat[2][1]=8.205479e-001f;curmat[2][2]=5.715732e-001f;curmat[2][3]=0.000000e+000f;
	curmat[3][0]=0.000000e+000f;curmat[3][1]=0.000000e+000f;curmat[3][2]=0.000000e+000f;curmat[3][3]=1.000000e+000f;

	vx[0] = (float)1.0; vx[1] = (float)0.0; vx[2] =(float) 0.0;
	vy[0] = (float)0.0; vy[1] = (float)1.0; vy[2] =(float) 0.0;
	vz[0] = (float)0.0; vz[1] = (float)0.0; vz[2] =(float) 1.0;

	scalefactor =2;
	
	savedmat.newmat=1;
	savedmat.viewpoint  =0;
}


//===========================================================
//===========================================================
void Panel::mkvectors()
{
	GLdouble	p0[3] = {W*0.5,   H*0.5  , 0};
	GLdouble	px[3] = {W*0.5+1, H*0.5  , 0};
	GLdouble	py[3] = {W*0.5,   H*0.5+1, 0};
	GLdouble	pz[3] = {W*0.5,   H*0.5, 0+(1./W)};
	GLdouble	pp0[3],ppx[3],ppy[3],ppz[3];

	gluUnProject(p0[0],p0[1],p0[2],savedmat.glconf.model,savedmat.glconf.proj,
		savedmat.glconf.view,pp0+0,pp0+1,pp0+2);

	gluUnProject(px[0],px[1],px[2],savedmat.glconf.model,savedmat.glconf.proj,
		savedmat.glconf.view,ppx+0,ppx+1,ppx+2);
	
	gluUnProject(py[0],py[1],py[2],savedmat.glconf.model,savedmat.glconf.proj,
		savedmat.glconf.view,ppy+0,ppy+1,ppy+2);
	
	gluUnProject(pz[0],pz[1],pz[2],savedmat.glconf.model,savedmat.glconf.proj,
		savedmat.glconf.view,ppz+0,ppz+1,ppz+2);
	
	for(int i=0;i<3;i++) 
	{
		vx[i] = (float)(ppx[i] - pp0[i]);
		vy[i] = (float)(ppy[i] - pp0[i]);
		vz[i] = (float)(ppz[i] - pp0[i]);
	}
}


//===========================================================
//===========================================================
void Panel::resetMatrix()
{
	matident(curmat);
	vx[0] = (float)1.0; vx[1] = (float)0.0; vx[2] =(float) 0.0;
	vy[0] = (float)0.0; vy[1] = (float)1.0; vy[2] =(float) 0.0;
	vz[0] = (float)0.0; vz[1] = (float)0.0; vz[2] =(float) 1.0;
	savedmat.newmat=1;
}


//===========================================================
//===========================================================

/*
void panel::resetOrtho()
{
	savedmat.glconf.ortho[0]=-1.f;
	savedmat.glconf.ortho[1]=+1.f;
	savedmat.glconf.ortho[2]=-1.f;
	savedmat.glconf.ortho[3]=+1.f;

	printf("PRIMA %f %f %f %f %f %f\n",
		(float)savedmat.glconf.ortho[0],
		(float)savedmat.glconf.ortho[1],
		(float)savedmat.glconf.ortho[2],
		(float)savedmat.glconf.ortho[3],
		(float)savedmat.glconf.ortho[4],
		(float)savedmat.glconf.ortho[5]);

	
	if (W > H) 
	{
		savedmat.glconf.ortho[0]*=((float)W/(float)H);
		savedmat.glconf.ortho[1]*=((float)W/(float)H);
	}
	else 
	{
		savedmat.glconf.ortho[3]*=((float)H/(float)W);
		savedmat.glconf.ortho[2]*=((float)H/(float)W);
	}

	printf("DOPO %f %f %f %f %f %f\n",
		(float)savedmat.glconf.ortho[0],
		(float)savedmat.glconf.ortho[1],
		(float)savedmat.glconf.ortho[2],
		(float)savedmat.glconf.ortho[3],
		(float)savedmat.glconf.ortho[4],
		(float)savedmat.glconf.ortho[5]);


	savedmat.newmat=1;
	if (win) win->redisplay();
}
*/

//===========================================================
//===========================================================
void Panel::special(int key, int x, int y)
{
	;
}


//===========================================================
//===========================================================
void Panel::keydown(unsigned char key, GLint x, GLint y)
{

	switch(key) 
	{
	case 's':case 'S': //reset matrix
		resetMatrix();
		win->redisplay();
		break;

	case '1':
	case '2':
	case '3':
	case '4':
		resetMatrix();
		setNewRotation(key-'1');
		win->redisplay();
		break;
	case '?':
	case 'h':
	case 'H':
			this->help();
		break;

	//default:
	//	win->keydown(key,x,y);
	}
}


void Panel::help(FILE * out){
	fprintf(out,"\n------------------------------------------------\n");
	fprintf(out,"'?'/'h'/'H': Show this help.\n");
	fprintf(out,"'W'/'w'    : Toggle wireframe edge display\n");
	fprintf(out,"\n------------------------------------------------\n");
	fflush(out);
}


//===========================================================
//===========================================================

Options::Options()
:Panel("options")
{
	this->n=0;
	this->a=0;
}

int Options::find(char* name)
{
	for (int i=0;i<n;i++)
	{
		if (!strcmp(options[i].name,name))
			return i;
	}
	return -1;
}

void Options::setActual(char* name)
{
	for (int i=0;i<n;i++)
	{
		if (!strcmp(options[i].name,name))
		{
			a=i;
			return;
		}
	}
}


void Options::set(char* name,bool value,char* explanation)
{
  int i;
	for (i=0;i<n;i++) 
		if (strcmp(this->options[i].name,name)==0) 
		{
			options[n].format=BOOL;
			options[i].value.b=value;
			if (explanation) strcpy(options[i].explanation,explanation);
			return;
		}

	options[n].format=BOOL;
	options[n].value.b=value;
	strcpy(options[n].name,name);
	options[n].explanation[0]=0;
	if (explanation) strcpy(options[i].explanation,explanation);
	n++;
}

void Options::set(char* name,int value,char* explanation)
{
  int i;
	for (i=0;i<n;i++) 
	{
		if (strcmp(this->options[i].name,name)==0) 
		{
			options[n].format=INT;
			options[i].value.i=value;
			if (explanation) strcpy(options[i].explanation,explanation);
			return;
		}
	}

	options[n].format=INT;
	options[n].value.i=value;
	strcpy(options[n].name,name);
	options[n].explanation[0]=0;
	if (explanation) strcpy(options[i].explanation,explanation);
	n++;

}


void Options::set(char* name,float value,char* explanation)
{
  int i;
	for (i=0;i<n;i++) 
	{
		if (strcmp(this->options[i].name,name)==0) 
		{
			options[n].format=FLOAT;
			options[i].value.f=value;
			if (explanation) strcpy(options[i].explanation,explanation);
			return;
		}
	}

	options[n].format=FLOAT;
	options[n].value.f=value;
	strcpy(options[n].name,name);
	options[n].explanation[0]=0;
	if (explanation) strcpy(options[i].explanation,explanation);
	n++;

}


void Options::set(char* name,char* value,char* explanation)
{
  int i;
	for (i=0;i<n;i++) 
	{
		if (strcmp(this->options[i].name,name)==0) 
		{
			options[n].format=STRING;
			strcpy(options[i].value.s,value);
			if (explanation) strcpy(options[i].explanation,explanation);
			return;
		}
	}	

	options[n].format=STRING;
	strcpy(options[n].value.s,value);
	strcpy(options[n].name,name);
	options[n].explanation[0]=0;
	if (explanation) strcpy(options[i].explanation,explanation);
	n++;
}

void Options::display()
{
	char s[2048*8];
	s[0]=0;

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, win->getWindowWidth(), 0, win->getWindowHeight(), -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	for (int cont=0;cont<n;cont++)
	{
		int i=(a+cont)%n;

		if (cont)
			sprintf(s+strlen(s),"%s ",options[i].name);
		else
			sprintf(s+strlen(s),"rgb(200,200,40)%srgb(255,30,30) ",options[i].name);
	

		switch(options[i].format)
		{
		case BOOL :sprintf(s+strlen(s),"%s ",options[i].value.b?"on":"off");break;
		case INT  :sprintf(s+strlen(s),"%d ",options[i].value.i);break;
		case FLOAT:
			{
				float value=options[i].value.f;

				if ((float)((int)value)==value)
					sprintf(s+strlen(s),"%d ",(int)value);
				else
					sprintf(s+strlen(s),"%f ",(float)value);
				break;
			}

		case STRING:
			{
				sprintf(s+strlen(s),"%s ",options[i].value.s);
			}
		}

		if (!cont)
			sprintf(s+strlen(s),"rgb(255,255,255)%s rgb(60,60,60)",options[i].explanation);

		break;
	}

	int x=0;
	int y=0;
	int W=win->getWindowWidth();
	int H=win->getWindowHeight();

	glEnable(GL_BLEND);
	glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA) ;

	glBegin(GL_QUADS);
	glColor4ub(0,0,0,100);glVertex2i(0,0);
	glColor4ub(0,0,0,100);glVertex2i(W,0);
	glColor4ub(0,0,0,100);glVertex2i(W,20);
	glColor4ub(0,0,0,100);glVertex2i(0,20);
	glEnd();
	glDisable(GL_BLEND);

	win->text2d((int)X,(int)Y,s,0.1,0.1,0.1);

	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);

	grabMatrices();
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);


}

